/*=============================================================================
	OpenGL.cpp: Unreal OpenGL support implementation for Windows and Linux.
	Copyright 1999 Epic Games, Inc. All Rights Reserved.

	OpenGL renderer by Daniel Vogel <vogel@lokigames.com>
	Loki Software, Inc.

	Other URenderDevice subclasses include:
	* USoftwareRenderDevice: Software renderer.
	* UGlideRenderDevice: 3dfx Glide renderer.
	* UDirect3DRenderDevice: Direct3D renderer.
	* UOpenGLRenderDevice: OpenGL renderer.

	Revision history:
	* Created by Daniel Vogel based on XMesaGLDrv
	* Changes (John Fulmer, Jeroen Janssen)
	* Major changes (Daniel Vogel)
	* Ported back to Win32 (Fredrik Gustafsson)
	* Unification and addition of vertex arrays (Daniel Vogel)
	* Actor triangle caching (Steve Sinclair)
	* One pass fogging (Daniel Vogel)
	* Windows gamma support (Daniel Vogel)
	* 2X blending support (Daniel Vogel)
	* Better detail texture handling (Daniel Vogel)
	* Scaleability (Daniel Vogel)
	* Texture LOD bias (Daniel Vogel)	
	* RefreshRate support on Windows (Jason Dick)
	* Finer control over gamma (Daniel Vogel)
	* (NOT ALWAYS) Fixed Windows bitdepth switching (Daniel Vogel)


	UseTNT			workaround for buggy TNT/TNT2 drivers
	UseTrilinear		whether to use trilinear filtering			
	UseAlphaPalette		set to 0 for buggy drivers (GeForce)
	UseS3TC			whether to use compressed textures
	Use4444Textures		speedup for real low end cards (G200)
	MaxAnisotropy		maximum level of anisotropy used
	UseFilterSGIS		whether to use the SGIS filters
	MaxTMUnits		maximum number of TMUs UT will try to use
	LODBias			texture lod bias
	RefreshRate		requested refresh rate (Windows only)
	GammaOffset		offset for the gamma correction


TODO:
	- DOCUMENTATION!!! (especially all subtle assumptions)
	- get rid of some unnecessary state changes (#ifdef out)

=============================================================================*/

#include "OpenGLDrv.h"
#include <math.h>
#include <stdio.h>


#include <utility>
#include <map>
#pragma warning(disable : 4663)
#pragma warning(disable : 4018)
#include <vector>

#pragma warning(disable : 4244)
#pragma warning(disable : 4245)

#pragma warning(disable : 4146)
#include <sstream>
#include <iostream>
#include <iomanip>

//Optional SSE code
#define INCLUDE_SSE_CODE

#ifdef INCLUDE_SSE_CODE
#include <xmmintrin.h>
#endif


//Build defines
#define UTGLR_DX_BUILD

//Experimental code defines
//#define UTGLR_SKYBOX_HACK

//Debug defines
//#define UTGLR_DEBUG_SHOW_TEX_CONVERT_COUNTS
//#define UTGLR_DEBUG_SHOW_CALL_COUNTS

// gruenschloss:
//#define UTGLR_GRUENSCHLOSS_DEBUG
//#define UTGLR_GRUENSCHLOSS_DEBUG_WIREFRAME

/*-----------------------------------------------------------------------------
	Globals.
-----------------------------------------------------------------------------*/

#define MAX_TMUNITS   4			// vogel: maximum number of texture mapping units supported

#define VERTEX_ARRAY_SIZE 2000		// vogel: better safe than sorry

#ifdef WIN32
#define STDGL 1					// Use standard GL driver or minidriver by default
#define DYNAMIC_BIND 1			// If 0, must static link to OpenGL32, Gdi32
#define GL_DLL (STDGL ? "OpenGL32.dll" : "3dfxgl.dll")
#endif
	

/*-----------------------------------------------------------------------------
	Vertex programs.
-----------------------------------------------------------------------------*/

static const char *g_vpComplexSurfaceSingleTexture =
	"!!ARBvp1.0\n"
	"OPTION ARB_position_invariant;\n"

	"ATTRIB iPos = vertex.position;\n"
	"ATTRIB fxInfo = vertex.attrib[6];\n"
	"ATTRIB fyInfo = vertex.attrib[7];\n"
	"ATTRIB texInfo0 = vertex.texcoord[0];\n"
	"PARAM defTexCoord = { 0, 0, 0, 1 };\n"
	"TEMP t1;\n"
	"TEMP t2;\n"
	"OUTPUT oTex0 = result.texcoord[0];\n"

	"MOV result.color, vertex.color;\n"

	"DP3 t1.x, fxInfo, iPos;\n"
	"DP3 t1.y, fyInfo, iPos;\n"
	"SUB t1.x, t1.x, fxInfo.w;\n"
	"SUB t1.y, t1.y, fyInfo.w;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo0.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo0.zwzw;\n"
	"MOV oTex0, t2;\n"

	"END\n"
;

static const char *g_vpComplexSurfaceDoubleTexture =
	"!!ARBvp1.0\n"
	"OPTION ARB_position_invariant;\n"

	"ATTRIB iPos = vertex.position;\n"
	"ATTRIB fxInfo = vertex.attrib[6];\n"
	"ATTRIB fyInfo = vertex.attrib[7];\n"
	"ATTRIB texInfo0 = vertex.texcoord[0];\n"
	"ATTRIB texInfo1 = vertex.texcoord[1];\n"
	"PARAM defTexCoord = { 0, 0, 0, 1 };\n"
	"TEMP t1;\n"
	"TEMP t2;\n"
	"OUTPUT oTex0 = result.texcoord[0];\n"
	"OUTPUT oTex1 = result.texcoord[1];\n"

	"MOV result.color, vertex.color;\n"

	"DP3 t1.x, fxInfo, iPos;\n"
	"DP3 t1.y, fyInfo, iPos;\n"
	"SUB t1.x, t1.x, fxInfo.w;\n"
	"SUB t1.y, t1.y, fyInfo.w;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo0.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo0.zwzw;\n"
	"MOV oTex0, t2;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo1.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo1.zwzw;\n"
	"MOV oTex1, t2;\n"

	"END\n"
;

static const char *g_vpComplexSurfaceTripleTexture =
	"!!ARBvp1.0\n"
	"OPTION ARB_position_invariant;\n"

	"ATTRIB iPos = vertex.position;\n"
	"ATTRIB fxInfo = vertex.attrib[6];\n"
	"ATTRIB fyInfo = vertex.attrib[7];\n"
	"ATTRIB texInfo0 = vertex.texcoord[0];\n"
	"ATTRIB texInfo1 = vertex.texcoord[1];\n"
	"ATTRIB texInfo2 = vertex.texcoord[2];\n"
	"PARAM defTexCoord = { 0, 0, 0, 1 };\n"
	"TEMP t1;\n"
	"TEMP t2;\n"
	"OUTPUT oTex0 = result.texcoord[0];\n"
	"OUTPUT oTex1 = result.texcoord[1];\n"
	"OUTPUT oTex2 = result.texcoord[2];\n"

	"MOV result.color, vertex.color;\n"

	"DP3 t1.x, fxInfo, iPos;\n"
	"DP3 t1.y, fyInfo, iPos;\n"
	"SUB t1.x, t1.x, fxInfo.w;\n"
	"SUB t1.y, t1.y, fyInfo.w;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo0.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo0.zwzw;\n"
	"MOV oTex0, t2;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo1.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo1.zwzw;\n"
	"MOV oTex1, t2;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo2.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo2.zwzw;\n"
	"MOV oTex2, t2;\n"

	"END\n"
;

static const char *g_vpComplexSurfaceQuadTexture =
	"!!ARBvp1.0\n"
	"OPTION ARB_position_invariant;\n"

	"ATTRIB iPos = vertex.position;\n"
	"ATTRIB fxInfo = vertex.attrib[6];\n"
	"ATTRIB fyInfo = vertex.attrib[7];\n"
	"ATTRIB texInfo0 = vertex.texcoord[0];\n"
	"ATTRIB texInfo1 = vertex.texcoord[1];\n"
	"ATTRIB texInfo2 = vertex.texcoord[2];\n"
	"ATTRIB texInfo3 = vertex.texcoord[3];\n"
	"PARAM defTexCoord = { 0, 0, 0, 1 };\n"
	"TEMP t1;\n"
	"TEMP t2;\n"
	"OUTPUT oTex0 = result.texcoord[0];\n"
	"OUTPUT oTex1 = result.texcoord[1];\n"
	"OUTPUT oTex2 = result.texcoord[2];\n"
	"OUTPUT oTex3 = result.texcoord[3];\n"

	"MOV result.color, vertex.color;\n"

	"DP3 t1.x, fxInfo, iPos;\n"
	"DP3 t1.y, fyInfo, iPos;\n"
	"SUB t1.x, t1.x, fxInfo.w;\n"
	"SUB t1.y, t1.y, fyInfo.w;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo0.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo0.zwzw;\n"
	"MOV oTex0, t2;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo1.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo1.zwzw;\n"
	"MOV oTex1, t2;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo2.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo2.zwzw;\n"
	"MOV oTex2, t2;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo3.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo3.zwzw;\n"
	"MOV oTex3, t2;\n"

	"END\n"
;

static const char *g_vpComplexSurfaceDetailAlpha =
	"!!ARBvp1.0\n"
	"OPTION ARB_position_invariant;\n"

	"ATTRIB iPos = vertex.position;\n"
	"ATTRIB fxInfo = vertex.attrib[6];\n"
	"ATTRIB fyInfo = vertex.attrib[7];\n"
	"ATTRIB texInfo1 = vertex.texcoord[1];\n"
	"PARAM defTexCoord = { 0, 0.5, 0, 1 };\n" //Shared with initializing V coord for alpha texture
	"PARAM RNearZ = 0.002631578947;\n"
	"TEMP t1;\n"
	"TEMP t2;\n"
	"OUTPUT oTex0 = result.texcoord[0];\n"
	"OUTPUT oTex1 = result.texcoord[1];\n"

	"MOV result.color, vertex.color;\n"

	"DP3 t1.x, fxInfo, iPos;\n"
	"DP3 t1.y, fyInfo, iPos;\n"
	"SUB t1.x, t1.x, fxInfo.w;\n"
	"SUB t1.y, t1.y, fyInfo.w;\n"

	"MOV t2, defTexCoord;\n"
	"MUL t2.x, iPos.z, RNearZ;\n"
	"MOV oTex0, t2;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo1.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo1.zwzw;\n"
	"MOV oTex1, t2;\n"

	"END\n"
;

static const char *g_vpComplexSurfaceDualTextureAndDetailTexture =
	"!!ARBvp1.0\n"
	"OPTION ARB_position_invariant;\n"

	"ATTRIB iPos = vertex.position;\n"
	"ATTRIB fxInfo = vertex.attrib[6];\n"
	"ATTRIB fyInfo = vertex.attrib[7];\n"
	"ATTRIB texInfo0 = vertex.texcoord[0];\n"
	"ATTRIB texInfo1 = vertex.texcoord[1];\n"
	"ATTRIB texInfo2 = vertex.texcoord[2];\n"
	"ATTRIB texInfo3 = vertex.texcoord[3];\n"
	"PARAM defTexCoord = { 0, 0.5, 0, 1 };\n" //Shared with initializing V coord for alpha texture
	"PARAM RNearZ = 0.002631578947;\n"
	"TEMP t1;\n"
	"TEMP t2;\n"
	"OUTPUT oTex0 = result.texcoord[0];\n"
	"OUTPUT oTex1 = result.texcoord[1];\n"
	"OUTPUT oTex2 = result.texcoord[2];\n"
	"OUTPUT oTex3 = result.texcoord[3];\n"

	"MOV result.color, vertex.color;\n"

	"DP3 t1.x, fxInfo, iPos;\n"
	"DP3 t1.y, fyInfo, iPos;\n"
	"SUB t1.x, t1.x, fxInfo.w;\n"
	"SUB t1.y, t1.y, fyInfo.w;\n"

	"MOV t2, defTexCoord;\n"
	"MUL t2.x, iPos.z, RNearZ;\n"
	"MOV oTex0, t2;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo1.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo1.zwzw;\n"
	"MOV oTex1, t2;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo2.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo2.zwzw;\n"
	"MOV oTex2, t2;\n"

	"MOV t2, defTexCoord;\n"
	"SUB t2.xy, t1.xyxy, texInfo3.xyxy;\n"
	"MUL t2.xy, t2.xyxy, texInfo3.zwzw;\n"
	"MOV oTex3, t2;\n"

	"END\n"
;

static const char *g_vpDefaultRenderingState =
	"!!ARBvp1.0\n"
	"OPTION ARB_position_invariant;\n"

	"MOV result.color, vertex.color;\n"
	"MOV result.texcoord[0], vertex.texcoord[0];\n"

	"END\n"
;


/*-----------------------------------------------------------------------------
	OpenGLDrv.
-----------------------------------------------------------------------------*/

struct FTexInfo {
	QWORD CurrentCacheID;
	FLOAT UMult;
	FLOAT VMult;
	FLOAT UPan;
	FLOAT VPan;
};

// Geometry
struct FGLVertex {
	FLOAT x;
	FLOAT y;
	FLOAT z;
};

// gruenschloss: Normals
struct FGLNormal {
	FLOAT x;
	FLOAT y;
	FLOAT z;
};

// Texcoords
struct FGLTexCoord {
	FLOAT u;
	FLOAT v;
};

// Primary and secondary (specular) color
struct FGLColor {
	INT color;
	INT specular;
};

struct FGLMapDot {
	FLOAT u;
	FLOAT v;
};

typedef struct {
	INT NewColorBytes;
	PFNWGLGETEXTENSIONSSTRINGARBPROC p_wglGetExtensionsStringARB;
	PFNWGLCHOOSEPIXELFORMATARBPROC p_wglChoosePixelFormatARB;
} InitARBPixelFormatWndProcParams_t;

//
// An OpenGL rendering device attached to a viewport.
//
static const TCHAR *g_pSection = TEXT("OpenGLDrv.OpenGLRenderDevice");
class UOpenGLRenderDevice : public URenderDevice {
	DECLARE_CLASS(UOpenGLRenderDevice,URenderDevice,CLASS_Config,OpenGLDrv)

	class ods_buf : public std::basic_stringbuf<TCHAR, std::char_traits<TCHAR> > {
	public:
		virtual ~ods_buf() {
			sync();
		}

	protected:
		int sync() {
			//Output the string
			TCHAR_CALL_OS(OutputDebugStringW(str().c_str()), OutputDebugStringA(appToAnsi(str().c_str())));

			//Clear the buffer
			str(std::basic_string<TCHAR>());
			
			return 0;
		}
	};
	class ods_stream : public std::basic_ostream<TCHAR, std::char_traits<TCHAR> > {
	public:
		ods_stream() : std::basic_ostream<TCHAR, std::char_traits<TCHAR> >(new ods_buf()) {
		}
		~ods_stream() {
			delete rdbuf();
		}
	};

	//Debug stream
	ods_stream dout;

	//Debug bits
	DWORD m_debugBits;
	inline bool DebugBit(DWORD debugBit) {
		return ((m_debugBits & debugBit) != 0);
	}
	enum {
		DEBUG_BIT_BASIC	= 0x00000001,
		DEBUG_BIT_ANY	= 0xFFFFFFFF
	};

#ifdef INCLUDE_SSE_CODE
	#define CPUID_SSE_BIT	0x02000000

	bool CPU_DetectCPUID(void) {
		//Check for cpuid instruction support
		__try {
			__asm {
				//CPUID function 0
				xor eax, eax
				cpuid
			}
		}
		__except (EXCEPTION_EXECUTE_HANDLER) {
			return false;
		}

		return true;
	}

	bool CPU_DetectSSE(void) {
		bool bSupportsSSE;

		//Check for cpuid instruction support
		if (CPU_DetectCPUID() != true) {
			return false;
		}
		
		//Check for SSE support
		bSupportsSSE = false;
		__asm {
			//CPUID function 1
			mov eax, 1
			cpuid

			//Check the SSE bit
			test edx, CPUID_SSE_BIT
			jz l_no_sse

			//Set bSupportsSSE to true
			mov bSupportsSSE, 1

l_no_sse:
		}

		//Return if no CPU SSE support
		if (bSupportsSSE == false) {
			return bSupportsSSE;
		}

		//Check for SSE OS support
		__try {
			__asm {
				//Execute SSE instruction
				xorps xmm0, xmm0
			}
		}
		__except (EXCEPTION_EXECUTE_HANDLER) {
			//Clear SSE support flag
			bSupportsSSE = false;
		}

		return bSupportsSSE;
	}
#endif //INCLUDE_SSE_CODE

	// Information about a cached texture.
	enum tex_type_t {
		TEX_TYPE_COMPRESSED_DXT1,
		TEX_TYPE_PALETTED,
		TEX_TYPE_HAS_PALETTE,
		TEX_TYPE_NORMAL
	};
	#define TEX_FLAG_NO_CLAMP	0x00000001
	struct FCachedTexture {
		GLuint Id;
		DWORD LastUsedFrameCount;
		INT BaseMip;
		INT UBits, VBits;
		DWORD UClampVal, VClampVal;
		FLOAT UMult, VMult;
		INT MaxLevel;
		tex_type_t texType;
		GLenum texSourceFormat;
		GLenum texInternalFormat;
		union {
			void (UOpenGLRenderDevice::*pConvertBGRA7777)(const FMipmapBase *, INT);
		};
	};

	struct FTexConvertCtx {
		BYTE *pCompose;
		INT stepBits;
		const FCachedTexture *pBind;
	} m_texConvertCtx;

	FMemMark m_texComposeMemMark;
	enum { LOCAL_TEX_COMPOSE_BUFFER_SIZE = 16384 };
	BYTE m_localTexComposeBuffer[LOCAL_TEX_COMPOSE_BUFFER_SIZE + 16];


	inline void *AlignMemPtr(void *ptr, DWORD align) {
		return (void *)(((DWORD)ptr + (align - 1)) & -align);
	}
	enum { SSE_ARRAY_ALIGN = 16 };

	// Geometry
	FGLVertex VertexArray[VERTEX_ARRAY_SIZE];

	// gruenschloss:
	// Normals
	FGLNormal NormalArray[VERTEX_ARRAY_SIZE];

	// Texcoords
	FGLTexCoord *TexCoordArray[MAX_TMUNITS];
	FGLTexCoord m_TexCoordArrayMem[MAX_TMUNITS][VERTEX_ARRAY_SIZE + SSE_ARRAY_ALIGN + 8];

	// Primary and secondary (specular) color
	FGLColor ColorArray[VERTEX_ARRAY_SIZE];

	FGLMapDot m_MapDotArrayMem[VERTEX_ARRAY_SIZE + SSE_ARRAY_ALIGN + 8];
	FGLMapDot *MapDotArray;

	//IsNear bits for detail texturing
	DWORD DetailTextureIsNearArray[VERTEX_ARRAY_SIZE / 3];
	DWORD DetailTextureAnyIsNearBits;

	//First and count arrays for glMultiDrawArrays
	GLint MultiDrawFirstArray[VERTEX_ARRAY_SIZE / 3];
	GLsizei MultiDrawCountArray[VERTEX_ARRAY_SIZE / 3];

	DWORD m_csPolyCount;
	INT m_csPtCount;


	// MultiPass rendering information
	struct FGLRenderPass {
		struct FGLSinglePass {
			FTextureInfo* Info;
			DWORD PolyFlags;
			FLOAT PanBias;
		} TMU[MAX_TMUNITS];
	} MultiPass;				// vogel: MULTIPASS!!! ;)

	//Texture state cache information
	DWORD m_texEnableBits;
	DWORD m_clientTexEnableBits;

	//Vertex program cache information
	bool m_vpModeEnabled;
	GLuint m_vpCurrent;
	//Vertex program ids
	GLuint m_vpComplexSurface[MAX_TMUNITS];
	GLuint m_vpComplexSurfaceDetailAlpha;
	GLuint m_vpComplexSurfaceDualTextureAndDetailTexture;
	GLuint m_vpDefaultRenderingState;


	struct FGammaRamp {
		_WORD red[256];
		_WORD green[256];
		_WORD blue[256];
	};

#ifdef WIN32
	// Permanent variables.
	HGLRC hRC;
	HWND hWnd;
	HDC hDC;

	// Gamma Ramp to restore at exit	
	FGammaRamp OriginalRamp; 
#endif

	UBOOL WasFullscreen;

	typedef std::map<DWORD, FCachedTexture> DWORD_CTMap_t;
	typedef std::map<QWORD, FCachedTexture> QWORD_CTMap_t;
	typedef std::vector<GLuint> TexIdPoolVector_t;
	typedef DWORD TexPoolMapKey_t;
	typedef std::vector<GLuint> TexPoolVector_t;
	typedef std::map<TexPoolMapKey_t, TexPoolVector_t> TexPoolMap_t;

	enum { NUM_CTMap_TREES = 16 }; //Must be a power of 2
	inline DWORD CTZeroPrefixCacheIDSuffixToTreeIndex(DWORD CacheIDSuffix) {
		return ((CacheIDSuffix >> 12) & (NUM_CTMap_TREES - 1));
	}
	inline DWORD CTNonZeroPrefixCacheIDSuffixToTreeIndex(DWORD CacheIDSuffix) {
		return ((CacheIDSuffix >> 20) & (NUM_CTMap_TREES - 1));
	}
	inline DWORD MakeTexPoolMapKey(DWORD UBits, DWORD VBits) {
		return ((UBits << 16) | VBits);
	}

	DWORD_CTMap_t m_localZeroPrefixBindMaps[NUM_CTMap_TREES], *m_zeroPrefixBindMaps;
	QWORD_CTMap_t m_localNonZeroPrefixBindMaps[NUM_CTMap_TREES], *m_nonZeroPrefixBindMaps;
	TexIdPoolVector_t m_localNonZeroPrefixTexIdPool, *m_nonZeroPrefixTexIdPool;
	TexPoolMap_t m_localRGBA8TexPool, *m_RGBA8TexPool;

	TArray<FPlane> Modes;
	UViewport* Viewport;

	// Timing.
	DWORD BindCycles, ImageCycles, ComplexCycles, GouraudCycles, TileCycles;

	// Hardware constraints.
	FLOAT LODBias;
	FLOAT GammaOffset;
	INT MaxLogUOverV;
	INT MaxLogVOverU;
	INT MinLogTextureSize;
	INT MaxLogTextureSize;
	INT MaxAnisotropy;
	INT TMUnits;
	INT MaxTMUnits;
	INT RefreshRate;
	UBOOL UseZTrick;
	UBOOL UseBGRATextures;
	UBOOL UseMultiTexture;
	UBOOL UsePalette;
	UBOOL UsePrecache;
	UBOOL ShareLists;
	UBOOL AlwaysMipmap;
	UBOOL UseTrilinear;
	UBOOL AutoGenerateMipmaps;
	UBOOL UseVertexSpecular;
	UBOOL UseAlphaPalette;
	UBOOL UseS3TC;
	UBOOL Use4444Textures;
	UBOOL UseTNT; 					// vogel: REMOVE ME - HACK!
	UBOOL UseCVA;
	UBOOL UseFilterSGIS;
	UBOOL NoFiltering;
	UBOOL UseDetailAlpha;
	UBOOL DetailClipping;
	UBOOL ColorizeDetailTextures;
	UBOOL SinglePassDetail;
	INT UseSSE;
	bool m_bEnableSSE;
	UBOOL UseTexIdPool;
	UBOOL UseTexPool;
	INT DynamicTexIdRecycleLevel;
	UBOOL TexDXT1ToDXT3;
	UBOOL UseMultiDrawArrays;
	UBOOL UseVertexProgram;
	INT SwapInterval;
	UBOOL MaskedTextureHack;

	PFNWGLGETEXTENSIONSSTRINGARBPROC m_wglGetExtensionsStringARB;
	PFNWGLCHOOSEPIXELFORMATARBPROC m_wglChoosePixelFormatARB;

	UBOOL UseAA;
	INT NumAASamples;
	INT AAFilterHint;

	UBOOL BufferActorTris;
	UBOOL BufferClippedActorTris;
	INT BufferedVerts;

	// gruenschloss:
	UBOOL UseTruForm;
	INT TruFormTessellation;

#ifdef UTGLR_SKYBOX_HACK
	UBOOL SkyboxHack;
	unsigned int m_csCurSkyboxTexNum;
	bool m_csAllSkyboxTexDetected;
	enum { NUM_SKYBOX_TEX_LIST_ENTRIES = 2 };
	struct FSkyboxTexList {
		QWORD texId;
		bool valid;
	} m_skyboxTexList[NUM_SKYBOX_TEX_LIST_ENTRIES];
	QWORD m_potentialSkyboxTexList[NUM_SKYBOX_TEX_LIST_ENTRIES];
#endif

	UBOOL ColorArrayEnabled;
	UBOOL RenderFog;

	UBOOL GammaFirstTime;

	INT AllocatedTextures;

	INT m_rpPassCount;
	bool m_rpForceSingle;
	bool m_rpMasked;
	bool m_rpSetDepthEqual;
	bool m_rpSetDepthMask;

	void (UOpenGLRenderDevice::*m_pRenderPassesNoCheckProc)(void);
	void (UOpenGLRenderDevice::*m_pRenderPassesNoCheck_DualTextureAndDetailTextureProc)(FTextureInfo &);

	// Hit info.
	BYTE* HitData;
	INT* HitSize;

	DWORD m_currentFrameCount;

	// Lock variables.
	UBOOL ZTrickToggle;
	INT ZTrickFunc;
	FPlane FlashScale, FlashFog;
	FLOAT RProjZ, Aspect;
	DWORD CurrentPolyFlags;
	DWORD m_lastPolyFlags;

	void (UOpenGLRenderDevice::*m_pBuffer3BasicVertsProc)(FTransTexture **);
	void (UOpenGLRenderDevice::*m_pBuffer3ColoredVertsProc)(FTransTexture **);
	void (UOpenGLRenderDevice::*m_pBuffer3FoggedVertsProc)(FTransTexture **);

	void (UOpenGLRenderDevice::*m_pBuffer3VertsProc)(FTransTexture **);

	DWORD CurrentEnvFlags[MAX_TMUNITS];	
	TArray<INT> GLHitData;
	FTexInfo TexInfo[MAX_TMUNITS];
	FLOAT RFX2, RFY2;
#ifdef UTGLR_DX_BUILD
	INT m_FrameX, m_FrameY;
#endif
	GLuint AlphaTextureId;
	GLuint NoTextureId;

	// Static variables.
	static DWORD_CTMap_t m_sharedZeroPrefixBindMaps[NUM_CTMap_TREES];
	static QWORD_CTMap_t m_sharedNonZeroPrefixBindMaps[NUM_CTMap_TREES];
	static TexIdPoolVector_t m_sharedNonZeroPrefixTexIdPool;
	static TexPoolMap_t m_sharedRGBA8TexPool;
	static INT NumDevices;
	static INT LockCount;

#ifdef WIN32
	static TArray<HGLRC> AllContexts;
	static HGLRC   hCurrentRC;
	static HMODULE hModuleGlMain;
	static HMODULE hModuleGlGdi;
#else
	static UBOOL GLLoaded;
#endif

	// GL functions.
	#define GL_EXT(name) static UBOOL SUPPORTS##name;
	#define GL_PROC(ext,ret,func,parms) static ret (STDCALL *func)parms;
	#include "OpenGLFuncs.h"
	#undef GL_EXT
	#undef GL_PROC


#ifdef RGBA_MAKE
#undef RGBA_MAKE
#endif

	inline int RGBA_MAKE(BYTE r, BYTE g, BYTE b, BYTE a)		// vogel: I hate macros...
	{
		return (a << 24) | (b << 16) | (g << 8) | r;
	}
																// vogel: ... and macros hate me

	// UObject interface.
	void StaticConstructor() {
		guard(UOpenGLRenderDevice::StaticConstructor);

#define CPP_PROPERTY_LOCAL(name) name, CPP_PROPERTY(name)

		//Set parameter defaults and add parameters
		SC_AddFloatConfigParam(TEXT("LODBias"), CPP_PROPERTY_LOCAL(LODBias), 0.0f);
		SC_AddFloatConfigParam(TEXT("GammaOffset"), CPP_PROPERTY_LOCAL(GammaOffset), 0.0f);
		SC_AddIntConfigParam(TEXT("MaxLogUOverV"), CPP_PROPERTY_LOCAL(MaxLogUOverV), 8);
		SC_AddIntConfigParam(TEXT("MaxLogVOverU"), CPP_PROPERTY_LOCAL(MaxLogVOverU), 8);
		SC_AddIntConfigParam(TEXT("MinLogTextureSize"), CPP_PROPERTY_LOCAL(MinLogTextureSize), 0);
		SC_AddIntConfigParam(TEXT("MaxLogTextureSize"), CPP_PROPERTY_LOCAL(MaxLogTextureSize), 8);
		SC_AddBoolConfigParam(TEXT("UseZTrick"), CPP_PROPERTY_LOCAL(UseZTrick), 0);
		SC_AddBoolConfigParam(TEXT("UseBGRATextures"), CPP_PROPERTY_LOCAL(UseBGRATextures), 1);
		SC_AddBoolConfigParam(TEXT("UseMultiTexture"), CPP_PROPERTY_LOCAL(UseMultiTexture), 1);
		SC_AddBoolConfigParam(TEXT("UsePalette"), CPP_PROPERTY_LOCAL(UsePalette), 0);
		SC_AddBoolConfigParam(TEXT("ShareLists"), CPP_PROPERTY_LOCAL(ShareLists), 0);
		SC_AddBoolConfigParam(TEXT("AlwaysMipmap"), CPP_PROPERTY_LOCAL(AlwaysMipmap), 0);
		SC_AddBoolConfigParam(TEXT("UsePrecache"), CPP_PROPERTY_LOCAL(UsePrecache), 0);
		SC_AddBoolConfigParam(TEXT("UseTrilinear"), CPP_PROPERTY_LOCAL(UseTrilinear), 0);
		SC_AddBoolConfigParam(TEXT("AutoGenerateMipmaps"), CPP_PROPERTY_LOCAL(AutoGenerateMipmaps), 0);
		SC_AddBoolConfigParam(TEXT("UseAlphaPalette"), CPP_PROPERTY_LOCAL(UseAlphaPalette), 0);
		SC_AddBoolConfigParam(TEXT("UseS3TC"), CPP_PROPERTY_LOCAL(UseS3TC), 0);
		SC_AddBoolConfigParam(TEXT("Use4444Textures"), CPP_PROPERTY_LOCAL(Use4444Textures), 0);
		SC_AddBoolConfigParam(TEXT("UseTNT"), CPP_PROPERTY_LOCAL(UseTNT), 0);
		SC_AddIntConfigParam(TEXT("MaxAnisotropy"), CPP_PROPERTY_LOCAL(MaxAnisotropy), 0);
		SC_AddBoolConfigParam(TEXT("UseFilterSGIS"), CPP_PROPERTY_LOCAL(UseFilterSGIS), 0);
		SC_AddBoolConfigParam(TEXT("NoFiltering"), CPP_PROPERTY_LOCAL(NoFiltering), 0);
		SC_AddIntConfigParam(TEXT("MaxTMUnits"), CPP_PROPERTY_LOCAL(MaxTMUnits), 0);
		SC_AddIntConfigParam(TEXT("RefreshRate"), CPP_PROPERTY_LOCAL(RefreshRate), 0);
		SC_AddBoolConfigParam(TEXT("UseDetailAlpha"), CPP_PROPERTY_LOCAL(UseDetailAlpha), 1);
		SC_AddBoolConfigParam(TEXT("DetailClipping"), CPP_PROPERTY_LOCAL(DetailClipping), 0);
		SC_AddBoolConfigParam(TEXT("ColorizeDetailTextures"), CPP_PROPERTY_LOCAL(ColorizeDetailTextures), 0);
		SC_AddBoolConfigParam(TEXT("SinglePassDetail"), CPP_PROPERTY_LOCAL(SinglePassDetail), 0);
		SC_AddBoolConfigParam(TEXT("BufferClippedActorTris"), CPP_PROPERTY_LOCAL(BufferClippedActorTris), 0);
		SC_AddIntConfigParam(TEXT("UseSSE"), CPP_PROPERTY_LOCAL(UseSSE), 2);
		SC_AddBoolConfigParam(TEXT("UseTexIdPool"), CPP_PROPERTY_LOCAL(UseTexIdPool), 1);
		SC_AddBoolConfigParam(TEXT("UseTexPool"), CPP_PROPERTY_LOCAL(UseTexPool), 1);
		SC_AddIntConfigParam(TEXT("DynamicTexIdRecycleLevel"), CPP_PROPERTY_LOCAL(DynamicTexIdRecycleLevel), 100);
		SC_AddBoolConfigParam(TEXT("TexDXT1ToDXT3"), CPP_PROPERTY_LOCAL(TexDXT1ToDXT3), 0);
		SC_AddIntConfigParam(TEXT("AAFilterHint"), CPP_PROPERTY_LOCAL(AAFilterHint), 0);
		SC_AddBoolConfigParam(TEXT("UseMultiDrawArrays"), CPP_PROPERTY_LOCAL(UseMultiDrawArrays), 0);
		SC_AddBoolConfigParam(TEXT("UseCVA"), CPP_PROPERTY_LOCAL(UseCVA), 0);
		SC_AddBoolConfigParam(TEXT("UseVertexProgram"), CPP_PROPERTY_LOCAL(UseVertexProgram), 0);
		SC_AddIntConfigParam(TEXT("SwapInterval"), CPP_PROPERTY_LOCAL(SwapInterval), -1);
		SC_AddBoolConfigParam(TEXT("MaskedTextureHack"), CPP_PROPERTY_LOCAL(MaskedTextureHack), 0);
		SC_AddBoolConfigParam(TEXT("UseAA"), CPP_PROPERTY_LOCAL(UseAA), 0);
		SC_AddIntConfigParam(TEXT("NumAASamples"), CPP_PROPERTY_LOCAL(NumAASamples), 0);
#ifdef UTGLR_SKYBOX_HACK
		SC_AddBoolConfigParam(TEXT("SkyboxHack"), CPP_PROPERTY_LOCAL(SkyboxHack), 0);
#endif
		// gruenschloss:
		SC_AddBoolConfigParam(TEXT("UseTruForm"), CPP_PROPERTY_LOCAL(UseTruForm), 0);
		SC_AddIntConfigParam(TEXT("TruFormTessellation"), CPP_PROPERTY_LOCAL(TruFormTessellation), 3);

#undef CPP_PROPERTY_LOCAL

		//Driver flags
		SpanBased				= 0;
		SupportsFogMaps			= 1;
		SupportsDistanceFog		= 0;
		FullscreenOnly			= 0;

		SupportsLazyTextures	= 0;
		PrefersDeferredLoad		= 0;

		unguard;
	}

	void SC_AddBoolConfigParam(const TCHAR *pName, UBOOL &param, ECppProperty EC_CppProperty, INT InOffset, UBOOL defaultValue);
	void SC_AddIntConfigParam(const TCHAR *pName, INT &param, ECppProperty EC_CppProperty, INT InOffset, INT defaultValue);
	void SC_AddFloatConfigParam(const TCHAR *pName, FLOAT &param, ECppProperty EC_CppProperty, INT InOffset, FLOAT defaultValue);


	// Implementation.
	void SetGamma(FLOAT GammaCorrection) {
		GammaCorrection += GammaOffset;

		FGammaRamp Ramp;
		for (INT i = 0; i < 256; i++) {
			Ramp.red[i] = Ramp.green[i] = Ramp.blue[i] = (_WORD)appRound((float)appPow(i / 255.0f, 1.0f / (2.5f * GammaCorrection)) * 65535.0f);
		}
#ifdef __LINUX__
		// vogel: FIXME (talk to Sam)
		// SDL_SetGammaRamp( Ramp.red, Ramp.green, Ramp.blue );
		FLOAT gamma = 0.4 + 2 * GammaCorrection; 
		SDL_SetGamma(gamma, gamma, gamma);
#else
		if (GammaFirstTime) {
			GetDeviceGammaRamp(hDC, &OriginalRamp);
			GammaFirstTime = 0;
		}
		SetDeviceGammaRamp(hDC, &Ramp);
#endif
	}

	void ResetGamma(void) {
		// vogel: grrr, UClient::destroy is called before this gets called so hDC is invalid
		SetDeviceGammaRamp(GetDC( GetDesktopWindow()), &OriginalRamp);

		return;
	}

	bool FindExt(const char *pName) {
		guard(UOpenGLRenderDevice::FindExt);

		if (strcmp(pName, "GL") == 0) {
			return true;
		}

		bool bRet = IsGLExtensionSupported((const char *)glGetString(GL_EXTENSIONS), pName);
		if (bRet) {
			debugf( NAME_Init, TEXT("Device supports: %s"), appFromAnsi(pName) );
		}
		if (DebugBit(DEBUG_BIT_BASIC)) {
			dout << L"utglr: GL_EXT: " << appFromAnsi(pName) << L" = " << bRet << std::endl;
		}

		return bRet;
		unguard;
	}
	void FindProc(void*& ProcAddress, char* Name, char* SupportName, UBOOL& Supports, UBOOL AllowExt) {
		guard(UOpenGLRenderDevice::FindProc);
#ifdef __LINUX__
		if( !ProcAddress )
			ProcAddress = (void*) SDL_GL_GetProcAddress( Name );
#else
#if DYNAMIC_BIND
		if( !ProcAddress )
			ProcAddress = GetProcAddress( hModuleGlMain, Name );
		if( !ProcAddress )
			ProcAddress = GetProcAddress( hModuleGlGdi, Name );
#endif
		if( !ProcAddress && Supports && AllowExt )
			ProcAddress = wglGetProcAddress( Name );
#endif
		if( !ProcAddress )
		{
			if( Supports )
				debugf( TEXT("   Missing function '%s' for '%s' support"), appFromAnsi(Name), appFromAnsi(SupportName) );
			Supports = 0;
		}
		unguard;
	}
	void FindProcs(UBOOL AllowExt) {
		guard(UOpenGLDriver::FindProcs);
		#define GL_EXT(name) if( AllowExt ) SUPPORTS##name = FindExt(#name+1);
		#define GL_PROC(ext,ret,func,parms) FindProc( *(void**)&func, #func, #ext, SUPPORTS##ext, AllowExt );
		#include "OpenGLFuncs.h"
		#undef GL_EXT
		#undef GL_PROC
		unguard;
	}
	UBOOL FailedInitf(const TCHAR* Fmt, ...) {
		TCHAR TempStr[4096];
		GET_VARARGS( TempStr, ARRAY_COUNT(TempStr), Fmt );
		debugf( NAME_Init, TempStr );
		Exit();
		return 0;
	}
	void MakeCurrent() {
		guard(UOpenGLRenderDevice::MakeCurrent);
#ifdef WIN32
		check(hRC);
		check(hDC);
		if(hCurrentRC != hRC) {
			verify(wglMakeCurrent(hDC,hRC));
			hCurrentRC = hRC;
		}
#endif
		unguard;
	}

	void Check(const char* Tag) {
		GLenum Error = glGetError();
		if (Error != GL_NO_ERROR) {
			const TCHAR* Msg;
			switch (Error) {
			case GL_INVALID_ENUM:
				Msg = TEXT("GL_INVALID_ENUM");
				break;
			case GL_INVALID_VALUE:
				Msg = TEXT("GL_INVALID_VALUE");
				break;
			case GL_INVALID_OPERATION:
				Msg = TEXT("GL_INVALID_OPERATION");
				break;
			case GL_STACK_OVERFLOW:
				Msg = TEXT("GL_STACK_OVERFLOW");
				break;
			case GL_STACK_UNDERFLOW:
				Msg = TEXT("GL_STACK_UNDERFLOW");
				break;
			case GL_OUT_OF_MEMORY:
				Msg = TEXT("GL_OUT_OF_MEMORY");
				break;
			default :
				Msg = TEXT("UNKNOWN");
			};
			//appErrorf( TEXT("OpenGL Error: %s (%s)"), Msg, Tag );
			debugf( TEXT("OpenGL Error: %s (%s)"), Msg, Tag );
		}
	}

	void SetNoTexture(INT Multi) {
		guard(UOpenGLRenderDevice::SetNoTexture);
		if (TexInfo[Multi].CurrentCacheID != NoTextureId) {
			// Set small white texture.
			clock(BindCycles);
			glBindTexture(GL_TEXTURE_2D, NoTextureId);
			TexInfo[Multi].CurrentCacheID = NoTextureId;
			unclock(BindCycles);
		}
		unguard;
	}

	void SetAlphaTexture(INT Multi) {
		guard(UOpenGLRenderDevice::SetAlphaTexture);
		if (TexInfo[Multi].CurrentCacheID != AlphaTextureId) {
			// Set alpha gradient texture.
			clock(BindCycles);
			glBindTexture(GL_TEXTURE_2D, AlphaTextureId);
			TexInfo[Multi].CurrentCacheID = AlphaTextureId;
			unclock(BindCycles);
		}
		unguard;
	}

	void CacheTextureInfo(FCachedTexture *pBind, const FTextureInfo &Info, DWORD PolyFlags);

	void ConvertDXT1_DXT3(const FMipmapBase *Mip, INT Level);
	void ConvertP8_P8(const FMipmapBase *Mip, INT Level);
	void ConvertP8_P8_NoStep(const FMipmapBase *Mip, INT Level);
	void ConvertP8_RGBA8888(const FMipmapBase *Mip, const FColor *Palette, INT Level);
	void ConvertP8_RGBA8888_NoStep(const FMipmapBase *Mip, const FColor *Palette, INT Level);
	void ConvertBGRA7777_BGRA8888(const FMipmapBase *Mip, INT Level);
	void ConvertBGRA7777_BGRA8888_NoClamp(const FMipmapBase *Mip, INT Level);
	void ConvertBGRA7777_RGBA8888(const FMipmapBase *Mip, INT Level);

	inline void SetTexture(INT Multi, FTextureInfo& Info, DWORD PolyFlags, FLOAT PanBias) {
		FTexInfo& Tex = TexInfo[Multi];

		// Set panning.
		Tex.UPan = Info.Pan.X + PanBias * Info.UScale;
		Tex.VPan = Info.Pan.Y + PanBias * Info.VScale;

		if (MaskedTextureHack && (PolyFlags & PF_Masked)) {
			Info.CacheID |= 1;
		}

		// Find in cache.
		if (Info.CacheID == Tex.CurrentCacheID && !Info.bRealtimeChanged) {
			return;
		}

		SetTextureNoCheck(Tex, Info, PolyFlags);

		if (MaskedTextureHack && (PolyFlags & PF_Masked)) {
			Info.CacheID &= ~1;
		}

		return;
	}

	inline void SetTextureNoPanBias(INT Multi, FTextureInfo& Info, DWORD PolyFlags) {
		FTexInfo& Tex = TexInfo[Multi];

		// Set panning.
		Tex.UPan = Info.Pan.X;
		Tex.VPan = Info.Pan.Y;

		if (MaskedTextureHack && (PolyFlags & PF_Masked)) {
			Info.CacheID |= 1;
		}

		// Find in cache.
		if (Info.CacheID == Tex.CurrentCacheID && !Info.bRealtimeChanged) {
			return;
		}

		SetTextureNoCheck(Tex, Info, PolyFlags);

		if (MaskedTextureHack && (PolyFlags & PF_Masked)) {
			Info.CacheID &= ~1;
		}

		return;
	}

	void SetTextureNoCheck(FTexInfo& Tex, FTextureInfo& Info, DWORD PolyFlags) {
		guard(UOpenGLRenderDevice::SetTexture);

		// Make current.
		clock(BindCycles);

		Tex.CurrentCacheID = Info.CacheID;
		bool isZeroPrefixCacheID = ((Tex.CurrentCacheID & 0xFFFFFFFF00000000) == 0) ? true : false;
		
		FCachedTexture *pBind = NULL;
		bool existingBind = false;
		bool needTexAllocate = true;

		if (isZeroPrefixCacheID) {
			DWORD CacheIDSuffix = (Tex.CurrentCacheID & 0x00000000FFFFFFFF);
			DWORD_CTMap_t *zeroPrefixBindMap = &m_zeroPrefixBindMaps[CTZeroPrefixCacheIDSuffixToTreeIndex(CacheIDSuffix)];
			DWORD_CTMap_t::iterator bindMapIter = zeroPrefixBindMap->find(CacheIDSuffix);
			if (bindMapIter != zeroPrefixBindMap->end()) {
				pBind = &bindMapIter->second;
				existingBind = true;
				needTexAllocate = false;
			}
			else {
				pBind = &zeroPrefixBindMap->insert(std::make_pair(CacheIDSuffix, FCachedTexture())).first->second;

				//Cache texture info for the new texture
				CacheTextureInfo(pBind, Info, PolyFlags);

				//Allocate a new texture Id
				glGenTextures(1, &pBind->Id);
				AllocatedTextures++;
			}
		}
		else {
			DWORD CacheIDSuffix = (Tex.CurrentCacheID & 0x00000000FFFFFFFF);
			QWORD_CTMap_t *nonZeroPrefixBindMap = &m_nonZeroPrefixBindMaps[CTNonZeroPrefixCacheIDSuffixToTreeIndex(CacheIDSuffix)];
			QWORD_CTMap_t::iterator bindMapIter = nonZeroPrefixBindMap->find(Info.CacheID);
			if (bindMapIter != nonZeroPrefixBindMap->end()) {
				pBind = &bindMapIter->second;
				pBind->LastUsedFrameCount = m_currentFrameCount;
				existingBind = true;
				needTexAllocate = false;
			}
			else {
				pBind = &nonZeroPrefixBindMap->insert(std::make_pair(Info.CacheID, FCachedTexture())).first->second;
				pBind->LastUsedFrameCount = m_currentFrameCount;

				//Cache texture info for the new texture
				CacheTextureInfo(pBind, Info, PolyFlags);

				//See if the tex pool is enabled
				bool needTexIdAllocate = true;
				if (UseTexPool) {
					//See if the format will be RGBA8
					if (pBind->texType == TEX_TYPE_NORMAL) {
						TexPoolMap_t::iterator texPoolIter;

						//Create a key from the lg2 width and height of the texture object
						TexPoolMapKey_t texPoolKey = MakeTexPoolMapKey(pBind->UBits, pBind->VBits);

						//Search for the key in the map
						texPoolIter = m_RGBA8TexPool->find(texPoolKey);
						if (texPoolIter != m_RGBA8TexPool->end()) {
							//Get a reference to the list of tex ids
							TexPoolVector_t &texIdList = texPoolIter->second;
							if (texIdList.size() != 0) {
								//Get the tex id from the tex pool
								pBind->Id = texIdList.back();
								texIdList.pop_back();

#if 0
{
	static int si;
	dout << L"utglr: TexPool retrieve = " << si++ << L", Id = " << Bind->Id
		<< L", u = " << texPoolKey.first << L", v = " << texPoolKey.second << std::endl;
}
#endif

								//Clear the need tex id allocate flag
								needTexIdAllocate = false;

								//Clear the need tex allocate flag
								needTexAllocate = false;
							}
						}
					}
				}
				if (needTexIdAllocate) {
					//Allocate a new texture Id
					if (UseTexIdPool && (m_nonZeroPrefixTexIdPool->size() != 0)) {
						pBind->Id = m_nonZeroPrefixTexIdPool->back();
						m_nonZeroPrefixTexIdPool->pop_back();
					}
					else {
						glGenTextures(1, &pBind->Id);
						AllocatedTextures++;
					}
				}
			}
		}

		glBindTexture(GL_TEXTURE_2D, pBind->Id);
		unclock(BindCycles);

		// Account for all the impact on scale normalization.
		Tex.UMult = pBind->UMult;
		Tex.VMult = pBind->VMult;

		// Upload if needed.
		if (!existingBind || Info.bRealtimeChanged) {
			FColor paletteIndex0;

			// Cleanup texture flags.
			if (SupportsLazyTextures) {
				Info.Load();
			}
			Info.bRealtimeChanged = 0;

			//Set palette index 0 to black for masked paletted textures
			if (Info.Palette && (PolyFlags & PF_Masked)) {
				paletteIndex0 = Info.Palette[0];
				Info.Palette[0] = FColor(0,0,0,0);
			}

			// Download the texture.
			clock(ImageCycles);

			if (pBind->texType == TEX_TYPE_PALETTED) {
				glColorTableEXT(GL_TEXTURE_2D, GL_RGBA, 256, GL_RGBA, GL_UNSIGNED_BYTE, Info.Palette);
			}

			// width * height * 4 bytes per pixel
			DWORD memAllocSize = 1 << (pBind->UBits + pBind->VBits + 2);
			if (memAllocSize > LOCAL_TEX_COMPOSE_BUFFER_SIZE) {
				m_texComposeMemMark = FMemMark(GMem);
				m_texConvertCtx.pCompose = New<BYTE>(GMem, memAllocSize);
			}
			else {
				m_texConvertCtx.pCompose = m_localTexComposeBuffer;
			}

			m_texConvertCtx.pBind = pBind;

			UBOOL SkipMipmaps = (Info.NumMips == 1) && !AlwaysMipmap;

			//Only update texture state for new textures
			if (!existingBind) {
				if (AutoGenerateMipmaps && (pBind->texType == TEX_TYPE_NORMAL)) {
					if (SkipMipmaps) {
						glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_FALSE);
					}
					else {
						glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
					}
				}
			}

			INT MaxLevel = pBind->MaxLevel;
			for (INT Level = 0; Level <= MaxLevel; Level++) {
				// Convert the mipmap.
				INT MipIndex = pBind->BaseMip + Level;
				INT stepBits = 0;
				if (MipIndex >= Info.NumMips) {
					stepBits = MipIndex - (Info.NumMips - 1);
					MipIndex = Info.NumMips - 1;
				}
				m_texConvertCtx.stepBits = stepBits;

				FMipmapBase* Mip = Info.Mips[MipIndex];
				if (Mip->DataPtr) {
					switch (pBind->texType) {
					case TEX_TYPE_COMPRESSED_DXT1:
						if (TexDXT1ToDXT3) {
							guard(ConvertDXT1_DXT3);
							ConvertDXT1_DXT3(Mip, Level);
							unguard;
						}
						else {
							//No conversion required for compressed DXT1 textures
						}
						break;

					case TEX_TYPE_PALETTED:
						guard(ConvertP8_P8);
						if (stepBits == 0) {
							ConvertP8_P8_NoStep(Mip, Level);
						}
						else {
							ConvertP8_P8(Mip, Level);
						}
						unguard;
						break;

					case TEX_TYPE_HAS_PALETTE:
						guard(ConvertP8_RGBA8888);
						if (stepBits == 0) {
							ConvertP8_RGBA8888_NoStep(Mip, Info.Palette, Level);
						}
						else {
							ConvertP8_RGBA8888(Mip, Info.Palette, Level);
						}
						unguard;
						break;

					default:
						guard(ConvertBGRA7777);
						(this->*pBind->pConvertBGRA7777)(Mip, Level);
						unguard;
					}

					BYTE *Src = (BYTE*)m_texConvertCtx.pCompose;
					INT texWidth, texHeight;

					texWidth = 1 << Max(0, pBind->UBits - Level);
					texHeight = 1 << Max(0, pBind->VBits - Level);

					if (!needTexAllocate) {
						if (pBind->texType == TEX_TYPE_COMPRESSED_DXT1) {
							guard(glCompressedTexSubImage2D);
#if 0
{
	static int si;
	dout << L"utglr: glCompressedTexSubImage2DARB = " << si++ << std::endl;
}
#endif
							glCompressedTexSubImage2DARB(
								GL_TEXTURE_2D,
								Level,
								0,
								0,
								texWidth,
								texHeight,
								pBind->texInternalFormat,
								TexDXT1ToDXT3 ? (texWidth * texHeight) : (texWidth * texHeight / 2),
								TexDXT1ToDXT3 ? Src : Mip->DataPtr);
							unguard;
						}
						else {
							guard(glTexSubImage2D);
#if 0
{
	static int si;
	dout << L"utglr: glTexSubImage2D = " << si++ << std::endl;
}
#endif
							glTexSubImage2D(
								GL_TEXTURE_2D,
								Level,
								0,
								0,
								texWidth,
								texHeight,
								pBind->texSourceFormat,
								GL_UNSIGNED_BYTE,
								Src);
							unguard;
						}
					}
					else {
						if (pBind->texType == TEX_TYPE_COMPRESSED_DXT1) {
							guard(glCompressedTexImage2D);
#if 0
{
	static int si;
	dout << L"utglr: glCompressedTexImage2DARB = " << si++ << std::endl;
}
#endif
							glCompressedTexImage2DARB(
								GL_TEXTURE_2D,
								Level,
								pBind->texInternalFormat,
								texWidth,
								texHeight,
								0,
								TexDXT1ToDXT3 ? (texWidth * texHeight) : (texWidth * texHeight / 2),
								TexDXT1ToDXT3 ? Src : Mip->DataPtr);
							unguard;
						}
						else {
							guard(glTexImage2D);
#if 0
{
	static int si;
	dout << L"utglr: glTexImage2D = " << si++ << std::endl;
}
#endif
							glTexImage2D(
								GL_TEXTURE_2D,
								Level,
								pBind->texInternalFormat,
								texWidth,
								texHeight,
								0,
								pBind->texSourceFormat,
								GL_UNSIGNED_BYTE,
								Src);
							unguard;
						}
					}
				}

				if (SkipMipmaps) {
					break;
				}
				if (AutoGenerateMipmaps && (pBind->texType == TEX_TYPE_NORMAL)) {
					break;
				}
			}

			if (memAllocSize > LOCAL_TEX_COMPOSE_BUFFER_SIZE) {
				m_texComposeMemMark.Pop();
			}

			unclock(ImageCycles);

			//Restore palette index 0 for masked paletted textures
			if (Info.Palette && (PolyFlags & PF_Masked)) {
				Info.Palette[0] = paletteIndex0;
			}

			//Only update texture state for new textures
			if (!existingBind) {
				// Set texture state.
				if (NoFiltering) {
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
				}
				else if (!(PolyFlags & PF_NoSmooth)) {
					if (UseFilterSGIS && !SkipMipmaps) {
						glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_FILTER4_SGIS);
						glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_FILTER4_SGIS);
					}
					else {
						glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, SkipMipmaps ? GL_LINEAR : 
							(UseTrilinear ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR_MIPMAP_NEAREST));
						glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
					}
					if (MaxAnisotropy) {
						glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, (GLfloat)MaxAnisotropy);
					}
				}
				else {
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, SkipMipmaps ? GL_NEAREST : GL_NEAREST_MIPMAP_NEAREST);
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
				}

				if (!UseTNT) {
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, MaxLevel < 0 ? 0 : MaxLevel);
				}
			}

			// Cleanup.
			if (SupportsLazyTextures) {
				Info.Unload();
			}
		}

		unguard;
	}

	inline void SetBlend(DWORD PolyFlags) {
		if (!(PolyFlags & (PF_Translucent | PF_Modulated | PF_Memorized)))
			PolyFlags |= PF_Occlude;
		else if (PolyFlags & PF_Translucent)
			PolyFlags &= ~PF_Masked;

		if (CurrentPolyFlags != PolyFlags) {
			SetBlendNoCheck(PolyFlags);
		}
	}
	void SetBlendNoCheck(DWORD PolyFlags);

	inline void SetTexEnv(INT Multi, DWORD PolyFlags) {
		if (CurrentEnvFlags[Multi] != PolyFlags) {
			SetTexEnvNoCheck(Multi, PolyFlags);
		}
	}
	void SetTexEnvNoCheck(INT Multi, DWORD PolyFlags);


#ifdef __LINUX__
	// URenderDevice interface.
	UBOOL Init(UViewport* InViewport, INT NewX, INT NewY, INT NewColorBytes, UBOOL Fullscreen) {
		guard(UOpenGLRenderDevice::Init);

		debugf( TEXT("Initializing OpenGLDrv...") );

		GammaFirstTime = 1;

		// Init global GL.
		if (NumDevices == 0) {
			// Bind the library.
			FString OpenGLLibName;
			// Default to libGL.so.1 if not defined
			if (!GConfig->GetString(g_pSection, TEXT("OpenGLLibName"), OpenGLLibName))
				OpenGLLibName = TEXT("libGL.so.1");

			if (!GLLoaded) {
				// Only call it once as succeeding calls will 'fail'.
				debugf( TEXT("binding %s"), *OpenGLLibName );
				if (SDL_GL_LoadLibrary( *OpenGLLibName ) == -1)
					appErrorf( TEXT(SDL_GetError()) );
				GLLoaded = true;
			}

			SUPPORTS_GL = 1;
			FindProcs( 0 );
			if(!SUPPORTS_GL)
				return 0;
		}
		NumDevices++;

		ZeroPrefixBindMap = ShareLists ? &SharedZeroPrefixBindMap : &LocalZeroPrefixBindMap;
		NonZeroPrefixBindMap = ShareLists ? &SharedNonZeroPrefixBindMap : &LocalNonZeroPrefixBindMap;
		NonZeroPrefixTexIdPool = ShareLists ? &SharedNonZeroPrefixTexIdPool : &LocalNonZeroPrefixTexIdPool;
		RGBA8TexPool = ShareLists ? &SharedRGBA8TexPool : &LocalRGBA8TexPool;
		Viewport = InViewport;

		// Try to change resolution to desired.
		if(!SetRes( NewX, NewY, NewColorBytes, Fullscreen ))
			return FailedInitf(LocalizeError("ResFailed"));

		return 1;
		unguard;
	}

	void UnsetRes() {
		guard(UOpenGLRenderDevice::UnsetRes);
		Flush( 1 );
		unguard;
	}
#else
	UBOOL Init(UViewport* InViewport, INT NewX, INT NewY, INT NewColorBytes, UBOOL Fullscreen) {
		guard(UOpenGLRenderDevice::Init);

		debugf( TEXT("Initializing OpenGLDrv...") );

		GammaFirstTime = 1;

		// Get list of device modes.
		for(INT i = 0; ; i++) {
#if defined(NO_UNICODE_OS_SUPPORT) || !defined(UNICODE)
			{
				DEVMODEA Tmp;
				appMemzero(&Tmp, sizeof(Tmp));
				Tmp.dmSize = sizeof(Tmp);
				if (!EnumDisplaySettingsA(NULL, i, &Tmp)) {
					break;
				}
				Modes.AddUniqueItem(FPlane(Tmp.dmPelsWidth, Tmp.dmPelsHeight, Tmp.dmBitsPerPel, Tmp.dmDisplayFrequency));
			}
#else
			{
				DEVMODEW Tmp;
				appMemzero(&Tmp, sizeof(Tmp));
				Tmp.dmSize = sizeof(Tmp);
				if (!EnumDisplaySettingsW(NULL, i, &Tmp)) {
					break;
				}
				Modes.AddUniqueItem(FPlane(Tmp.dmPelsWidth, Tmp.dmPelsHeight, Tmp.dmBitsPerPel, Tmp.dmDisplayFrequency));
			}
#endif
		}

		// Init global GL.
		if (NumDevices == 0) {
#if DYNAMIC_BIND
			// Find DLL's.
			hModuleGlMain = LoadLibraryA(GL_DLL);
			if (!hModuleGlMain) {
				debugf(NAME_Init, LocalizeError("NoFindGL"), appFromAnsi(GL_DLL));
				return 0;
			}
			hModuleGlGdi = LoadLibraryA("GDI32.dll");
			check(hModuleGlGdi);

			// Find functions.
			SUPPORTS_GL = 1;
			FindProcs(0);
			if (!SUPPORTS_GL) {
				return 0;
			}
#endif
		}
		NumDevices++;

		// Init this GL rendering context.
		m_zeroPrefixBindMaps = ShareLists ? m_sharedZeroPrefixBindMaps : m_localZeroPrefixBindMaps;
		m_nonZeroPrefixBindMaps = ShareLists ? m_sharedNonZeroPrefixBindMaps : m_localNonZeroPrefixBindMaps;
		m_nonZeroPrefixTexIdPool = ShareLists ? &m_sharedNonZeroPrefixTexIdPool : &m_localNonZeroPrefixTexIdPool;
		m_RGBA8TexPool = ShareLists ? &m_sharedRGBA8TexPool : &m_localRGBA8TexPool;
		Viewport = InViewport;
		hWnd = (HWND)InViewport->GetWindow();
		check(hWnd);
		hDC = GetDC(hWnd);
		check(hDC);
#if 0 /* Print all PFD's exposed */
		INT Count = DescribePixelFormat(hDC, 0, 0, NULL);
		for (i = 1; i < Count; i++)
			PrintFormat(hDC, i);
#endif
		if (!SetRes(NewX, NewY, NewColorBytes, Fullscreen))
			return FailedInitf(LocalizeError("ResFailed"));

		return 1;
		unguard;
	}

	void UnsetRes() {
		guard(UOpenGLRenderDevice::UnsetRes);
		check(hRC)
		hCurrentRC = NULL;
		verify(wglMakeCurrent(NULL, NULL));
		verify(wglDeleteContext(hRC));
		verify(AllContexts.RemoveItem(hRC) == 1);
		hRC = NULL;
		if (WasFullscreen) {
			TCHAR_CALL_OS(ChangeDisplaySettings(NULL,0), ChangeDisplaySettingsA(NULL,0));
		}
		unguard;
	}

	void PrintFormat(HDC hDC, INT nPixelFormat) {
		guard(UOpenGlRenderDevice::PrintFormat);
		TCHAR Flags[1024] = TEXT("");
		PIXELFORMATDESCRIPTOR pfd;
		DescribePixelFormat(hDC, nPixelFormat, sizeof(pfd), &pfd);
		if (pfd.dwFlags & PFD_DRAW_TO_WINDOW) appStrcat(Flags, TEXT(" PFD_DRAW_TO_WINDOW"));
		if (pfd.dwFlags & PFD_DRAW_TO_BITMAP) appStrcat(Flags, TEXT(" PFD_DRAW_TO_BITMAP"));
		if (pfd.dwFlags & PFD_SUPPORT_GDI) appStrcat(Flags, TEXT(" PFD_SUPPORT_GDI"));
		if (pfd.dwFlags & PFD_SUPPORT_OPENGL) appStrcat(Flags, TEXT(" PFD_SUPPORT_OPENGL"));
		if (pfd.dwFlags & PFD_GENERIC_ACCELERATED) appStrcat(Flags, TEXT(" PFD_GENERIC_ACCELERATED"));
		if (pfd.dwFlags & PFD_GENERIC_FORMAT) appStrcat(Flags, TEXT(" PFD_GENERIC_FORMAT"));
		if (pfd.dwFlags & PFD_NEED_PALETTE) appStrcat(Flags, TEXT(" PFD_NEED_PALETTE"));
		if (pfd.dwFlags & PFD_NEED_SYSTEM_PALETTE) appStrcat(Flags, TEXT(" PFD_NEED_SYSTEM_PALETTE"));
		if (pfd.dwFlags & PFD_DOUBLEBUFFER) appStrcat(Flags, TEXT(" PFD_DOUBLEBUFFER"));
		if (pfd.dwFlags & PFD_STEREO) appStrcat(Flags, TEXT(" PFD_STEREO"));
		if (pfd.dwFlags & PFD_SWAP_LAYER_BUFFERS) appStrcat(Flags, TEXT("PFD_SWAP_LAYER_BUFFERS"));
		debugf(NAME_Init, TEXT("Pixel format %i:"), nPixelFormat);
		debugf(NAME_Init, TEXT("   Flags:%s"), Flags);
		debugf(NAME_Init, TEXT("   Pixel Type: %i"), pfd.iPixelType);
		debugf(NAME_Init, TEXT("   Bits: Color=%i R=%i G=%i B=%i A=%i"), pfd.cColorBits, pfd.cRedBits, pfd.cGreenBits, pfd.cBlueBits, pfd.cAlphaBits);
		debugf(NAME_Init, TEXT("   Bits: Accum=%i Depth=%i Stencil=%i"), pfd.cAccumBits, pfd.cDepthBits, pfd.cStencilBits);
		unguard;
	}
#endif

	UBOOL SetRes(INT NewX, INT NewY, INT NewColorBytes, UBOOL Fullscreen) {
		unsigned int u;

		guard(UOpenGlRenderDevice::SetRes);
		
		//Get debug bits
		{
			INT i = 0;
			if (!GConfig->GetInt(g_pSection, TEXT("DebugBits"), i)) i = 0;
			m_debugBits = i;
		}
		//Display debug bits
		if (DebugBit(DEBUG_BIT_ANY)) dout << L"utglr: DebugBits = " << m_debugBits << std::endl;

#ifdef __LINUX__
		UnsetRes();

		INT MinDepthBits;
		
		// Minimum size of the depth buffer
		if (!GConfig->GetInt(g_pSection, TEXT("MinDepthBits"), MinDepthBits)) MinDepthBits = 16;
		//debugf( TEXT("MinDepthBits = %i"), MinDepthBits );
		// 16 is the bare minimum.
		if (MinDepthBits < 16) MinDepthBits = 16; 

		INT RequestDoubleBuffer;
		if (!GConfig->GetInt(g_pSection, TEXT("RequestDoubleBuffer"), RequestDoubleBuffer)) RequestDoubleBuffer = 1;

		SDL_GL_SetAttribute(SDL_GL_RED_SIZE, (NewColorBytes <= 2) ? 5 : 8);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, (NewColorBytes <= 2) ? 5 : 8);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, (NewColorBytes <= 2) ? 5 : 8);
		SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, MinDepthBits);
		if (RequestDoubleBuffer) {
			SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
		}
		else {
			SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 0);
		}

		// Change window size.
		Viewport->ResizeViewport(Fullscreen ? (BLIT_Fullscreen | BLIT_OpenGL) : (BLIT_HardwarePaint | BLIT_OpenGL), NewX, NewY, NewColorBytes);
#else
		debugf(TEXT("Enter SetRes()"));

		// If not fullscreen, and color bytes hasn't changed, do nothing.
		if (hRC && !Fullscreen && !WasFullscreen && NewColorBytes == Viewport->ColorBytes) {
			if (!Viewport->ResizeViewport(BLIT_HardwarePaint | BLIT_OpenGL, NewX, NewY, NewColorBytes)) {
				return 0;
			}
			glViewport(0, 0, NewX, NewY);

			return 1;
		}

		// Exit res.
		if (hRC) {
			debugf(TEXT("UnSetRes() -> hRc != NULL"));
			UnsetRes();
		}

#if !STDGL
		Fullscreen = 1; /* Minidrivers are fullscreen only!! */
#endif

		// Change display settings.
		if (Fullscreen && STDGL) {
			INT FindX = NewX, FindY = NewY, BestError = MAXINT;
			for (INT i = 0; i < Modes.Num(); i++) {
				if (Modes(i).Z==NewColorBytes*8) {
					INT Error
					=	(Modes(i).X-FindX)*(Modes(i).X-FindX)
					+	(Modes(i).Y-FindY)*(Modes(i).Y-FindY);
					if (Error < BestError) {
						NewX      = Modes(i).X;
						NewY      = Modes(i).Y;
						BestError = Error;
					}
				}
			}
#if defined(NO_UNICODE_OS_SUPPORT) || !defined(UNICODE)
			{
				DEVMODEA dm;
				ZeroMemory(&dm, sizeof(dm));
				dm.dmSize       = sizeof(dm);
				dm.dmPelsWidth  = NewX;
				dm.dmPelsHeight = NewY;
				dm.dmBitsPerPel = NewColorBytes * 8;
				if (RefreshRate) {
					dm.dmDisplayFrequency = RefreshRate;
					dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_DISPLAYFREQUENCY;// | DM_BITSPERPEL;
				}
				else {
					dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT;// | DM_BITSPERPEL;
				}
				if (ChangeDisplaySettingsA(&dm, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL) {
					debugf( TEXT("ChangeDisplaySettingsA failed: %ix%i"), NewX, NewY );

					return 0;
				}
			}
#else
			{
				DEVMODE dm;
				ZeroMemory(&dm, sizeof(dm));
				dm.dmSize       = sizeof(dm);
				dm.dmPelsWidth  = NewX;
				dm.dmPelsHeight = NewY;
				dm.dmBitsPerPel = NewColorBytes * 8;
				if (RefreshRate) {
					dm.dmDisplayFrequency = RefreshRate;
					dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_DISPLAYFREQUENCY;// | DM_BITSPERPEL;
				}
				else {
					dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT;// | DM_BITSPERPEL;
				}
				if (ChangeDisplaySettingsW(&dm, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL) {
					debugf( TEXT("ChangeDisplaySettingsW failed: %ix%i"), NewX, NewY );

					return 0;
				}
			}
#endif
		}

		// Change window size.
		UBOOL Result = Viewport->ResizeViewport(Fullscreen ? (BLIT_Fullscreen | BLIT_OpenGL) : (BLIT_HardwarePaint | BLIT_OpenGL), NewX, NewY, NewColorBytes);
		if (!Result) {
			if (Fullscreen) {
				TCHAR_CALL_OS(ChangeDisplaySettings(NULL,0),ChangeDisplaySettingsA(NULL,0));
			}

			return 0;
		}

		//Restrict NumAASamples range
		if (NumAASamples < 1) NumAASamples = 1;

		m_wglGetExtensionsStringARB = NULL;
		m_wglChoosePixelFormatARB = NULL;
		if (UseAA) {
			InitARBPixelFormat(NewColorBytes);
		}

		const char *pWGLExtensions = NULL;
		if (m_wglGetExtensionsStringARB != NULL) {
			pWGLExtensions = m_wglGetExtensionsStringARB(hDC);
		}

		bool haveWGLMultisampleARB = false;
		if (UseAA && (pWGLExtensions != NULL)) {
			haveWGLMultisampleARB = IsGLExtensionSupported(pWGLExtensions, "WGL_ARB_multisample");
		}

		if (UseAA && (m_wglChoosePixelFormatARB != NULL) && (haveWGLMultisampleARB == true)) {
			BOOL bRet;
			int iFormats[1];
			int iAttributes[30];
			UINT nNumFormats;

			if (DebugBit(DEBUG_BIT_BASIC)) dout << L"utglr: AAInit" << std::endl;

			iAttributes[0] = WGL_SUPPORT_OPENGL_ARB;
			iAttributes[1] = GL_TRUE;
			iAttributes[2] = WGL_DRAW_TO_WINDOW_ARB;
			iAttributes[3] = GL_TRUE;
			iAttributes[4] = WGL_COLOR_BITS_ARB;
			iAttributes[5] = (NewColorBytes <= 2) ? 16 : 24;
			iAttributes[6] = WGL_DEPTH_BITS_ARB;
			iAttributes[7] = (NewColorBytes <= 2) ? 16 : 24;
			iAttributes[8] = WGL_DOUBLE_BUFFER_ARB;
			iAttributes[9] = GL_TRUE;
			iAttributes[10] = WGL_SAMPLE_BUFFERS_ARB;
			iAttributes[11] = GL_TRUE;
			iAttributes[12] = WGL_SAMPLES_ARB;
			iAttributes[13] = NumAASamples;
			iAttributes[14] = 0;
			iAttributes[15] = 0;

			bRet = m_wglChoosePixelFormatARB(hDC, iAttributes, NULL, 1, iFormats, &nNumFormats);
			if ((bRet == FALSE) || (nNumFormats == 0)) {
				if (DebugBit(DEBUG_BIT_BASIC)) dout << L"utglr: AAInit failed" << std::endl;
				goto basic_GL_init;
			}

			PIXELFORMATDESCRIPTOR tempPfd;
			verify(SetPixelFormat(hDC, iFormats[0], &tempPfd));
			hRC = wglCreateContext(hDC);
			check(hRC);
		}
		else {
basic_GL_init:
			if (DebugBit(DEBUG_BIT_BASIC)) dout << L"utglr: BasicInit" << std::endl;

			// Set res.
			INT DesiredColorBits   = (NewColorBytes <= 2) ? 16 : 24; // vogel: changed to saner values 
			INT DesiredStencilBits = 0;                              // NewColorBytes<=2 ? 0  : 8;
			INT DesiredDepthBits   = (NewColorBytes <= 2) ? 16 : 24; // NewColorBytes<=2 ? 16 : 32;
			PIXELFORMATDESCRIPTOR pfd =
			{
				sizeof(PIXELFORMATDESCRIPTOR),
				1,
				PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
				PFD_TYPE_RGBA,
				DesiredColorBits,
				0,0,0,0,0,0,
				0,0,
				0,0,0,0,0,
				DesiredDepthBits,
				DesiredStencilBits,
				0,
				PFD_MAIN_PLANE,
				0,
				0,0,0
			};
			INT nPixelFormat = ChoosePixelFormat(hDC, &pfd);
			Parse(appCmdLine(), TEXT("PIXELFORMAT="), nPixelFormat);
			debugf(NAME_Init, TEXT("Using pixel format %i"), nPixelFormat);
			check(nPixelFormat);
			verify(SetPixelFormat(hDC, nPixelFormat, &pfd));
			hRC = wglCreateContext(hDC);
			check(hRC);
		}

		MakeCurrent();
		if(ShareLists && AllContexts.Num())
			verify(wglShareLists(AllContexts(0),hRC) == 1);
		AllContexts.AddItem(hRC);
#endif

		// Get info and extensions.
	
		//PrintFormat( hDC, nPixelFormat );
		debugf(NAME_Init, TEXT("GL_VENDOR     : %s"), appFromAnsi((ANSICHAR*)glGetString(GL_VENDOR)));
		debugf(NAME_Init, TEXT("GL_RENDERER   : %s"), appFromAnsi((ANSICHAR*)glGetString(GL_RENDERER)));
		debugf(NAME_Init, TEXT("GL_VERSION    : %s"), appFromAnsi((ANSICHAR*)glGetString(GL_VERSION)));
		// vogel: logging of more than 1024 characters is dangerous at the moment.
		if (strlen((char*) glGetString(GL_EXTENSIONS)) < 1024)
			debugf( NAME_Init, TEXT("GL_EXTENSIONS : %s"), appFromAnsi((ANSICHAR*)glGetString(GL_EXTENSIONS)));
#ifdef __LINUX__
		else
			printf("GL_EXTENSIONS : %s", glGetString(GL_EXTENSIONS));
#endif		

		FindProcs(1);

		// Set modelview.
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		FLOAT Matrix[16] =
		{
			+1, +0, +0, +0,
			+0, -1, +0, +0,
			+0, +0, -1, +0,
			+0, +0, +0, +1,
		};
		glMultMatrixf(Matrix);

		//Debug parameter listing
		if (DebugBit(DEBUG_BIT_BASIC)) {
			#define UTGLR_DEBUG_SHOW_PARAM(name) dout << L"utglr: " L#name L" = " << name << std::endl;

			UTGLR_DEBUG_SHOW_PARAM(LODBias);
			UTGLR_DEBUG_SHOW_PARAM(GammaOffset);
			UTGLR_DEBUG_SHOW_PARAM(MaxLogUOverV);
			UTGLR_DEBUG_SHOW_PARAM(MaxLogVOverU);
			UTGLR_DEBUG_SHOW_PARAM(MinLogTextureSize);
			UTGLR_DEBUG_SHOW_PARAM(MaxLogTextureSize);
			UTGLR_DEBUG_SHOW_PARAM(MaxAnisotropy);
			UTGLR_DEBUG_SHOW_PARAM(TMUnits);
			UTGLR_DEBUG_SHOW_PARAM(MaxTMUnits);
			UTGLR_DEBUG_SHOW_PARAM(RefreshRate);
			UTGLR_DEBUG_SHOW_PARAM(UseZTrick);
			UTGLR_DEBUG_SHOW_PARAM(UseBGRATextures);
			UTGLR_DEBUG_SHOW_PARAM(UseMultiTexture);
			UTGLR_DEBUG_SHOW_PARAM(UsePalette);
			UTGLR_DEBUG_SHOW_PARAM(UsePrecache);
			UTGLR_DEBUG_SHOW_PARAM(ShareLists);
			UTGLR_DEBUG_SHOW_PARAM(AlwaysMipmap);
			UTGLR_DEBUG_SHOW_PARAM(UseTrilinear);
			UTGLR_DEBUG_SHOW_PARAM(AutoGenerateMipmaps);
//			UTGLR_DEBUG_SHOW_PARAM(UseVertexSpecular);
			UTGLR_DEBUG_SHOW_PARAM(UseAlphaPalette);
			UTGLR_DEBUG_SHOW_PARAM(UseS3TC);
			UTGLR_DEBUG_SHOW_PARAM(Use4444Textures);
			UTGLR_DEBUG_SHOW_PARAM(UseTNT);
			UTGLR_DEBUG_SHOW_PARAM(UseCVA);
			UTGLR_DEBUG_SHOW_PARAM(UseFilterSGIS);
			UTGLR_DEBUG_SHOW_PARAM(NoFiltering);
			UTGLR_DEBUG_SHOW_PARAM(UseDetailAlpha);
			UTGLR_DEBUG_SHOW_PARAM(DetailClipping);
			UTGLR_DEBUG_SHOW_PARAM(ColorizeDetailTextures);
			UTGLR_DEBUG_SHOW_PARAM(SinglePassDetail);
			UTGLR_DEBUG_SHOW_PARAM(UseSSE);
			UTGLR_DEBUG_SHOW_PARAM(UseTexIdPool);
			UTGLR_DEBUG_SHOW_PARAM(UseTexPool);
			UTGLR_DEBUG_SHOW_PARAM(DynamicTexIdRecycleLevel);
			UTGLR_DEBUG_SHOW_PARAM(TexDXT1ToDXT3);
			UTGLR_DEBUG_SHOW_PARAM(UseMultiDrawArrays);
			UTGLR_DEBUG_SHOW_PARAM(UseVertexProgram);
			UTGLR_DEBUG_SHOW_PARAM(SwapInterval);
			UTGLR_DEBUG_SHOW_PARAM(MaskedTextureHack);
//			UTGLR_DEBUG_SHOW_PARAM(BufferActorTris);
			UTGLR_DEBUG_SHOW_PARAM(BufferClippedActorTris);
			UTGLR_DEBUG_SHOW_PARAM(UseAA);
			UTGLR_DEBUG_SHOW_PARAM(NumAASamples);
			UTGLR_DEBUG_SHOW_PARAM(AAFilterHint);
#ifdef UTGLR_SKYBOX_HACK
			UTGLR_DEBUG_SHOW_PARAM(SkyboxHack);
#endif
			// gruenschloss:
			UTGLR_DEBUG_SHOW_PARAM(UseTruForm);
			UTGLR_DEBUG_SHOW_PARAM(TruFormTessellation);

			#undef UTGLR_DEBUG_SHOW_PARAM
		}

		//Restrict dynamic tex id recycle level range
		if (DynamicTexIdRecycleLevel < 10) DynamicTexIdRecycleLevel = 10;

		//Special handling for WGL_EXT_swap_control
		//Restricted to a maximum of 10 for now
		if ((SwapInterval >= 0) && (SwapInterval <= 10)) {
			//Setup pWGLExtensions (may or may not have been setup before for AAInit)
			pWGLExtensions = NULL;
			PFNWGLGETEXTENSIONSSTRINGARBPROC p_wglGetExtensionsStringARB = reinterpret_cast<PFNWGLGETEXTENSIONSSTRINGARBPROC>(wglGetProcAddress("wglGetExtensionsStringARB"));
			if (p_wglGetExtensionsStringARB != NULL) {
				pWGLExtensions = p_wglGetExtensionsStringARB(hDC);
			}

			//See if WGL extension string was retrieved
			if (pWGLExtensions != NULL) {
				//See if WGL_EXT_swap_control is supported
				if (IsGLExtensionSupported(pWGLExtensions, "WGL_EXT_swap_control")) {
					//Get the extension function pointer and use it if successful
					PFNWGLSWAPINTERVALEXTPROC p_wglSwapIntervalEXT = reinterpret_cast<PFNWGLSWAPINTERVALEXTPROC>(wglGetProcAddress("wglSwapIntervalEXT"));
					if (p_wglSwapIntervalEXT != NULL) {
						//Set the swap interval
						p_wglSwapIntervalEXT(SwapInterval);
					}
				}
			}
		}

		if (SUPPORTS_GL_NV_multisample_filter_hint) {
			if (AAFilterHint != 0) {
				switch (AAFilterHint) {
				case 1:
					glHint(GL_MULTISAMPLE_FILTER_HINT_NV, GL_FASTEST);
					break;
				case 2:
					glHint(GL_MULTISAMPLE_FILTER_HINT_NV, GL_NICEST);
					break;
				default:
					;
				}
			}
		}


#ifdef INCLUDE_SSE_CODE
		if (UseSSE == 1) {
			m_bEnableSSE = true;
		}
		else if (UseSSE == 2) {
			if (CPU_DetectSSE() == true) {
				m_bEnableSSE = true;
			}
			else {
				m_bEnableSSE = false;
			}
		}
		else {
			m_bEnableSSE = false;
		}
#else
		m_bEnableSSE = false;
#endif
		if (DebugBit(DEBUG_BIT_BASIC)) dout << L"utglr: m_bEnableSSE = " << m_bEnableSSE << std::endl;

		SetGamma(Viewport->GetOuterUClient()->Brightness);

		UseVertexSpecular	= 1;
		AlwaysMipmap		= 0;

		SupportsTC			= UseS3TC;

		//ARB extension uses similar tokens
		SUPPORTS_GL_EXT_texture_env_combine |= SUPPORTS_GL_ARB_texture_env_combine;

		// Validate flags.
		if (!SUPPORTS_GL_ARB_vertex_program) UseVertexProgram = 0;
		m_vpModeEnabled = false;
		m_vpCurrent = 0;

		if (!SUPPORTS_GL_EXT_multi_draw_arrays) UseMultiDrawArrays = 0;
		if (!SUPPORTS_GL_EXT_bgra) UseBGRATextures = 0;
		if (!SUPPORTS_GL_ARB_multitexture) UseMultiTexture = 0;
		if (!SUPPORTS_GL_SGIS_generate_mipmap) AutoGenerateMipmaps = 0;
		if (!SUPPORTS_GL_EXT_paletted_texture) UsePalette = 0;
		if (!SUPPORTS_GL_EXT_texture_compression_s3tc || !SUPPORTS_GL_ARB_texture_compression) SupportsTC = 0;
		if (!SUPPORTS_GL_EXT_texture_env_combine) DetailTextures = 0;
		if (!SUPPORTS_GL_EXT_compiled_vertex_array) UseCVA = 0;
		if (!SUPPORTS_GL_EXT_secondary_color) UseVertexSpecular = 0;
		if (!SUPPORTS_GL_EXT_texture_filter_anisotropic) MaxAnisotropy  = 0;
		if (!SUPPORTS_GL_SGIS_texture_lod) UseFilterSGIS = 0;
		if (!SUPPORTS_GL_EXT_texture_env_combine) UseDetailAlpha = 0;
		if (!SUPPORTS_GL_EXT_texture_lod_bias) LODBias = 0;

		// gruenschloss: check if TruForm is supported by hardware
		if (!SUPPORTS_GL_ATI_pn_triangles && UseTruForm) {
			UseTruForm = false;
			debugf(TEXT("OpenGL Error: No support for TruForm! (GL_ATI_pn_triangles)"));
		}

#ifdef UTGLR_SKYBOX_HACK
		//Reset skybox tex list
		for (u = 0; u < NUM_SKYBOX_TEX_LIST_ENTRIES; u++) {
			m_skyboxTexList[u].valid = false;
		}
#endif

		if (!MaxTMUnits || (MaxTMUnits > MAX_TMUNITS)) {
			MaxTMUnits = MAX_TMUNITS;
		}

		if (UseMultiTexture) {
			glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &TMUnits);
			debugf(TEXT("%i Texture Mapping Units found"), TMUnits);
			if (TMUnits > MaxTMUnits) {
				TMUnits = MaxTMUnits;
			}
		}
		else {
			TMUnits = 1;
			UseDetailAlpha = 0;
		}

		//Must use detail alpha for vertex program detail textures
		if (!UseDetailAlpha) {
			if (UseVertexProgram) {
				DetailTextures = 0;
			}
		}

		//Single pass detail texturing requires at least 4 texture units
		if (TMUnits < 4) {
			SinglePassDetail = 0;
		}
		//Single pass detail texturing requires detail alpha
		if (!UseDetailAlpha) {
			SinglePassDetail = 0;
		}
		
		if (MaxAnisotropy) {
			INT tmp;
			glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &tmp);
			debugf(TEXT("MaxAnisotropy = %i"), tmp); 
			if (MaxAnisotropy > tmp) {
				MaxAnisotropy = tmp;
			}

			UseTrilinear = true; // Anisotropic filtering doesn't make much sense without trilinear filtering
		}

		BufferActorTris = UseVertexSpecular; // Only buffer when we can use 1 pass fogging

		if (SupportsTC) {
			debugf(TEXT("Trying to use S3TC extension."));
			UseTNT = 0;
		}

		//Special case MaxLogTextureSize == 0
		if (MaxLogTextureSize == 0) MaxLogTextureSize = 8;

		INT MaxTextureSize;
		glGetIntegerv(GL_MAX_TEXTURE_SIZE, &MaxTextureSize);
		INT Dummy = -1;
		while (MaxTextureSize > 0) {
			MaxTextureSize >>= 1;
			Dummy++;
		}
		
		if ((MaxLogTextureSize > Dummy) || (SupportsTC)) MaxLogTextureSize = Dummy;
		if ((MinLogTextureSize < 2) || (SupportsTC)) MinLogTextureSize = 2;

		if (SupportsTC) {
			MaxLogUOverV = MaxLogTextureSize;
			MaxLogVOverU = MaxLogTextureSize;
		}
		else {
			MaxLogUOverV = 8;
			MaxLogVOverU = 8;
		}

		debugf(TEXT("MinLogTextureSize = %i"), MinLogTextureSize);
		debugf(TEXT("MaxLogTextureSize = %i"), MaxLogTextureSize);


		//Set pointers to aligned memory
		for (u = 0; u < MAX_TMUNITS; u++) {
			TexCoordArray[u] = (FGLTexCoord *)AlignMemPtr(m_TexCoordArrayMem[u], SSE_ARRAY_ALIGN);
		}
		MapDotArray = (FGLMapDot *)AlignMemPtr(m_MapDotArrayMem, SSE_ARRAY_ALIGN);


		// Verify hardware defaults.
		check(MinLogTextureSize >= 0);
		check(MaxLogTextureSize >= 0);
		check(MinLogTextureSize < MaxLogTextureSize);
		check(MinLogTextureSize <= MaxLogTextureSize);

		// Flush textures.
		Flush(1);

		// Bind little white RGBA texture to ID 0.
		FColor Data[8*8];
		for (INT i = 0; i < ARRAY_COUNT(Data); i++) {
			Data[i] = FColor(255,255,255,255);
		}
		glGenTextures(1, &NoTextureId);
		SetNoTexture(0);
		for (INT Level = 0; 8 >> Level; Level++) {
			glTexImage2D(GL_TEXTURE_2D, Level, 4, 8 >> Level, 8 >> Level, 0, GL_RGBA, GL_UNSIGNED_BYTE, Data);
		}

		// Set permanent state.
		glEnable(GL_DEPTH_TEST);
		glShadeModel(GL_SMOOTH);
		glEnable(GL_TEXTURE_2D);
		glAlphaFunc(GL_GREATER, 0.5);
		glDepthMask(GL_TRUE);
		glDepthFunc(GL_LEQUAL);
		glPolygonOffset(-1.0f, -1.0f);
		glBlendFunc(GL_ONE, GL_ZERO);
		glDisable(GL_BLEND);
		glDisable(GL_DITHER);
		if (LODBias) {
			glTexEnvf(GL_TEXTURE_FILTER_CONTROL_EXT, GL_TEXTURE_LOD_BIAS_EXT, LODBias);
			if (SUPPORTS_GL_ARB_multitexture) {
				for (INT i = 1; i < TMUnits; i++) {
					glActiveTextureARB(GL_TEXTURE0_ARB + i);
					glTexEnvf(GL_TEXTURE_FILTER_CONTROL_EXT, GL_TEXTURE_LOD_BIAS_EXT, LODBias);
				}
				glActiveTextureARB(GL_TEXTURE0_ARB);
			}
		}
		if (UseDetailAlpha) {			// vogel: alpha texture for better detail textures (no vertex alpha)
			BYTE AlphaData[256];
			for (INT ac = 0; ac < 256; ac++) {
				AlphaData[ac] = 255 - ac;
			}
			glGenTextures(1, &AlphaTextureId);
			// vogel: could use 1D texture but opted against (for minor reasons)
			glBindTexture(GL_TEXTURE_2D, AlphaTextureId);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, 256, 1, 0, GL_ALPHA, GL_UNSIGNED_BYTE, AlphaData);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		}
		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, sizeof(FGLVertex), &VertexArray[0].x);

		// gruenschloss:
		if (UseTruForm)
			glNormalPointer(GL_FLOAT, sizeof(FGLNormal), &NormalArray[0].x);

		// gruenschloss:
		if (UseTruForm) {
#ifdef UTGLR_GRUENSCHLOSS_DEBUG
			debugf(TEXT("OpenGL: setting TruForm modes"));
#endif
			glPNTrianglesiATI(GL_PN_TRIANGLES_NORMAL_MODE_ATI, GL_PN_TRIANGLES_NORMAL_MODE_LINEAR_ATI);
			glPNTrianglesiATI(GL_PN_TRIANGLES_POINT_MODE_ATI, GL_PN_TRIANGLES_POINT_MODE_CUBIC_ATI);
			glPNTrianglesiATI(GL_PN_TRIANGLES_TESSELATION_LEVEL_ATI, TruFormTessellation);
		}

		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, sizeof(FGLTexCoord), &TexCoordArray[0][0].u);
		if (UseMultiTexture) {
			glClientActiveTextureARB(GL_TEXTURE1_ARB);
			glTexCoordPointer(2, GL_FLOAT, sizeof(FGLTexCoord), &TexCoordArray[1][0].u);
			if (TMUnits > 2) {
				glClientActiveTextureARB(GL_TEXTURE2_ARB);
				glTexCoordPointer(2, GL_FLOAT, sizeof(FGLTexCoord), &TexCoordArray[2][0].u);
			}
			if (TMUnits > 3) {
				glClientActiveTextureARB(GL_TEXTURE3_ARB);
				glTexCoordPointer(2, GL_FLOAT, sizeof(FGLTexCoord), &TexCoordArray[3][0].u);
			}
			glClientActiveTextureARB(GL_TEXTURE0_ARB);
		}
		glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(FGLColor), &ColorArray[0].color);
		if (UseVertexSpecular) {
			glSecondaryColorPointerEXT(3, GL_UNSIGNED_BYTE, sizeof(FGLColor), &ColorArray[0].specular);
		}

		//Initialize texture state cache information
		m_texEnableBits = 0x1;
		m_clientTexEnableBits = 0x1;

		// Init variables.
		BufferedVerts = 0;
		ColorArrayEnabled = 0;
		RenderFog = 0;

		CurrentPolyFlags = PF_Occlude;
		m_lastPolyFlags = 0;


		//Initialize default proc pointers
		// gruenschloss:
		if (UseTruForm) {
			m_pBuffer3BasicVertsProc = &UOpenGLRenderDevice::Buffer3Verts;
			m_pBuffer3ColoredVertsProc = &UOpenGLRenderDevice::Buffer3Verts;
			m_pBuffer3FoggedVertsProc = &UOpenGLRenderDevice::Buffer3Verts;
		}
		else {
			m_pBuffer3BasicVertsProc = &UOpenGLRenderDevice::Buffer3BasicVerts;
			m_pBuffer3ColoredVertsProc = &UOpenGLRenderDevice::Buffer3ColoredVerts;
			m_pBuffer3FoggedVertsProc = &UOpenGLRenderDevice::Buffer3FoggedVerts;
		}

#ifdef INCLUDE_SSE_CODE
		//Initialize SSE proc pointers
		if (m_bEnableSSE == true && !UseTruForm) {
			m_pBuffer3ColoredVertsProc = &UOpenGLRenderDevice::Buffer3ColoredVerts_SSE;
		}
#endif //INCLUDE_SSE_CODE

		m_pBuffer3VertsProc = NULL;

		//Initialize render passes no check proc pointer
		if (UseVertexProgram) {
			m_pRenderPassesNoCheckProc = &UOpenGLRenderDevice::RenderPassesNoCheck_VP;
			m_pRenderPassesNoCheck_DualTextureAndDetailTextureProc = &UOpenGLRenderDevice::RenderPassesNoCheck_DualTextureAndDetailTexture_VP;
		}
		else {
			m_pRenderPassesNoCheckProc = &UOpenGLRenderDevice::RenderPassesNoCheck;
			m_pRenderPassesNoCheck_DualTextureAndDetailTextureProc = &UOpenGLRenderDevice::RenderPassesNoCheck_DualTextureAndDetailTexture;
#ifdef INCLUDE_SSE_CODE
			if (m_bEnableSSE == true) {
				m_pRenderPassesNoCheckProc = &UOpenGLRenderDevice::RenderPassesNoCheck_SSE;
				m_pRenderPassesNoCheck_DualTextureAndDetailTextureProc = &UOpenGLRenderDevice::RenderPassesNoCheck_DualTextureAndDetailTexture_SSE;
			}
#endif //INCLUDE_SSE_CODE
		}


		for (INT TMU = 0; TMU < MAX_TMUNITS; TMU++) {
			CurrentEnvFlags[TMU] = 0;
		}

		if (UseVertexProgram) {
			//Initialize vertex programs
			AllocateVertexProgramNames();
			InitializeVertexPrograms();
		}

		//Reset current frame count
		m_currentFrameCount = 0;

		// Remember fullscreenness.
		WasFullscreen = Fullscreen;
		return 1;

		unguard;
	}

	void InitARBPixelFormat(INT NewColorBytes) {
		guard(UOpenGLRenderDevice::InitARBPixelFormat);

		HINSTANCE hInstance = TCHAR_CALL_OS(GetModuleHandleW(NULL), GetModuleHandleA(NULL));
		struct tagWNDCLASSW wcW;
		struct tagWNDCLASSA wcA;
		static const WCHAR *pClassNameW = L"UOpenGLRenderDevice::InitARBPixelFormat";
		static const CHAR *pClassNameA = "UOpenGLRenderDevice::InitARBPixelFormat";
		InitARBPixelFormatWndProcParams_t initParams;
		HWND hWnd;

#if defined(NO_UNICODE_OS_SUPPORT) || !defined(UNICODE)
		wcA.style = CS_HREDRAW | CS_VREDRAW;
		wcA.lpfnWndProc = UOpenGLRenderDevice::InitARBPixelFormatWndProc;
		wcA.cbClsExtra = 0;
		wcA.cbWndExtra = 0;
		wcA.hInstance = hInstance;
		wcA.hIcon = NULL;
		wcA.hCursor = LoadCursorA(NULL, MAKEINTRESOURCEA(32512)/*IDC_ARROW*/);
		wcA.hbrBackground = NULL;
		wcA.lpszMenuName = NULL;
		wcA.lpszClassName = pClassNameA;
#else
		wcW.style = CS_HREDRAW | CS_VREDRAW;
		wcW.lpfnWndProc = UOpenGLRenderDevice::InitARBPixelFormatWndProc;
		wcW.cbClsExtra = 0;
		wcW.cbWndExtra = 0;
		wcW.hInstance = hInstance;
		wcW.hIcon = NULL;
		wcW.hCursor = LoadCursorW(NULL, MAKEINTRESOURCEW(32512)/*IDC_ARROW*/);
		wcW.hbrBackground = NULL;
		wcW.lpszMenuName = NULL;
		wcW.lpszClassName = pClassNameW;
#endif

		if (TCHAR_CALL_OS(RegisterClassW(&wcW), RegisterClassA(&wcA)) == 0) {
			return;
		}

		initParams.NewColorBytes = NewColorBytes;
		initParams.p_wglGetExtensionsStringARB = NULL;
		initParams.p_wglChoosePixelFormatARB = NULL;

		hWnd = TCHAR_CALL_OS(
			CreateWindowW(
			pClassNameW,
			pClassNameW,
			WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
			0, 0,
			20, 20,
			NULL,
			NULL,
			hInstance,
			&initParams),
			CreateWindowA(
			pClassNameA,
			pClassNameA,
			WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
			0, 0,
			20, 20,
			NULL,
			NULL,
			hInstance,
			&initParams)
			);
		if (hWnd == NULL) {
			TCHAR_CALL_OS(UnregisterClassW(pClassNameW, hInstance), UnregisterClassA(pClassNameA, hInstance));
			return;
		}

		DestroyWindow(hWnd);

		TCHAR_CALL_OS(UnregisterClassW(pClassNameW, hInstance), UnregisterClassA(pClassNameA, hInstance));

		m_wglGetExtensionsStringARB = initParams.p_wglGetExtensionsStringARB;
		m_wglChoosePixelFormatARB = initParams.p_wglChoosePixelFormatARB;

		return;

		unguard;
	}

	static LRESULT CALLBACK InitARBPixelFormatWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
		switch(uMsg) {
		case WM_CREATE:
			{
				InitARBPixelFormatWndProcParams_t *pInitParams = (InitARBPixelFormatWndProcParams_t *)((LPCREATESTRUCT)lParam)->lpCreateParams;

				HDC hDC = GetDC(hWnd);

				INT NewColorBytes = pInitParams->NewColorBytes;
				INT DesiredColorBits   = (NewColorBytes <= 2) ? 16 : 24;
				INT DesiredStencilBits = 0;
				INT DesiredDepthBits   = (NewColorBytes <= 2) ? 16 : 24;
				PIXELFORMATDESCRIPTOR pfd =
				{
					sizeof(PIXELFORMATDESCRIPTOR),
					1,
					PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
					PFD_TYPE_RGBA,
					DesiredColorBits,
					0,0,0,0,0,0,
					0,0,
					0,0,0,0,0,
					DesiredDepthBits,
					DesiredStencilBits,
					0,
					PFD_MAIN_PLANE,
					0,
					0,0,0
				};
				INT nPixelFormat = ChoosePixelFormat(hDC, &pfd);
				if (nPixelFormat == 0) {
					ReleaseDC(hWnd, hDC);
					break;
				}
				if (SetPixelFormat(hDC, nPixelFormat, &pfd) == FALSE) {
					ReleaseDC(hWnd, hDC);
					break;
				}
				HGLRC hGLRC = wglCreateContext(hDC);
				if (hGLRC == NULL) {
					ReleaseDC(hWnd, hDC);
					break;
				}
				if (wglMakeCurrent(hDC, hGLRC) == FALSE) {
					wglDeleteContext(hGLRC);
					ReleaseDC(hWnd, hDC);
					break;
				}

				pInitParams->p_wglGetExtensionsStringARB = reinterpret_cast<PFNWGLGETEXTENSIONSSTRINGARBPROC>(wglGetProcAddress("wglGetExtensionsStringARB"));
				pInitParams->p_wglChoosePixelFormatARB = reinterpret_cast<PFNWGLCHOOSEPIXELFORMATARBPROC>(wglGetProcAddress("wglChoosePixelFormatARB"));

				wglMakeCurrent(NULL, NULL);
				wglDeleteContext(hGLRC);

				ReleaseDC(hWnd, hDC);
			}
			break;

		default:
			return TCHAR_CALL_OS(DefWindowProcW(hWnd, uMsg, wParam, lParam), DefWindowProcA(hWnd, uMsg, wParam, lParam));
		}

		return 0;
	}

	static bool IsGLExtensionSupported(const char *pExtensionsString, const char *pExtensionName) {
		const char *pStart;
		const char *pWhere, *pTerminator;

		pStart = pExtensionsString;
		while (1) {
			pWhere = strstr(pStart, pExtensionName);
			if (pWhere == NULL) {
				break;
			}
			pTerminator = pWhere + strlen(pExtensionName);
			if ((pWhere == pStart) || (*(pWhere - 1) == ' ')) {
				if ((*pTerminator == ' ') || (*pTerminator == '\0')) {
					return true;
				}
			}
			pStart = pTerminator;
		}

		return false;
	}

	void Exit() {
		guard(UOpenGLRenderDevice::Exit);
		check(NumDevices > 0);

		// Shut down RC.
		Flush(0);

#ifdef __LINUX__
		UnsetRes();

		// Shut down global GL.
		if (--NumDevices == 0) {
			{
				unsigned int u;

				for (u = 0; u < NUM_CTMap_TREES; u++) {
					m_sharedZeroPrefixBindMaps[u].clear();
				}
				for (u = 0; u < NUM_CTMap_TREES; u++) {
					m_sharedNonZeroPrefixBindMaps[u].clear();
				}
			}
			m_sharedNonZeroPrefixTexIdPool.clear();
			m_sharedRGBA8TexPool.clear();
		}
#else
		if (hRC) {
			UnsetRes();
		}

		//Reset gamma ramp
		ResetGamma();

		// Shut down this GL context. May fail if window was already destroyed.
		if (hDC) {
			ReleaseDC(hWnd, hDC);
		}

		// Shut down global GL.
		if (--NumDevices == 0) {
#if DYNAMIC_BIND && 0 /* Broken on some drivers */
			// Free modules.
			if (hModuleGlMain)
				verify(FreeLibrary(hModuleGlMain));
			if (hModuleGlGdi)
				verify(FreeLibrary(hModuleGlGdi));
#endif
			{
				unsigned int u;

				for (u = 0; u < NUM_CTMap_TREES; u++) {
					m_sharedZeroPrefixBindMaps[u].clear();
				}
				for (u = 0; u < NUM_CTMap_TREES; u++) {
					m_sharedNonZeroPrefixBindMaps[u].clear();
				}
			}
			m_sharedNonZeroPrefixTexIdPool.clear();
			m_sharedRGBA8TexPool.clear();
			AllContexts.~TArray<HGLRC>();
		}
#endif
		unguard;
	}

	void ShutdownAfterError() {
		guard(UOpenGLRenderDevice::ShutdownAfterError);

		debugf(NAME_Exit, TEXT("UOpenGLRenderDevice::ShutdownAfterError"));

		if (DebugBit(DEBUG_BIT_BASIC)) {
			dout << L"ShutdownAfterError" << std::endl;
		}

		//ChangeDisplaySettings( NULL, 0 );

		//Reset gamma ramp
		ResetGamma();

		unguard;
	}

	void Flush(UBOOL AllowPrecache) {
		guard(UOpenGLRenderDevice::Flush);
		unsigned int u;
		TArray<GLuint> Binds;

		for (u = 0; u < NUM_CTMap_TREES; u++) {
			DWORD_CTMap_t *zeroPrefixBindMap = &m_zeroPrefixBindMaps[u];
			for (DWORD_CTMap_t::const_iterator zpbmIt = zeroPrefixBindMap->begin(); zpbmIt != zeroPrefixBindMap->end(); ++zpbmIt) {
				Binds.AddItem(zpbmIt->second.Id);
			}
			zeroPrefixBindMap->clear();
		}

		for (u = 0; u < NUM_CTMap_TREES; u++) {
			QWORD_CTMap_t *nonZeroPrefixBindMap = &m_nonZeroPrefixBindMaps[u];
			for (QWORD_CTMap_t::const_iterator nzpbmIt = nonZeroPrefixBindMap->begin(); nzpbmIt != nonZeroPrefixBindMap->end(); ++nzpbmIt) {
				Binds.AddItem(nzpbmIt->second.Id);
			}
			nonZeroPrefixBindMap->clear();
		}

		for (TexIdPoolVector_t::const_iterator nzptipIt = m_nonZeroPrefixTexIdPool->begin(); nzptipIt != m_nonZeroPrefixTexIdPool->end(); ++nzptipIt) {
			Binds.AddItem(*nzptipIt);
		}
		m_nonZeroPrefixTexIdPool->clear();

		for (TexPoolMap_t::const_iterator RGBA8TpIt = m_RGBA8TexPool->begin(); RGBA8TpIt != m_RGBA8TexPool->end(); ++RGBA8TpIt) {
			for (TexPoolVector_t::const_iterator texIdListIter = RGBA8TpIt->second.begin(); texIdListIter != RGBA8TpIt->second.end(); ++texIdListIter) {
				Binds.AddItem(*texIdListIter);
			}
		}
		m_RGBA8TexPool->clear();

		// vogel: FIXME: add 0 and AlphaTextureId

		if (Binds.Num()) {
			glDeleteTextures(Binds.Num(), (GLuint*)&Binds(0));
		}
		AllocatedTextures = 0;

		//Reset current texture ids to hopefully unused values
		for (u = 0; u < MAX_TMUNITS; u++) {
			TexInfo[u].CurrentCacheID = 0xFFFFFFFFFFFFFFFF;
		}

#ifdef UTGLR_SKYBOX_HACK
		if (SkyboxHack) {
			unsigned int u;

			//Mark skybox textures as invalid
			for (u = 0; u < NUM_SKYBOX_TEX_LIST_ENTRIES; u++) {
				m_skyboxTexList[u].valid = false;
			}
		}
#endif

		if (AllowPrecache && UsePrecache && !GIsEditor) {
			PrecacheOnFlip = 1;
		}

		SetGamma(Viewport->GetOuterUClient()->Brightness);

		unguard;
	}

	static QSORT_RETURN CDECL CompareRes(const FPlane* A, const FPlane* B) {
		return (QSORT_RETURN) ( (A->X-B->X)!=0.0 ? (A->X-B->X) : (A->Y-B->Y) );
	}

	UBOOL Exec(const TCHAR* Cmd, FOutputDevice& Ar) {
		guard(UOpenGLRenderDevice::Exec);
		if (URenderDevice::Exec(Cmd, Ar)) {
			return 1;
		}
		if (ParseCommand(&Cmd,TEXT("DGL"))) {
			if( ParseCommand(&Cmd,TEXT("BUFFERTRIS")) ) {
				BufferActorTris = !BufferActorTris;
				if ( !UseVertexSpecular )
					BufferActorTris = 0;
				debugf( TEXT("BUFFERTRIS [%i]"), BufferActorTris );
			}
			if( ParseCommand(&Cmd,TEXT("CVA")) ) {
				UseCVA = !UseCVA;
				if ( !SUPPORTS_GL_EXT_compiled_vertex_array )
					UseCVA = 0;
				debugf( TEXT("CVA [%i]"), UseCVA );
			}
			if( ParseCommand(&Cmd,TEXT("TRILINEAR")) ) {
				UseTrilinear = !UseTrilinear;
				debugf( TEXT("TRILINEAR [%i]"), UseTrilinear );
				Flush(1);
			}
			if( ParseCommand(&Cmd,TEXT("RGBA4444")) ) {
				Use4444Textures = !Use4444Textures;
				debugf( TEXT("RGBA4444 [%i]"), Use4444Textures );
				Flush(1);
			}
			if( ParseCommand(&Cmd,TEXT("TNTFIX")) ) {
				UseTNT = !UseTNT;
				debugf( TEXT("TNTFIX [%i]"), UseTNT );
				Flush(1);
			}
			if( ParseCommand(&Cmd,TEXT("MULTITEX")) ) {
				UseMultiTexture = !UseMultiTexture;
				if( !SUPPORTS_GL_ARB_multitexture )
					UseMultiTexture = 0;
				if ( UseMultiTexture )
					glGetIntegerv( GL_MAX_TEXTURE_UNITS_ARB, &TMUnits );
				else
					TMUnits = 1;
				debugf( TEXT("MULTITEX [%i]"), UseMultiTexture );
			}
			if( ParseCommand(&Cmd,TEXT("DETAILTEX")) ) {
				DetailTextures = !DetailTextures;
				if( !SUPPORTS_GL_EXT_texture_env_combine )
					DetailTextures = 0;
				debugf( TEXT("DETAILTEX [%i]"), DetailTextures );
			}
			if( ParseCommand(&Cmd,TEXT("DETAILALPHA")) ) {
				UseDetailAlpha = !UseDetailAlpha;
				if( !SUPPORTS_GL_EXT_texture_env_combine )
					UseDetailAlpha = 0;
				debugf( TEXT("DETAILALPHA [%i]"), UseDetailAlpha );
			}
			if( ParseCommand(&Cmd,TEXT("BUILD")) ) {
#ifdef __LINUX__
				debugf( TEXT("OpenGL renderer built: %s %s"), __DATE__, __TIME__ );
#else
				debugf( TEXT("OpenGL renderer built: VOGEL FIXME") );
#endif
			}
			return 1;	// vogel: FIXME
		}
		else if( ParseCommand(&Cmd,TEXT("GetRes")) ) {
#ifdef __LINUX__
			// Changing Resolutions:
			// Entries in the resolution box in the console is
			// apparently controled building a string of relevant 
			// resolutions and sending it to the engine via Ar.Log()

			// Here I am querying SDL_ListModes for available resolutions,
			// and populating the dropbox with its output.
			FString Str = "";
			SDL_Rect **modes;
                        INT i,j;

                	// Available fullscreen video modes
               		modes = SDL_ListModes(NULL, SDL_FULLSCREEN);
                        
			if ( modes == (SDL_Rect **)0 ) {
				debugf( NAME_Init, TEXT("No available fullscreen video modes") );
			}
			else if ( modes == (SDL_Rect **)-1 ) {
				debugf( NAME_Init, TEXT("No special fullscreen video modes") );
			}
			else {
				// count the number of available modes
				for ( i=0,j=0; modes[i]; ++i )
					++j;

				// Load the string with resolutions from smallest to 
				// largest. SDL_ListModes() provides them from lg
				// to sm...
				for ( i=(j-1); i >= 0; --i )
					Str += FString::Printf( TEXT("%ix%i "), modes[i]->w, modes[i]->h);
			}

			// Send the resolution string to the engine.
			Ar.Log( *Str.LeftChop(1) );
			return 1;
#else
			TArray<FPlane> Relevant;
			INT i;
			for( i=0; i<Modes.Num(); i++ )
				if( Modes(i).Z==Viewport->ColorBytes*8 )
					if
					(	(Modes(i).X!=320 || Modes(i).Y!=200)
					&&	(Modes(i).X!=640 || Modes(i).Y!=400) )
					Relevant.AddUniqueItem(FPlane(Modes(i).X,Modes(i).Y,0,0));
			appQsort( &Relevant(0), Relevant.Num(), sizeof(FPlane), (QSORT_COMPARE)CompareRes );
			FString Str;
			for( i=0; i<Relevant.Num(); i++ )
				Str += FString::Printf( TEXT("%ix%i "), (INT)Relevant(i).X, (INT)Relevant(i).Y );
			Ar.Log( *Str.LeftChop(1) );
			return 1;
#endif
		}
		return 0;
		unguard;
	}


	void Lock(FPlane InFlashScale, FPlane InFlashFog, FPlane ScreenClear, DWORD RenderLockFlags, BYTE* InHitData, INT* InHitSize) {
#ifdef UTGLR_DEBUG_SHOW_CALL_COUNTS
{
	static int si;
	dout << L"utglr: Lock = " << si++ << std::endl;
}
#endif
		guard(UOpenGLRenderDevice::Lock);
		check(LockCount == 0);
		BindCycles = ImageCycles = ComplexCycles = GouraudCycles = TileCycles = 0;
		++LockCount;

		// Make this context current.
		MakeCurrent();

		// Clear the Z buffer if needed.
		if(!UseZTrick || GIsEditor || (RenderLockFlags & LOCKR_ClearScreen)) {
			glClearColor(ScreenClear.X, ScreenClear.Y, ScreenClear.Z, ScreenClear.W);
			glClearDepth(1.0);
			glDepthRange(0.0, 1.0);
			ZTrickFunc = GL_LEQUAL;
			glPolygonOffset(-1.0f, -1.0f);
			SetBlend(PF_Occlude);
			glClear(GL_DEPTH_BUFFER_BIT | ((RenderLockFlags & LOCKR_ClearScreen) ? GL_COLOR_BUFFER_BIT : 0));
		}
		else if (ZTrickToggle) {
			ZTrickToggle = 0;
			glClearDepth(0.5);
			glDepthRange(0.0, 0.5);
			ZTrickFunc = GL_LEQUAL;
			glPolygonOffset(-1.0f, -1.0f);
		}
		else {
			ZTrickToggle = 1;
			glClearDepth(0.5);
			glDepthRange(1.0, 0.5);
			ZTrickFunc = GL_GEQUAL;
			glPolygonOffset(1.0f, 1.0f);
		}
		glDepthFunc((GLenum)ZTrickFunc);

		//Increment current frame count
		m_currentFrameCount++;

		//Scan for old textures
		if (UseTexIdPool) {
			unsigned int u;

			for (u = 0; u < NUM_CTMap_TREES; u++) {
				QWORD_CTMap_t *nonZeroPrefixBindMap = &m_nonZeroPrefixBindMaps[u];

				for (QWORD_CTMap_t::iterator nzpbmIt = nonZeroPrefixBindMap->begin(); nzpbmIt != nonZeroPrefixBindMap->end(); ++nzpbmIt) {
					//Get a reference to the cached texture object
					const FCachedTexture &cachedTexture = nzpbmIt->second;

					DWORD numFramesSinceUsed = m_currentFrameCount - cachedTexture.LastUsedFrameCount;
					if (numFramesSinceUsed > DynamicTexIdRecycleLevel) {
						//See if the tex pool is not enabled or the internal format is not RGBA8
						if (!UseTexPool || (cachedTexture.texInternalFormat != GL_RGBA8)) {
							//Add the texture id to the texture id pool
							m_nonZeroPrefixTexIdPool->push_back(cachedTexture.Id);
						}
						else {
							TexPoolMap_t::iterator texPoolIter;

							//Create a key from the lg2 width and height of the texture object
							TexPoolMapKey_t texPoolKey = MakeTexPoolMapKey(cachedTexture.UBits, cachedTexture.VBits);

							//See if the key does not yet exist
							texPoolIter = m_RGBA8TexPool->find(texPoolKey);
							//If the key does not yet exist, add an empty vector in its place
							if (texPoolIter == m_RGBA8TexPool->end()) {
								texPoolIter = m_RGBA8TexPool->insert(m_RGBA8TexPool->begin(), std::make_pair(texPoolKey, TexPoolVector_t()));
							}

							//Add the texture id to a list in the tex pool based on its dimensions
							texPoolIter->second.push_back(cachedTexture.Id);
#if 0
{
	static int si;
	dout << L"utglr: TexPool free = " << si++ << L", Id = " << cachedTexture.Id
		<< L", u = " << texPoolKey.first << L", v = " << texPoolKey.second << std::endl;
}
#endif
						}
	
						//Remove the cached texture object from the bind map
						nzpbmIt = nonZeroPrefixBindMap->erase(nzpbmIt);
						if (nzpbmIt == nonZeroPrefixBindMap->end()) break;
					}
				}
			}
		}

#ifdef UTGLR_SKYBOX_HACK
		if (SkyboxHack) {
			//Reset skybox flags
			m_csCurSkyboxTexNum = 0;
			m_csAllSkyboxTexDetected = true;
		}
#endif

		// Remember stuff.
		FlashScale = InFlashScale;
		FlashFog   = InFlashFog;
		HitData    = InHitData;
		HitSize    = InHitSize;
		if (HitData) {
			*HitSize = 0;
			if (!GLHitData.Num())
				GLHitData.Add(16384);
			glSelectBuffer(GLHitData.Num(), (GLuint*)&GLHitData(0));
			glRenderMode(GL_SELECT);
			glInitNames();
		}
		unguard;
	}

	void SetSceneNode(FSceneNode* Frame) {
#if 0
{
	static int si;
	dout << L"utglr: SetSceneNode = " << si++ << std::endl;
}
#endif
		guard(UOpenGLRenderDevice::SetSceneNode);

		EndBuffering();		// Flush vertex array before changing the projection matrix!

		SetDefaultVertexProgramState();
		SetDefaultTextureState();

		// gruenschloss:
/*
#ifdef UTGLR_GRUENSCHLOSS_DEBUG
		dout << L"[Frame]: " <<
			L"(X: " << Frame->X << L"), " <<
			L"(Y: " << Frame->Y << L"), " <<
			L"(FX: " << Frame->FX << L"), " <<
			L"(FY: " << Frame->FY << L"), " <<
			L"(FX2: " << Frame->FX2 << L"), " <<
			L"(FY2: " << Frame->FY2 << L"), " <<
			L"(XB: " << Frame->XB << L"), " <<
			L"(YB: " << Frame->YB << L")" << std::endl;
#endif
*/

		// Precompute stuff.
		FLOAT One_Over_FX = 1.0f / Frame->FX;
		Aspect = Frame->FY * One_Over_FX;

		// gruenschloss:
#ifdef UTGLR_DX_BUILD
		// standard or cinemativ view
		INT offsetY = Frame->YB;
		FLOAT magicFOV = 97.7419; // => 75.0
		// computer console mode
		if (Frame->XB != 0) {
			offsetY = m_FrameY - Frame->YB - Frame->Y;
			magicFOV = 90.0;
		}

		// gruenschloss:
		RProjZ = appTan(Frame->PrjXM * magicFOV * PI / 360.0);
#else
		RProjZ = appTan(Viewport->Actor->FovAngle * PI / 360.0);
#endif
		RFX2 = 2.0f * RProjZ * One_Over_FX;
		RFY2 = 2.0f * RProjZ * One_Over_FX;

		// Set viewport.
#ifdef UTGLR_DX_BUILD
		// gruenschloss:
		glViewport(Frame->XB, offsetY, Frame->X, Frame->Y);
#else
		glViewport(Frame->XB, Viewport->SizeY - Frame->Y - Frame->YB, Frame->X, Frame->Y);
#endif

		// gruenschloss (changed position):
#ifdef UTGLR_DX_BUILD
		m_FrameX = Frame->X;
		m_FrameY = Frame->Y;
#endif

		// Set projection.
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		if (Frame->Viewport->IsOrtho()) {
			glOrtho(-RProjZ / 2.0, +RProjZ / 2.0, -Aspect*RProjZ / 2.0, +Aspect*RProjZ / 2.0, 1.0 / 2.0, 32768.0);
		}
		else {
			glFrustum(-RProjZ / 2.0, +RProjZ / 2.0, -Aspect*RProjZ / 2.0, +Aspect*RProjZ / 2.0, 1.0 / 2.0, 32768.0);
		}

		// Set clip planes.
		if (HitData) {
			FVector N[4];
			N[0] = (FVector((Viewport->HitX-Frame->FX2)*Frame->RProj.Z,0,1)^FVector(0,-1,0)).SafeNormal();
			N[1] = (FVector((Viewport->HitX+Viewport->HitXL-Frame->FX2)*Frame->RProj.Z,0,1)^FVector(0,+1,0)).SafeNormal();
			N[2] = (FVector(0,(Viewport->HitY-Frame->FY2)*Frame->RProj.Z,1)^FVector(+1,0,0)).SafeNormal();
			N[3] = (FVector(0,(Viewport->HitY+Viewport->HitYL-Frame->FY2)*Frame->RProj.Z,1)^FVector(-1,0,0)).SafeNormal();
			for (INT i = 0; i < 4; i++) {
				double D0[4] = {N[i].X, N[i].Y, N[i].Z, 0};
				glClipPlane((GLenum)(GL_CLIP_PLANE0 + i), D0);
				glEnable((GLenum)(GL_CLIP_PLANE0 + i));
			}
		}
		unguard;
	}

	void Unlock(UBOOL Blit) {
#ifdef UTGLR_DEBUG_SHOW_CALL_COUNTS
{
	static int si;
	dout << L"utglr: Unlock = " << si++ << std::endl;
}
#endif
		guard(UOpenGLRenderDevice::Unlock);

		EndBuffering();

		SetDefaultVertexProgramState();
		SetDefaultTextureState();

		// Unlock and render.
		check(LockCount == 1);
		//glFlush();
		if (Blit) {
			Check("please report this bug");
#ifdef __LINUX__
			SDL_GL_SwapBuffers();
#else
			verify(SwapBuffers( hDC ));
#endif
		}
		--LockCount;

		// Hits.
		if (HitData) {
			INT Records = glRenderMode(GL_RENDER);
			INT* Ptr = &GLHitData(0);
			DWORD BestDepth = MAXDWORD;
			for (INT i = 0; i < Records; i++) {
				INT   NameCount = *Ptr++;
				DWORD MinDepth  = *Ptr++;
				DWORD MaxDepth  = *Ptr++;
				if (MinDepth <= BestDepth) {
					BestDepth = MinDepth;
					*HitSize = 0;
					for (INT i = 0; i < NameCount; ) {
						INT Count = Ptr[i++];
						for (INT j = 0; j < Count; j += 4)
							*(INT*)(HitData + *HitSize + j) = Ptr[i++];
						*HitSize += Count;
					}
					check(i == NameCount);
				}
				Ptr += NameCount;
				(void)MaxDepth;
			}
			for (i = 0; i < 4; i++)
				glDisable(GL_CLIP_PLANE0 + i);
		}

		unguard;
	}

	inline void SetDefaultTextureState(void) {
		//Check if only texture unit zero is enabled
		if (m_texEnableBits != 0x1) {
			DWORD texUnit;
			DWORD texBit;

			//Disable all texture units except texture unit zero
			for (texUnit = 1, texBit = 0x2; m_texEnableBits != 0x1; texUnit++, texBit <<= 1) {
				//See if the texture unit is enabled
				if (texBit & m_texEnableBits) {
					//Disable the texture unit
					glActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
					glDisable(GL_TEXTURE_2D);

					m_texEnableBits &= ~texBit;
				}
			}

			//Leave texture unit zero active
			glActiveTextureARB(GL_TEXTURE0_ARB);
		}

		//Check if only client texture zero is enabled
		if (m_clientTexEnableBits != 0x1) {
			DWORD texUnit;
			DWORD texBit;

			//Enable client texture zero if it is disabled
			if ((m_clientTexEnableBits & 0x1) == 0) {
				//Enable client texture zero
				glClientActiveTextureARB(GL_TEXTURE0_ARB);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits |= 0x1;
			}

			//Disable all client textures except client texture zero
			for (texUnit = 1, texBit = 0x2; m_clientTexEnableBits != 0x1; texUnit++, texBit <<= 1) {
				//Disable the client texture
				glClientActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits &= ~texBit;
			}
		}

		return;
	}

	inline void SetDefaultVertexProgramState(void) {
		//See if vertex program mode is enabled
		if (m_vpModeEnabled == true) {
			//Mark vertex program mode as disabled
			m_vpModeEnabled = false;

			//Disable vertex program mode
			glDisable(GL_VERTEX_PROGRAM_ARB);
		}

		return;
	}

	void SetVertexProgram(GLuint program) {
		//See if vertex program mode is disabled
		if (m_vpModeEnabled == false) {
			//Mark vertex program mode as enabled
			m_vpModeEnabled = true;

			//Enable vertex program mode
			glEnable(GL_VERTEX_PROGRAM_ARB);
		}

		//See if the requested program is not already bound
		if (program != m_vpCurrent) {
			//Save the new current vertex program
			m_vpCurrent = program;

			//Bind the vertex program
			glBindProgramARB(GL_VERTEX_PROGRAM_ARB, program);
		}

		return;
	}

	void AllocateVertexProgramNames(void) {
		//Allocate complex surface names
		glGenProgramsARB(MAX_TMUNITS, m_vpComplexSurface);
		glGenProgramsARB(1, &m_vpComplexSurfaceDetailAlpha);
		glGenProgramsARB(1, &m_vpComplexSurfaceDualTextureAndDetailTexture);
		glGenProgramsARB(1, &m_vpDefaultRenderingState);
	}

	void FreeVertexProgramNames(void) {
		//Free complex surface names
		glDeleteProgramsARB(MAX_TMUNITS, m_vpComplexSurface);
		glDeleteProgramsARB(1, &m_vpComplexSurfaceDetailAlpha);
		glDeleteProgramsARB(1, &m_vpComplexSurfaceDualTextureAndDetailTexture);
		glDeleteProgramsARB(1, &m_vpDefaultRenderingState);
	}

	void InitializeVertexPrograms(void) {
		//Complex surface single texture
		LoadVertexProgram(m_vpComplexSurface[0], g_vpComplexSurfaceSingleTexture,
			L"Complex surface single texture");

		if (TMUnits >= 2) {
			//Complex surface dual texture
			LoadVertexProgram(m_vpComplexSurface[1], g_vpComplexSurfaceDoubleTexture,
				L"Complex surface dual texture");
		}

		if (TMUnits >= 3) {
			//Complex surface triple texture
			LoadVertexProgram(m_vpComplexSurface[2], g_vpComplexSurfaceTripleTexture,
				L"Complex surface triple texture");
		}

		if (TMUnits >= 4) {
			//Complex surface quad texture
			LoadVertexProgram(m_vpComplexSurface[3], g_vpComplexSurfaceQuadTexture,
				L"Complex surface quad texture");
		}


		if (UseDetailAlpha) {
			//Complex surface detail alpha
			LoadVertexProgram(m_vpComplexSurfaceDetailAlpha, g_vpComplexSurfaceDetailAlpha,
				L"Complex surface detail alpha");
		}

		if (SinglePassDetail) {
			//Complex surface dual texture and detail texture
			LoadVertexProgram(m_vpComplexSurfaceDualTextureAndDetailTexture, g_vpComplexSurfaceDualTextureAndDetailTexture,
				L"Complex surface dual texture and detail texture");
		}


		//Default rendering state
		LoadVertexProgram(m_vpDefaultRenderingState, g_vpDefaultRenderingState,
			L"Default rendering state");


		//Reset to default vertex program and udpate current vertex program variable
		glBindProgramARB(GL_VERTEX_PROGRAM_ARB, 0);
		m_vpCurrent = 0;
	}

	void LoadVertexProgram(GLuint vpId, const char *pProgram, const TCHAR *pName) {
		if (DebugBit(DEBUG_BIT_BASIC)) {
			dout << L"utglr: Loading vertex program \"" << pName << L"\"" << std::endl;
		}

		glBindProgramARB(GL_VERTEX_PROGRAM_ARB, vpId);
		glProgramStringARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB,
			strlen(pProgram), pProgram);

		if (DebugBit(DEBUG_BIT_BASIC)) {
			GLint i;

			glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &i);
			if (i != -1) {
				dout << L"utglr: Vertex program error at offset " << i << std::endl;
				dout << L"utglr: Vertex program text from error offset:\n" << appFromAnsi(pProgram + i) << std::endl;
			}
		}
	}

	inline void RenderPasses(void) {
		if (m_rpPassCount != 0) {
			if (m_rpMasked && m_rpForceSingle && !m_rpSetDepthEqual) {
				glDepthFunc(GL_EQUAL);
				m_rpSetDepthEqual = true;
			}

			//Call the render passes no check proc
			(this->*m_pRenderPassesNoCheckProc)();

			m_rpForceSingle = true;
		}
	}
	void RenderPassesNoCheck(void);
#ifdef INCLUDE_SSE_CODE
	void RenderPassesNoCheck_SSE(void);
#endif //INCLUDE_SSE_CODE
	void RenderPassesNoCheck_VP(void);
	void RenderPassesNoCheck_DualTextureAndDetailTexture(FTextureInfo &);
#ifdef INCLUDE_SSE_CODE
	void RenderPassesNoCheck_DualTextureAndDetailTexture_SSE(FTextureInfo &);
#endif //INCLUDE_SSE_CODE
	void RenderPassesNoCheck_DualTextureAndDetailTexture_VP(FTextureInfo &);

	inline void AddRenderPass(FTextureInfo* Info, DWORD PolyFlags, FLOAT PanBias) {
		MultiPass.TMU[m_rpPassCount].Info      = Info;
		MultiPass.TMU[m_rpPassCount].PolyFlags = PolyFlags;
		MultiPass.TMU[m_rpPassCount].PanBias   = PanBias;

		m_rpPassCount++;
		if ((m_rpPassCount >= TMUnits) || m_rpForceSingle) {
			RenderPasses();
		}
	}

	INT BufferStaticComplexSurfaceGeometry(const FSurfaceFacet&);
	INT BufferStaticComplexSurfaceGeometry_VP(const FSurfaceFacet&);
#ifdef INCLUDE_SSE_CODE
	INT BufferStaticComplexSurfaceGeometry_SSE(const FSurfaceFacet&);
#endif
	INT BufferStaticComplexSurfaceGeometryAndDetailTextureData(const FSurfaceFacet&);
	INT BufferStaticComplexSurfaceGeometryAndDetailTextureData_VP(const FSurfaceFacet&);

	void DrawComplexSurface(FSceneNode* Frame, FSurfaceInfo& Surface, FSurfaceFacet& Facet) {
#ifdef UTGLR_DEBUG_SHOW_CALL_COUNTS
{
	static int si;
	dout << L"utglr: DrawComplexSurface = " << si++ << std::endl;
}
#endif
		guard(UOpenGLRenderDevice::DrawComplexSurface);
		check(Surface.Texture);

		EndBuffering();		// vogel: might have still been locked (can happen!)

		//This function uses cached vertex program state information
		//This function uses cached texture state information

		clock(ComplexCycles);

		DWORD PolyFlags = Surface.PolyFlags;

#ifdef UTGLR_SKYBOX_HACK
		if (SkyboxHack) {
			if (m_csCurSkyboxTexNum < NUM_SKYBOX_TEX_LIST_ENTRIES) {
				//Save texture id
				m_potentialSkyboxTexList[m_csCurSkyboxTexNum] = Surface.Texture->CacheID;

				//Check for skybox texture match
				if ((m_skyboxTexList[m_csCurSkyboxTexNum].valid == true) &&
					(m_potentialSkyboxTexList[m_csCurSkyboxTexNum] == m_skyboxTexList[m_csCurSkyboxTexNum].texId)) {
					
					//Adjust flags so that Z values will not be written
					PolyFlags |= PF_Memorized;
				}
				else {
					m_csAllSkyboxTexDetected = false;
				}

				m_csCurSkyboxTexNum++;
			}
			else {
				m_csAllSkyboxTexDetected = false;
			}
		}
#endif

		//Set initial draw detail texture flag base on config setting and detail texture presence
		bool drawDetailTexture = (DetailTextures != 0) && Surface.DetailTexture;

		// Mutually exclusive effects.
		if (drawDetailTexture && Surface.FogMap) {
			drawDetailTexture = false;
		}

		// Buffer "static" geometry.
		INT Index;
		if (drawDetailTexture == true) {
			if (UseVertexProgram) {
				Index = BufferStaticComplexSurfaceGeometryAndDetailTextureData_VP(Facet);
			}
			else {
				Index = BufferStaticComplexSurfaceGeometryAndDetailTextureData(Facet);
			}

			//Do not draw detail texture if no vertices are near
			if (DetailTextureAnyIsNearBits == 0) {
				drawDetailTexture = false;
			}
		}
		else {
			if (UseVertexProgram) {
				Index = BufferStaticComplexSurfaceGeometry_VP(Facet);
			}
			else {
				Index = BufferStaticComplexSurfaceGeometry(Facet);
			}
		}

		//Save number of points
		m_csPtCount = Index;


		if (UseCVA) {
			DWORD texUnit;
			DWORD texBit;
			DWORD clientTexEnableBitsCopy;

			//Disable texture arrays
			for (texUnit = 0, texBit = 0x1, clientTexEnableBitsCopy = m_clientTexEnableBits; clientTexEnableBitsCopy != 0; texUnit++, texBit <<= 1) {
				//Disable the texture unit
				glClientActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);

				clientTexEnableBitsCopy &= ~texBit;
			}

			//Lock arrays
			glLockArraysEXT(0, Index);

			//Enable texture arrays
			for (texUnit = 0, texBit = 0x1, clientTexEnableBitsCopy = m_clientTexEnableBits; clientTexEnableBitsCopy != 0; texUnit++, texBit <<= 1) {
				//Enable the texture unit
				glClientActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);

				clientTexEnableBitsCopy &= ~texBit;
			}
		}


		//Initialize render passes state information
		m_rpPassCount = 0;
		m_rpForceSingle = false;
		m_rpMasked = ((PolyFlags & PF_Masked) == 0) ? false : true;
		m_rpSetDepthEqual = false;
		m_rpSetDepthMask = false;


		//Do static render passes state setup
		if (UseVertexProgram) {
			const FVector &XAxis = Facet.MapCoords.XAxis;
			const FVector &YAxis = Facet.MapCoords.YAxis;
			FLOAT UDot = XAxis | Facet.MapCoords.Origin;
			FLOAT VDot = YAxis | Facet.MapCoords.Origin;

			glVertexAttrib4fARB(6, XAxis.X, XAxis.Y, XAxis.Z, UDot);
			glVertexAttrib4fARB(7, YAxis.X, YAxis.Y, YAxis.Z, VDot);

			glColor3f(1.0f, 1.0f, 1.0f);
		}
		else {
			glColor3f(1.0f, 1.0f, 1.0f);
		}

		AddRenderPass(Surface.Texture, PolyFlags, 0);

		if (Surface.MacroTexture) {
			AddRenderPass(Surface.MacroTexture, PF_Modulated, -0.5);
		}

		if (Surface.LightMap) {
			AddRenderPass(Surface.LightMap, PF_Modulated, -0.5);
		}
		
		if (Surface.FogMap) {
			//No SetTexEnv support for fog pass
			//It must be the first pass
			RenderPasses();

			AddRenderPass(Surface.FogMap, PF_Highlighted, -0.5);
		}

		// Draw detail texture overlaid, in a separate pass.
		if (drawDetailTexture == true) {
			bool singlePassDetail = (SinglePassDetail != 0);

			//Single pass detail must be done with two normal passes
			singlePassDetail &= (m_rpPassCount == 2);

			if (singlePassDetail) {
				if (m_rpMasked && m_rpForceSingle && !m_rpSetDepthEqual) {
					glDepthFunc(GL_EQUAL);
					m_rpSetDepthEqual = true;
				}

				m_rpPassCount = 4;
				(this->*m_pRenderPassesNoCheck_DualTextureAndDetailTextureProc)(*Surface.DetailTexture);

				if (UseCVA) {
					//Unlock arrays
					glUnlockArraysEXT();
				}
			}
			else {
				RenderPasses();

				if (UseCVA) {
					//Unlock arrays
					glUnlockArraysEXT();
				}

				bool clipDetailTexture = (DetailClipping != 0);

				if (m_rpMasked) {
					//Cannot use detail texture clipping with masked mode
					//It will not work with glDepthFunc(GL_EQUAL)
					clipDetailTexture = false;

					if (m_rpSetDepthEqual == false) {
						glDepthFunc(GL_EQUAL);
						m_rpSetDepthEqual = true;
					}
				}

				if (m_rpSetDepthMask == false) {
					glDepthMask(GL_FALSE);
					m_rpSetDepthMask = true;
				}

				//This function should only be called if at least one polygon will be detail textured
				if (UseVertexProgram) {
					DrawDetailTexture_VP(*Surface.DetailTexture);
				}
				else {
					DrawDetailTexture(*Surface.DetailTexture, Index, clipDetailTexture);
				}
			}
		}
		else {
			RenderPasses();

			if (UseCVA) {
				//Unlock arrays
				glUnlockArraysEXT();
			}
		}

		// UnrealEd selection.
		if (GIsEditor && (PolyFlags & PF_Selected)) {
			SetDefaultVertexProgramState();
			SetDefaultTextureState();

			SetNoTexture(0);
			SetBlend(PF_Highlighted);
			glBegin(GL_TRIANGLE_FAN);
			glColor4f(0.0, 0.0, 0.5, 0.5);			
			for (FSavedPoly* Poly = Facet.Polys; Poly; Poly = Poly->Next) {
				glBegin(GL_TRIANGLE_FAN);
				for (INT i = 0; i < Poly->NumPts; i++)
					glVertex3fv(&Poly->Pts[i]->Point.X);
				glEnd();
			}
			glEnd();
		}

		if (m_rpSetDepthEqual == true) {
			glDepthFunc((GLenum)ZTrickFunc);
		}
		if (m_rpSetDepthMask == true) {
			glDepthMask(GL_TRUE);
			CurrentPolyFlags |= PF_Occlude;
		}

		unclock(ComplexCycles);
		unguard;
	}

	//Modified this routine to always set up detail texture state
	//It should only be called if at least on polygon will be detail textured
	void DrawDetailTexture(FTextureInfo &DetailTextureInfo, INT &BaseClipIndex, bool clipDetailTexture) {
		INT Index = 0;
		FLOAT NearZ  = 380.0f;
		FLOAT RNearZ = 1.0f / NearZ;
		INT NextClipIndex = BaseClipIndex;
		DWORD detailColor;


		//Setup detail texture state
		SetBlend(PF_Modulated);
		if ((clipDetailTexture == false) && UseDetailAlpha) {
			glActiveTextureARB(GL_TEXTURE0_ARB);

			if (ColorizeDetailTextures) {
				glColor3f(0.5f, 1.0f, 0.5f);
			}
			else {
				glColor3f(0.5f, 0.5f, 0.5f);
			}
			SetAlphaTexture(0);

			SetTexEnv(0, PF_Modulated);

			glActiveTextureARB(GL_TEXTURE1_ARB);
			if ((m_texEnableBits & 0x2) == 0) {
				glEnable(GL_TEXTURE_2D);

				m_texEnableBits |= 0x2;
			}
			if ((m_clientTexEnableBits & 0x2) == 0) {
				glClientActiveTextureARB(GL_TEXTURE1_ARB);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits |= 0x2;
			}

			SetTextureNoPanBias(1, DetailTextureInfo, PF_Modulated);
			SetTexEnv(1, PF_Memorized);

			//Enable client texture zero if it is disabled
			if ((m_clientTexEnableBits & 0x1) == 0) {
				//Enable client texture zero
				glClientActiveTextureARB(GL_TEXTURE0_ARB);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits |= 0x1;
			}

			//Check for additional enabled texture units that should be disabled
			DWORD desiredTexBits = ~(~0U << 2);
			if (m_texEnableBits != desiredTexBits) {
				DWORD texUnit;

				//Disable subsequent texture units
				for (texUnit = 2; m_texEnableBits != desiredTexBits; texUnit++) {
					DWORD texBit;

					//See if the texture unit is enabled
					texBit = 1U << texUnit;
					if (texBit & m_texEnableBits) {
						//Disable the texture unit
						glActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
						glDisable(GL_TEXTURE_2D);

						m_texEnableBits &= ~texBit;
					}
				}
			}
			if (m_clientTexEnableBits != desiredTexBits) {
				DWORD texUnit;

				//Disable subsequent client textures
				for (texUnit = 2; m_clientTexEnableBits != desiredTexBits; texUnit++) {
					DWORD texBit;

					//See if the client texture is enabled
					texBit = 1U << texUnit;
					if (texBit & m_clientTexEnableBits) {
						//Disable the client texture
						glClientActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
						glDisableClientState(GL_TEXTURE_COORD_ARRAY);

						m_clientTexEnableBits &= ~texBit;
					}
				}
			}
		}
		else {
			glActiveTextureARB(GL_TEXTURE0_ARB);

			SetTextureNoPanBias(0, DetailTextureInfo, PF_Modulated);
			SetTexEnv(0, PF_Memorized);
			glEnableClientState(GL_COLOR_ARRAY);
			if (clipDetailTexture == true) {
				glEnable(GL_POLYGON_OFFSET_FILL);
			}

			//Enable client texture zero if it is disabled
			if ((m_clientTexEnableBits & 0x1) == 0) {
				//Enable client texture zero
				glClientActiveTextureARB(GL_TEXTURE0_ARB);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits |= 0x1;
			}

			//Check for additional enabled texture units that should be disabled
			DWORD desiredTexBits = ~(~0U << 1);
			if (m_texEnableBits != desiredTexBits) {
				DWORD texUnit;

				//Disable subsequent texture units
				for (texUnit = 1; m_texEnableBits != desiredTexBits; texUnit++) {
					DWORD texBit;

					//See if the texture unit is enabled
					texBit = 1U << texUnit;
					if (texBit & m_texEnableBits) {
						//Disable the texture unit
						glActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
						glDisable(GL_TEXTURE_2D);

						m_texEnableBits &= ~texBit;
					}
				}

				glActiveTextureARB(GL_TEXTURE0_ARB);
			}
			if (m_clientTexEnableBits != desiredTexBits) {
				DWORD texUnit;

				//Disable subsequent client textures
				for (texUnit = 1; m_clientTexEnableBits != desiredTexBits; texUnit++) {
					DWORD texBit;

					//See if the client texture is enabled
					texBit = 1U << texUnit;
					if (texBit & m_clientTexEnableBits) {
						//Disable the client texture
						glClientActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
						glDisableClientState(GL_TEXTURE_COORD_ARRAY);

						m_clientTexEnableBits &= ~texBit;
					}
				}
			}
		}


		detailColor = (ColorizeDetailTextures) ? 0x0080F080 : 0x00808080;
		INT *pNumPts = &MultiDrawCountArray[0];
		DWORD *pDetailTextureIsNear = DetailTextureIsNearArray;
		for (DWORD PolyNum = 0; PolyNum < m_csPolyCount; PolyNum++, pNumPts++, pDetailTextureIsNear++) {
			DWORD NumPts = *pNumPts;
			DWORD isNearBits = *pDetailTextureIsNear;
			//Skip the polygon if it will not be detail textured
			if (isNearBits == 0) {
				Index += NumPts;
				continue;
			}
			INT StartIndex = Index;

			DWORD allPtsBits = ~(~0U << NumPts);
			if ((clipDetailTexture == false) || (isNearBits == allPtsBits)) {
				for (INT i = 0; i < NumPts; i++) {
					const FGLVertex &Point = VertexArray[StartIndex + i];
					FLOAT U = MapDotArray[Index].u;
					FLOAT V = MapDotArray[Index].v;

					if ((clipDetailTexture == false) && UseDetailAlpha) {
						FLOAT PointZ_Times_RNearZ = Point.z * RNearZ;
						TexCoordArray[0][Index].u = PointZ_Times_RNearZ;
						TexCoordArray[0][Index].v = 0.5f;
						TexCoordArray[1][Index].u = (U - TexInfo[1].UPan) * TexInfo[1].UMult;
						TexCoordArray[1][Index].v = (V - TexInfo[1].VPan) * TexInfo[1].VMult;
					}
					else {
						TexCoordArray[0][Index].u = (U - TexInfo[0].UPan) * TexInfo[0].UMult;
						TexCoordArray[0][Index].v = (V - TexInfo[0].VPan) * TexInfo[0].VMult;
						DWORD alpha = appRound((1.0f - (Clamp(Point.z, 0.0f, NearZ) * RNearZ)) * 255.0f);
						ColorArray[Index].color = detailColor | (alpha << 24);
					}
					Index++;
				}
				glDrawArrays(GL_TRIANGLE_FAN, StartIndex, Index - StartIndex);
			}
			else {
				DWORD NextIndex = 0;
				GLuint IndexList[64];
				DWORD isNear_i_bit = 1U << (NumPts - 1);
				DWORD isNear_j_bit = 1U;
				for (INT i = 0, j = NumPts - 1; i < NumPts; j = i++, isNear_j_bit = isNear_i_bit, isNear_i_bit >>= 1) {
					const FGLVertex &Point = VertexArray[StartIndex + i];
					FLOAT U = MapDotArray[Index].u;
					FLOAT V = MapDotArray[Index].v;

					if (((isNear_i_bit & isNearBits) != 0) && ((isNear_j_bit & isNearBits) == 0)) {
						const FGLVertex &PrevPoint = VertexArray[StartIndex + j];
						FLOAT PrevU = MapDotArray[StartIndex + j].u;
						FLOAT PrevV = MapDotArray[StartIndex + j].v;

						FLOAT dist = PrevPoint.z - Point.z;
						if (dist > 0.001f) {
							FLOAT m = (NearZ - Point.z) / dist;
							VertexArray[NextClipIndex].x = (m * (PrevPoint.x - Point.x)) + Point.x;
							VertexArray[NextClipIndex].y = (m * (PrevPoint.y - Point.y)) + Point.y;
							VertexArray[NextClipIndex].z = NearZ;
							TexCoordArray[0][NextClipIndex].u = ((m * (PrevU - U)) + U - TexInfo[0].UPan) * TexInfo[0].UMult;
							TexCoordArray[0][NextClipIndex].v = ((m * (PrevV - V)) + V - TexInfo[0].VPan) * TexInfo[0].VMult;
						}
						else {
							VertexArray[NextClipIndex].x = PrevPoint.x;
							VertexArray[NextClipIndex].y = PrevPoint.y;
							VertexArray[NextClipIndex].z = NearZ;
							TexCoordArray[0][NextClipIndex].u = (PrevU - TexInfo[0].UPan) * TexInfo[0].UMult;
							TexCoordArray[0][NextClipIndex].v = (PrevV - TexInfo[0].VPan) * TexInfo[0].VMult;
						}
						DWORD alpha = 0;
						ColorArray[NextClipIndex].color = detailColor | (alpha << 24);
						IndexList[NextIndex++] = NextClipIndex++;
					}

					if ((isNear_i_bit & isNearBits) != 0) {
						TexCoordArray[0][Index].u = (U - TexInfo[0].UPan) * TexInfo[0].UMult;
						TexCoordArray[0][Index].v = (V - TexInfo[0].VPan) * TexInfo[0].VMult;
						DWORD alpha = appRound((1.0f - (Clamp(Point.z, 0.0f, NearZ) * RNearZ)) * 255.0f);
						ColorArray[Index].color = detailColor | (alpha << 24);
						IndexList[NextIndex++] = Index;
					}

					if (((isNear_i_bit & isNearBits) == 0) && ((isNear_j_bit & isNearBits) != 0)) {
						const FGLVertex &PrevPoint = VertexArray[StartIndex + j];
						FLOAT PrevU = MapDotArray[StartIndex + j].u;
						FLOAT PrevV = MapDotArray[StartIndex + j].v;

						FLOAT dist = Point.z - PrevPoint.z;
						if (dist > 0.001f) {
							FLOAT m = (NearZ - PrevPoint.z) / dist;
							VertexArray[NextClipIndex].x = (m * (Point.x - PrevPoint.x)) + PrevPoint.x;
							VertexArray[NextClipIndex].y = (m * (Point.y - PrevPoint.y)) + PrevPoint.y;
							VertexArray[NextClipIndex].z = NearZ;
							TexCoordArray[0][NextClipIndex].u = ((m * (U - PrevU)) + PrevU - TexInfo[0].UPan) * TexInfo[0].UMult;
							TexCoordArray[0][NextClipIndex].v = ((m * (V - PrevV)) + PrevV - TexInfo[0].VPan) * TexInfo[0].VMult;
						}
						else {
							VertexArray[NextClipIndex].x = Point.x;
							VertexArray[NextClipIndex].y = Point.y;
							VertexArray[NextClipIndex].z = NearZ;
							TexCoordArray[0][NextClipIndex].u = (U - TexInfo[0].UPan) * TexInfo[0].UMult;
							TexCoordArray[0][NextClipIndex].v = (V - TexInfo[0].VPan) * TexInfo[0].VMult;
						}
						DWORD alpha = 0;
						ColorArray[NextClipIndex].color = detailColor | (alpha << 24);
						IndexList[NextIndex++] = NextClipIndex++;
					}
					Index++;
				}
				glDrawElements(GL_TRIANGLE_FAN, NextIndex, GL_UNSIGNED_INT, IndexList);
			}
		}


		//Clear detail texture state
		if ((clipDetailTexture == false) && UseDetailAlpha) {
			glActiveTextureARB(GL_TEXTURE0_ARB);
		}
		else {
			glDisableClientState(GL_COLOR_ARRAY);
			if (clipDetailTexture == true) {
				glDisable(GL_POLYGON_OFFSET_FILL);
			}
		}
		SetTexEnv(0, PF_Modulated);


		BaseClipIndex = NextClipIndex;
	}

	//Modified this routine to always set up detail texture state
	//It should only be called if at least on polygon will be detail textured
	void DrawDetailTexture_VP(FTextureInfo &DetailTextureInfo) {
		INT Index = 0;

		//Setup detail texture state
		SetVertexProgram(m_vpComplexSurfaceDetailAlpha);
		SetBlend(PF_Modulated);
		{
			glActiveTextureARB(GL_TEXTURE0_ARB);

			if (ColorizeDetailTextures) {
				glColor3f(0.5f, 1.0f, 0.5f);
			}
			else {
				glColor3f(0.5f, 0.5f, 0.5f);
			}
			SetAlphaTexture(0);

			SetTexEnv(0, PF_Modulated);

			glActiveTextureARB(GL_TEXTURE1_ARB);
			if ((m_texEnableBits & 0x2) == 0) {
				glEnable(GL_TEXTURE_2D);

				m_texEnableBits |= 0x2;
			}

			SetTextureNoPanBias(1, DetailTextureInfo, PF_Modulated);
			glVertexAttrib4fARB(9, TexInfo[1].UPan, TexInfo[1].VPan, TexInfo[1].UMult, TexInfo[1].VMult);
			SetTexEnv(1, PF_Memorized);

			//Check for additional enabled texture units that should be disabled
			DWORD desiredTexBits = ~(~0U << 2);
			if (m_texEnableBits != desiredTexBits) {
				DWORD texUnit;

				//Disable subsequent texture units
				for (texUnit = 2; m_texEnableBits != desiredTexBits; texUnit++) {
					DWORD texBit;

					//See if the texture unit is enabled
					texBit = 1U << texUnit;
					if (texBit & m_texEnableBits) {
						//Disable the texture unit
						glActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
						glDisable(GL_TEXTURE_2D);

						m_texEnableBits &= ~texBit;
					}
				}
			}
			DWORD desiredClientTexBits = 0;
			if (m_clientTexEnableBits != desiredClientTexBits) {
				DWORD texUnit;

				//Disable subsequent client textures
				for (texUnit = 0; m_clientTexEnableBits != desiredClientTexBits; texUnit++) {
					DWORD texBit;

					//See if the client texture is enabled
					texBit = 1U << texUnit;
					if (texBit & m_clientTexEnableBits) {
						//Disable the client texture
						glClientActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
						glDisableClientState(GL_TEXTURE_COORD_ARRAY);

						m_clientTexEnableBits &= ~texBit;
					}
				}
			}
		}


		INT *pNumPts = &MultiDrawCountArray[0];
		DWORD *pDetailTextureIsNear = DetailTextureIsNearArray;
		for (DWORD PolyNum = 0; PolyNum < m_csPolyCount; PolyNum++, pNumPts++, pDetailTextureIsNear++) {
			DWORD NumPts = *pNumPts;
			DWORD isNearBits = *pDetailTextureIsNear;
			//Skip the polygon if it will not be detail textured
			if (isNearBits == 0) {
				Index += NumPts;
				continue;
			}
			glDrawArrays(GL_TRIANGLE_FAN, Index, NumPts);
			Index += NumPts;
		}


		//Clear detail texture state
		{
			glActiveTextureARB(GL_TEXTURE0_ARB);
		}
		SetTexEnv(0, PF_Modulated);

		return;
	}

	inline void checkAndSetNormal(const FGLNormal *src, FGLNormal *dst) {
		// copy from source to destination
		dst->x = src->x;
		dst->y = src->y;
		dst->z = src->z;

		// calculate length of normal (not squared)
		FLOAT normalLength = dst->x * dst->x + dst->y * dst->y + dst->z * dst->z;
		if (normalLength > 1) { // normal is not valid, so normalize it
			normalLength = sqrtf(normalLength);
			dst->x /= normalLength;
			dst->y /= normalLength;
			dst->z /= normalLength;
		}
	}

	void DrawGouraudPolygonOld(FSceneNode* Frame, FTextureInfo& Info, FTransTexture** Pts, INT NumPts, DWORD PolyFlags, FSpanBuffer* Span) {
#ifdef UTGLR_DEBUG_SHOW_CALL_COUNTS
{
	static int si;
	dout << L"utglr: DrawGouraudPolygonOld = " << si++ << std::endl;
}
#endif

		guard(UOpenGLRenderDevice::DrawGouraudPolygonOld);
		clock(GouraudCycles);

		INT Index     = 0;
		UBOOL Enabled = false;

		SetBlend(PolyFlags);
		SetTextureNoPanBias(0, Info, PolyFlags);
		if (PolyFlags & PF_Modulated) {
			glColor3f(1.0f, 1.0f, 1.0f);
		}
		else {
			glEnableClientState(GL_COLOR_ARRAY);
			Enabled = true;
		}

		for (INT i = 0; i < NumPts; i++) {
			FTransTexture* P = Pts[i];
			TexCoordArray[0][Index].u = P->U * TexInfo[0].UMult;
			TexCoordArray[0][Index].v = P->V * TexInfo[0].VMult;
			if (Enabled)
				ColorArray[Index].color = RGBA_MAKE(
					appRound(P->Light.X * 255.0f), 
					appRound(P->Light.Y * 255.0f), 
					appRound(P->Light.Z * 255.0f),
					255);
			FGLVertex &destVertexArray = VertexArray[Index];
			destVertexArray.x = P->Point.X;
			destVertexArray.y = P->Point.Y;
			destVertexArray.z = P->Point.Z;

			// gruenschloss:
			if (UseTruForm)
				checkAndSetNormal((FGLNormal *) &(P->Normal), &NormalArray[Index]); // conversion *should* work

			Index++;
		}

		// gruenschloss:
#ifdef UTGLR_GRUENSCHLOSS_DEBUG_WIREFRAME
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
#endif

		// gruenschloss:
		if (UseTruForm) {
			glEnable(GL_PN_TRIANGLES_ATI);
			glEnableClientState(GL_NORMAL_ARRAY);
		}

		glDrawArrays(GL_TRIANGLE_FAN, 0, Index);

		// gruenschloss:
#ifdef UTGLR_GRUENSCHLOSS_DEBUG_WIREFRAME
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
#endif

		if ((PolyFlags & (PF_RenderFog | PF_Translucent | PF_Modulated)) == PF_RenderFog) {
			Index = 0;
			if (!Enabled) {
				glEnableClientState(GL_COLOR_ARRAY);
				Enabled = true;
			}
			SetNoTexture(0);
			SetBlend(PF_Highlighted);
			for (INT i = 0; i < NumPts; i++) {
				const FTransTexture* P = Pts[i];
				ColorArray[Index].color = RGBA_MAKE(
					appRound(255 * P->Fog.X), 
					appRound(255 * P->Fog.Y),
					appRound(255 * P->Fog.Z),
					appRound(255 * P->Fog.W));
				Index++;
			}

			glDrawArrays(GL_TRIANGLE_FAN, 0, Index);
		}

		// gruenschloss:
		if (UseTruForm) {
			glDisable(GL_PN_TRIANGLES_ATI);
			glDisableClientState(GL_NORMAL_ARRAY);
		}

		if (Enabled)
			glDisableClientState(GL_COLOR_ARRAY);

		unclock(GouraudCycles);
		unguard;
	}

	inline void EndBuffering(void) {
		if (BufferedVerts > 0) {
			EndBufferingNoCheck();
		}
	}
	void EndBufferingNoCheck(void);

	void Buffer3Verts(FTransTexture** Pts);
	void Buffer3BasicVerts(FTransTexture** Pts);
	void Buffer3ColoredVerts(FTransTexture** Pts);
	void Buffer3FoggedVerts(FTransTexture** Pts);

	void BufferAdditionalClippedVerts(FTransTexture** Pts, INT NumPts);

#ifdef INCLUDE_SSE_CODE
	void Buffer3ColoredVerts_SSE(FTransTexture** Pts);
#endif

	void DrawGouraudPolygon(FSceneNode* Frame, FTextureInfo& Info, FTransTexture** Pts, INT NumPts, DWORD PolyFlags, FSpanBuffer* Span) {
#ifdef UTGLR_DEBUG_SHOW_CALL_COUNTS
{
	static int si;
	dout << L"utglr: DrawGouraudPolygon = " << si++ << std::endl;
}
#endif
		guard(UOpenGLRenderDevice::DrawGouraudPolygon);      
		if (!BufferActorTris || ((NumPts != 3) && !BufferClippedActorTris) || (NumPts > 5)) {
			EndBuffering();

			SetDefaultVertexProgramState();
			SetDefaultTextureState();

			DrawGouraudPolygonOld(Frame, Info, Pts, NumPts, PolyFlags, Span);

			return;
		}

		if (!((m_lastPolyFlags == PolyFlags) && (TexInfo[0].CurrentCacheID == Info.CacheID) &&
			((BufferedVerts + NumPts) < (VERTEX_ARRAY_SIZE - 1 - 4)) && (BufferedVerts > 0)))
		// flush drawing and set the state!
		{
			EndBuffering(); //Flush the vertex array

			//Default vertex program state not needed for polygon caching stage
			//Default texture state not needed for polygon caching stage

			SetBlend(PolyFlags);
			SetTextureNoPanBias(0, Info, PolyFlags);
			RenderFog = false;
			if (PolyFlags & PF_Modulated) {
				glColor3f(1.0f, 1.0f, 1.0f);
				ColorArrayEnabled = false;
			}
			else {
				glEnableClientState(GL_COLOR_ARRAY);
				ColorArrayEnabled = true;

				if (((PolyFlags & (PF_RenderFog | PF_Translucent | PF_Modulated)) == PF_RenderFog) && UseVertexSpecular) {
					RenderFog = true;
				}
			}

			//Save last used poly flags
			m_lastPolyFlags = PolyFlags;

			//Select a buffered verts proc
			if (RenderFog) {
				m_pBuffer3VertsProc = m_pBuffer3FoggedVertsProc;
			}
			else if (ColorArrayEnabled) {
				m_pBuffer3VertsProc = m_pBuffer3ColoredVertsProc;
			}
			else {
				m_pBuffer3VertsProc = m_pBuffer3BasicVertsProc;
			}
		}

		clock(GouraudCycles);

		if (NumPts >= 3) {
			//Buffer 3 vertices from the first (and perhaps only) triangle
			(this->*m_pBuffer3VertsProc)(Pts);

			if (NumPts > 3) {
				//Buffer additional vertices from a clipped triangle
				BufferAdditionalClippedVerts(Pts, NumPts);
			}
		}
		unclock(GouraudCycles);
		unguard;
	}

	void DrawTile(FSceneNode* Frame, FTextureInfo& Info, FLOAT X, FLOAT Y, FLOAT XL, FLOAT YL, FLOAT U, FLOAT V, FLOAT UL, FLOAT VL, class FSpanBuffer* Span, FLOAT Z, FPlane Color, FPlane Fog, DWORD PolyFlags) {
#ifdef UTGLR_DEBUG_SHOW_CALL_COUNTS
{
	static int si;
	dout << L"utglr: DrawTile = " << si++ << std::endl;
}
#endif
		guard(UOpenGLRenderDevice::DrawTile);

		EndBuffering();

		SetDefaultVertexProgramState();
		SetDefaultTextureState();

		clock(TileCycles);

		if (Info.Palette && Info.Palette[128].A != 255 && !(PolyFlags & PF_Translucent)) {
			PolyFlags |= PF_Highlighted;
		}

		SetBlend(PolyFlags);
		SetTextureNoPanBias(0, Info, PolyFlags);
		if (PolyFlags & PF_Modulated) {
			glColor3f(1.0f, 1.0f, 1.0f);
		}
		else {
			glColor3fv(&Color.X);
		}

		FLOAT RFX2_Times_Z = RFX2 * Z;
		FLOAT RFY2_Times_Z = RFY2 * Z;
		FLOAT X_Minus_FrameFX2 = X - Frame->FX2;
		FLOAT Y_Minus_FrameFY2 = Y - Frame->FY2;
		FLOAT TexInfoUMult = TexInfo[0].UMult;
		FLOAT TexInfoVMult = TexInfo[0].VMult;

		glBegin(GL_TRIANGLE_FAN);

		glTexCoord2f((U) * TexInfoUMult, (V) * TexInfoVMult);
		glVertex3f(RFX2_Times_Z * (X_Minus_FrameFX2), RFY2_Times_Z * (Y_Minus_FrameFY2), Z);

		glTexCoord2f((U + UL) * TexInfoUMult, (V) * TexInfoVMult);
		glVertex3f(RFX2_Times_Z * (X_Minus_FrameFX2 + XL), RFY2_Times_Z * (Y_Minus_FrameFY2), Z);

		glTexCoord2f((U + UL) * TexInfoUMult, (V + VL) * TexInfoVMult);
		glVertex3f(RFX2_Times_Z * (X_Minus_FrameFX2 + XL), RFY2_Times_Z * (Y_Minus_FrameFY2 + YL), Z);

		glTexCoord2f((U) * TexInfoUMult, (V + VL) * TexInfoVMult);
		glVertex3f(RFX2_Times_Z * (X_Minus_FrameFX2), RFY2_Times_Z * (Y_Minus_FrameFY2 + YL), Z);

		glEnd();

		unclock(TileCycles);
		unguard;
	}

	void Draw2DLine(FSceneNode* Frame, FPlane Color, DWORD LineFlags, FVector P1, FVector P2) {
#ifdef UTGLR_DEBUG_SHOW_CALL_COUNTS
{
	static int si;
	dout << L"utglr: Draw2DLine = " << si++ << std::endl;
}
#endif
		guard(UOpenGLRenderDevice::Draw2DLine);

		EndBuffering();

		SetDefaultVertexProgramState();
		SetDefaultTextureState();

		SetNoTexture( 0 );
		SetBlend( PF_Highlighted );
		glColor3fv( &Color.X );
		glBegin( GL_LINES );
		glVertex3f( RFX2*P1.Z*(P1.X-Frame->FX2), RFY2*P1.Z*(P1.Y-Frame->FY2), P1.Z );
		glVertex3f( RFX2*P2.Z*(P2.X-Frame->FX2), RFY2*P2.Z*(P2.Y-Frame->FY2), P2.Z );
		glEnd();
		unguard;
	}

	void Draw3DLine(FSceneNode* Frame, FPlane Color, DWORD LineFlags, FVector P1, FVector P2) {
#ifdef UTGLR_DEBUG_SHOW_CALL_COUNTS
{
	static int si;
	dout << L"utglr: Draw3DLine = " << si++ << std::endl;
}
#endif
		guard(UOpenGLRenderDevice::Draw3DLine);

		EndBuffering();

		SetDefaultVertexProgramState();
		SetDefaultTextureState();

		P1 = P1.TransformPointBy( Frame->Coords );
		P2 = P2.TransformPointBy( Frame->Coords );
		if (Frame->Viewport->IsOrtho()) {
			// Zoom.
			P1.X = (P1.X) / Frame->Zoom + Frame->FX2;
			P1.Y = (P1.Y) / Frame->Zoom + Frame->FY2;
			P2.X = (P2.X) / Frame->Zoom + Frame->FX2;
			P2.Y = (P2.Y) / Frame->Zoom + Frame->FY2;
			P1.Z = P2.Z = 1;

			// See if points form a line parallel to our line of sight (i.e. line appears as a dot).
			if( Abs(P2.X-P1.X)+Abs(P2.Y-P1.Y)>=0.2 )
				Draw2DLine( Frame, Color, LineFlags, P1, P2 );
			else if( Frame->Viewport->Actor->OrthoZoom < ORTHO_LOW_DETAIL )
				Draw2DPoint( Frame, Color, LINE_None, P1.X-1, P1.Y-1, P1.X+1, P1.Y+1, P1.Z );
		}
		else {
			SetNoTexture( 0 );
			SetBlend( PF_Highlighted );
			glColor3fv( &Color.X );
			glBegin( GL_LINES );
			glVertex3fv( &P1.X );
			glVertex3fv( &P2.X );
			glEnd();
		}
		unguard;
	}

	void Draw2DPoint(FSceneNode* Frame, FPlane Color, DWORD LineFlags, FLOAT X1, FLOAT Y1, FLOAT X2, FLOAT Y2, FLOAT Z) {
#ifdef UTGLR_DEBUG_SHOW_CALL_COUNTS
{
	static int si;
	dout << L"utglr: Draw2DPoint = " << si++ << std::endl;
}
#endif
		guard(UOpenGLRenderDevice::Draw2DPoint);

		EndBuffering();

		SetDefaultVertexProgramState();
		SetDefaultTextureState();

		SetBlend( PF_Highlighted );
		SetNoTexture( 0 );
		glColor3fv( &Color.X ); // vogel: was 4 - ONLY FOR UT!
		glBegin( GL_TRIANGLE_FAN );
		glVertex3f( RFX2*Z*(X1-Frame->FX2), RFY2*Z*(Y1-Frame->FY2), Z );
		glVertex3f( RFX2*Z*(X2-Frame->FX2), RFY2*Z*(Y1-Frame->FY2), Z );
		glVertex3f( RFX2*Z*(X2-Frame->FX2), RFY2*Z*(Y2-Frame->FY2), Z );
		glVertex3f( RFX2*Z*(X1-Frame->FX2), RFY2*Z*(Y2-Frame->FY2), Z );
		glEnd();
		unguard;
	}

	void ClearZ(FSceneNode* Frame) {
#ifdef UTGLR_DEBUG_SHOW_CALL_COUNTS
{
	static int si;
	dout << L"utglr: ClearZ = " << si++ << std::endl;
}
#endif
		guard(UOpenGLRenderDevice::ClearZ);

		EndBuffering();

#ifdef UTGLR_SKYBOX_HACK
		if (SkyboxHack) {
			unsigned int u;

			//Save and mark skybox textures as valid
			for (u = 0; u < m_csCurSkyboxTexNum; u++) {
				m_skyboxTexList[u].texId = m_potentialSkyboxTexList[u];
				m_skyboxTexList[u].valid = true;
			}

			if (m_csAllSkyboxTexDetected == true) {
				return;
			}
		}
#endif

		//Default vertex program state not required for glClear
		//Default texture state not required for glClear

		SetBlend(PF_Occlude);
		glClear(GL_DEPTH_BUFFER_BIT);
		unguard;
	}

	void PushHit(const BYTE* Data, INT Count) {
		guard(UOpenGLRenderDevice::PushHit);
		glPushName(Count);
		for(INT i = 0; i < Count; i += 4)
			glPushName(*(INT*)(Data + i));
		unguard;
	}

	void PopHit(INT Count, UBOOL bForce) {
		guard(UOpenGLRenderDevice::PopHit);
		glPopName();
		for(INT i = 0; i < Count; i += 4)
			glPopName();
		//!!implement bforce
		unguard;
	}

	void GetStats(TCHAR* Result) {
		guard(UOpenGLRenderDevice::GetStats);
		appSprintf
		(
			Result,
			TEXT("OpenGL stats: Bind=%04.1f Image=%04.1f Complex=%04.1f Gouraud=%04.1f Tile=%04.1f"),
#ifdef UTGLR_DX_BUILD
			(float)BindCycles,
			(float)ImageCycles,
			(float)ComplexCycles,
			(float)GouraudCycles,
			(float)TileCycles
#else
			GSecondsPerCycle*1000 * BindCycles,
			GSecondsPerCycle*1000 * ImageCycles,
			GSecondsPerCycle*1000 * ComplexCycles,
			GSecondsPerCycle*1000 * GouraudCycles,
			GSecondsPerCycle*1000 * TileCycles
#endif
		);
		unguard;
	}

	void ReadPixels(FColor* Pixels) {
		guard(UOpenGLRenderDevice::ReadPixels);
		INT SizeX, SizeY;
#ifdef UTGLR_DX_BUILD
		SizeX = m_FrameX;
		SizeY = m_FrameY;
#else
		SizeX = Viewport->SizeX;
		SizeY = Viewport->SizeY;
#endif
		glReadPixels(0, 0, SizeX, SizeY, GL_RGBA, GL_UNSIGNED_BYTE, Pixels);
		for (INT i = 0; i < SizeY / 2; i++) {
			for (INT j = 0; j < SizeX; j++) {
				Exchange(Pixels[j + i * SizeX].R, Pixels[j + (SizeY - 1 - i) * SizeX].B);
				Exchange(Pixels[j + i * SizeX].G, Pixels[j + (SizeY - 1 - i) * SizeX].G);
				Exchange(Pixels[j + i * SizeX].B, Pixels[j + (SizeY - 1 - i) * SizeX].R);
			}
		}
		unguard;
	}

	void EndFlash() {
#ifdef UTGLR_DEBUG_SHOW_CALL_COUNTS
{
	static int si;
	dout << L"utglr: EndFlash = " << si++ << std::endl;
}
#endif
		if (FlashScale != FPlane(.5,.5,.5,0) || FlashFog != FPlane(0,0,0,0)) {
			EndBuffering();

			SetDefaultVertexProgramState();
			SetDefaultTextureState();

			SetBlend(PF_Highlighted);
			SetNoTexture(0);
			glColor4f(FlashFog.X, FlashFog.Y, FlashFog.Z, 1.0 - Min(FlashScale.X * 2.0f, 1.0f));
			FLOAT RFX2 = RProjZ;
			FLOAT RFY2 = RProjZ * Aspect;
			glBegin(GL_TRIANGLE_FAN);
				glVertex3f(RFX2 * (-1.0f), RFY2 * (-1.0f), 1.0f);
				glVertex3f(RFX2 * (+1.0f), RFY2 * (-1.0f), 1.0f);
				glVertex3f(RFX2 * (+1.0f), RFY2 * (+1.0f), 1.0f);
				glVertex3f(RFX2 * (-1.0f), RFY2 * (+1.0f), 1.0f);
			glEnd();
		}
	}

	void PrecacheTexture(FTextureInfo& Info, DWORD PolyFlags) {
		guard(UOpenGLRenderDevice::PrecacheTexture);
		SetTextureNoPanBias(0, Info, PolyFlags);
		unguard;
	}
};
IMPLEMENT_CLASS(UOpenGLRenderDevice);


void UOpenGLRenderDevice::SC_AddBoolConfigParam(const TCHAR *pName, UBOOL &param, ECppProperty EC_CppProperty, INT InOffset, UBOOL defaultValue) {
	param = ((defaultValue) != 0) ? -1 : 0; //Not exactly handled as "// Boolean 0 (false) or 1 (true).", as listed at the UBOOL definition, in all places
	new(GetClass(), pName, RF_Public)UBoolProperty(EC_CppProperty, InOffset, TEXT("Options"), CPF_Config);
	GConfig->GetBool(g_pSection, pName, param);
}

void UOpenGLRenderDevice::SC_AddIntConfigParam(const TCHAR *pName, INT &param, ECppProperty EC_CppProperty, INT InOffset, INT defaultValue) {
	param = defaultValue;
	new(GetClass(), pName, RF_Public)UIntProperty(EC_CppProperty, InOffset, TEXT("Options"), CPF_Config);
}

void UOpenGLRenderDevice::SC_AddFloatConfigParam(const TCHAR *pName, FLOAT &param, ECppProperty EC_CppProperty, INT InOffset, FLOAT defaultValue) {
	param = defaultValue;
	new(GetClass(), pName, RF_Public)UFloatProperty(EC_CppProperty, InOffset, TEXT("Options"), CPF_Config);
}


void UOpenGLRenderDevice::CacheTextureInfo(FCachedTexture *pBind, const FTextureInfo &Info, DWORD PolyFlags) {
#if 0
{
	dout.fill('0'); dout << std::hex;
	dout << L"utglr: CacheId = "
		<< std::setw(8) << (DWORD)((QWORD)Info.CacheID >> 32) << L":"
		<< std::setw(8) << (DWORD)((QWORD)Info.CacheID & 0xFFFFFFFF) << std::endl;
	dout.fill(' '); dout << std::dec;
}
{
	dout << L"utglr: NumMips = " << Info.NumMips << std::endl;
}
{
	unsigned int u;

	for (u = 0; u < NUM_CTMap_TREES; u++) {
		dout << L"utglr: ZeroPrefixBindMap[" << u << L"] Size = " << m_zeroPrefixBindMaps[u].size() << std::endl;
	}
	for (u = 0; u < NUM_CTMap_TREES; u++) {
		dout << L"utglr: NonZeroPrefixBindMap[" << u << L"] Size = " << m_nonZeroPrefixBindMaps[u].size() << std::endl;
	}
}
#endif

	// Figure out OpenGL-related scaling for the texture.
	DWORD texFlags = 0;
	pBind->BaseMip = 0;
	INT UBits = Info.Mips[0]->UBits;
	INT VBits = Info.Mips[0]->VBits;
	INT UCopyBits = 0;
	INT VCopyBits = 0;
	if ((UBits - VBits) > MaxLogUOverV) {
		VCopyBits += (UBits - VBits) - MaxLogUOverV;
		VBits = UBits - MaxLogUOverV;
	}
	if ((VBits - UBits) > MaxLogVOverU) {
		UCopyBits += (VBits - UBits) - MaxLogVOverU;
		UBits = VBits - MaxLogVOverU;
	}
	if (UBits < MinLogTextureSize) {
		UCopyBits += MinLogTextureSize - UBits;
		UBits += MinLogTextureSize - UBits;
	}
	if (VBits < MinLogTextureSize) {
		VCopyBits += MinLogTextureSize - VBits;
		VBits += MinLogTextureSize - VBits;
	}
	if (UBits > MaxLogTextureSize) {
		pBind->BaseMip += UBits - MaxLogTextureSize;
		VBits -= UBits - MaxLogTextureSize;
		UBits = MaxLogTextureSize;
		if (VBits < 0) {
			VCopyBits = -VBits;
			VBits = 0;
		}
	}
	if (VBits > MaxLogTextureSize) {
		pBind->BaseMip += VBits - MaxLogTextureSize;
		UBits -= VBits - MaxLogTextureSize;
		VBits = MaxLogTextureSize;
		if (UBits < 0) {
			UCopyBits = -UBits;
			UBits = 0;
		}
	}

	pBind->UBits = UBits;
	pBind->VBits = VBits;

	pBind->UMult = 1.0f / (Info.UScale * (Info.USize << UCopyBits));
	pBind->VMult = 1.0f / (Info.VScale * (Info.VSize << VCopyBits));

	pBind->UClampVal = Info.UClamp - 1;
	pBind->VClampVal = Info.VClamp - 1;

	//Check for texture that does not require clamping
	//No clamp required if ((Info.UClamp == Info.USize) & (Info.VClamp == Info.VSize))
	if (((Info.UClamp ^ Info.USize) | (Info.VClamp ^ Info.VSize)) == 0) {
		texFlags |= TEX_FLAG_NO_CLAMP;
	}

	if (UseTNT) {
		pBind->MaxLevel = Max(UBits, VBits);
	}
	else {
		pBind->MaxLevel = Min(UBits, VBits) - MinLogTextureSize;
	}


	//Determine texture type
	//PolyFlags PF_Masked cannot change if existing texture is updated as it caches texture type information here
	bool paletted = UsePalette && Info.Palette;
	if (!UseAlphaPalette) {
		paletted = paletted && !(PolyFlags & PF_Masked) && (Info.Palette[0].A == 255);
	}

	if ((Info.Format == TEXF_DXT1) && SupportsTC) {
		if (TexDXT1ToDXT3) {
			pBind->texType = TEX_TYPE_COMPRESSED_DXT1;
			//texSourceFormat not used for compressed textures
			pBind->texInternalFormat = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
		}
		else {
			pBind->texType = TEX_TYPE_COMPRESSED_DXT1;
			//texSourceFormat not used for compressed textures
			pBind->texInternalFormat = GL_COMPRESSED_RGB_S3TC_DXT1_EXT;
		}
	}
	else if (paletted) {
		pBind->texType = TEX_TYPE_PALETTED;
		pBind->texSourceFormat = GL_COLOR_INDEX;
		pBind->texInternalFormat = GL_COLOR_INDEX8_EXT;
	}
	else if (Info.Palette) {
		pBind->texType = TEX_TYPE_HAS_PALETTE;
		pBind->texSourceFormat = GL_RGBA;
		if (Use4444Textures) {
			pBind->texInternalFormat = GL_RGBA4; // vogel: will cause banding in menus
		}
		else {
			pBind->texInternalFormat = GL_RGBA8;
		}
	}
	else {
		pBind->texType = TEX_TYPE_NORMAL;
		if (UseBGRATextures) {
			pBind->texSourceFormat = GL_BGRA_EXT;
			if (texFlags & TEX_FLAG_NO_CLAMP) {
				pBind->pConvertBGRA7777 = ConvertBGRA7777_BGRA8888_NoClamp;
			}
			else {
				pBind->pConvertBGRA7777 = ConvertBGRA7777_BGRA8888;
			}
		}
		else {
			pBind->texSourceFormat = GL_RGBA;
			pBind->pConvertBGRA7777 = ConvertBGRA7777_RGBA8888;
		}
		pBind->texInternalFormat = GL_RGBA8;
	}

	return;
}


void UOpenGLRenderDevice::ConvertDXT1_DXT3(const FMipmapBase *Mip, INT Level) {
	DWORD *pSrc = (DWORD *)Mip->DataPtr;
	DWORD *pDest = (DWORD *)m_texConvertCtx.pCompose;
	DWORD numBlocks = 1 << (Max(0, m_texConvertCtx.pBind->UBits - Level - 2) + Max(0, m_texConvertCtx.pBind->VBits - Level - 2));
	for (DWORD block = 0; block < numBlocks; block++) {
		*pDest = 0xFFFFFFFF;
		*(pDest + 1) = 0xFFFFFFFF;
		*(pDest + 2) = *pSrc;
		*(pDest + 3) = *(pSrc + 1);
		pSrc += 2;
		pDest += 4;
	}

#ifdef UTGLR_DEBUG_SHOW_TEX_CONVERT_COUNTS
	{
		static int si;
		dout << L"utglr: ConvertDXT1_DXT3 = " << si++ << std::endl;
	}
#endif
}

void UOpenGLRenderDevice::ConvertP8_P8(const FMipmapBase *Mip, INT Level) {
	BYTE* Ptr = (BYTE*)m_texConvertCtx.pCompose;
	INT StepBits = m_texConvertCtx.stepBits;
	DWORD UMask = Mip->USize - 1;
	DWORD VMask = Mip->VSize - 1;
	INT ij_inc = 1 << StepBits;
	INT i_stop = 1 << Max(0, m_texConvertCtx.pBind->VBits - Level + StepBits);
	INT j_stop = 1 << Max(0, m_texConvertCtx.pBind->UBits - Level + StepBits);
	INT i = 0;
	do { //i_stop always >= 1
		BYTE* Base = (BYTE*)Mip->DataPtr + (i & VMask) * Mip->USize;
		INT j = 0;
		do { //j_stop always >= 1
			*Ptr++ = Base[j & UMask];
		} while ((j += ij_inc) < j_stop);
	} while ((i += ij_inc) < i_stop);

#ifdef UTGLR_DEBUG_SHOW_TEX_CONVERT_COUNTS
	{
		static int si;
		dout << L"utglr: ConvertP8_P8 = " << si++ << std::endl;
	}
#endif
}

void UOpenGLRenderDevice::ConvertP8_P8_NoStep(const FMipmapBase *Mip, INT Level) {
	BYTE* Ptr = (BYTE*)m_texConvertCtx.pCompose;
	DWORD UMask = Mip->USize - 1;
	DWORD VMask = Mip->VSize - 1;
	INT i_stop = 1 << Max(0, m_texConvertCtx.pBind->VBits - Level);
	INT j_stop = 1 << Max(0, m_texConvertCtx.pBind->UBits - Level);
	INT i = 0;
	do { //i_stop always >= 1
		BYTE* Base = (BYTE*)Mip->DataPtr + (i & VMask) * Mip->USize;
		INT j = 0;
		do { //j_stop always >= 1
			*Ptr++ = Base[j & UMask];
		} while ((j += 1) < j_stop);
	} while ((i += 1) < i_stop);

#ifdef UTGLR_DEBUG_SHOW_TEX_CONVERT_COUNTS
	{
		static int si;
		dout << L"utglr: ConvertP8_P8_NoStep = " << si++ << std::endl;
	}
#endif
}

void UOpenGLRenderDevice::ConvertP8_RGBA8888(const FMipmapBase *Mip, const FColor *Palette, INT Level) {
	FColor* Ptr = (FColor*)m_texConvertCtx.pCompose;
	INT StepBits = m_texConvertCtx.stepBits;
	DWORD UMask = Mip->USize - 1;
	DWORD VMask = Mip->VSize - 1;
	INT ij_inc = 1 << StepBits;
	INT i_stop = 1 << Max(0, m_texConvertCtx.pBind->VBits - Level + StepBits);
	INT j_stop = 1 << Max(0, m_texConvertCtx.pBind->UBits - Level + StepBits);
	INT i = 0;
	do { //i_stop always >= 1
		BYTE* Base = (BYTE*)Mip->DataPtr + (i & VMask) * Mip->USize;
		INT j = 0;
		do { //j_stop always >= 1;
			*Ptr++ = Palette[Base[j & UMask]];
		} while ((j += ij_inc) < j_stop);
	} while ((i += ij_inc) < i_stop);

#ifdef UTGLR_DEBUG_SHOW_TEX_CONVERT_COUNTS
	{
		static int si;
		dout << L"utglr: ConvertP8_RGBA8888 = " << si++ << std::endl;
	}
#endif
}

void UOpenGLRenderDevice::ConvertP8_RGBA8888_NoStep(const FMipmapBase *Mip, const FColor *Palette, INT Level) {
	FColor* Ptr = (FColor*)m_texConvertCtx.pCompose;
	DWORD UMask = Mip->USize - 1;
	DWORD VMask = Mip->VSize - 1;
	INT i_stop = 1 << Max(0, m_texConvertCtx.pBind->VBits - Level);
	INT j_stop = 1 << Max(0, m_texConvertCtx.pBind->UBits - Level);
	INT i = 0;
	do { //i_stop always >= 1
		BYTE* Base = (BYTE*)Mip->DataPtr + (i & VMask) * Mip->USize;
		INT j = 0;
		do { //j_stop always >= 1;
			*Ptr++ = Palette[Base[j & UMask]];
		} while ((j += 1) < j_stop);
	} while ((i += 1) < i_stop);

#ifdef UTGLR_DEBUG_SHOW_TEX_CONVERT_COUNTS
	{
		static int si;
		dout << L"utglr: ConvertP8_RGBA8888_NoStep = " << si++ << std::endl;
	}
#endif
}

void UOpenGLRenderDevice::ConvertBGRA7777_BGRA8888(const FMipmapBase *Mip, INT Level) {
	FColor* Ptr = (FColor*)m_texConvertCtx.pCompose;
	INT StepBits = m_texConvertCtx.stepBits;
	DWORD VMask = Mip->VSize - 1;
	DWORD VClampVal = m_texConvertCtx.pBind->VClampVal;
	DWORD UMask = Mip->USize - 1;
	DWORD UClampVal = m_texConvertCtx.pBind->UClampVal;
	INT ij_inc = 1 << StepBits;
	INT i_stop = 1 << Max(0, m_texConvertCtx.pBind->VBits - Level + StepBits);
	INT j_stop = 1 << Max(0, m_texConvertCtx.pBind->UBits - Level + StepBits);
	INT i = 0;
	do { //i_stop always >= 1
		FColor* Base = (FColor*)Mip->DataPtr + Min<DWORD>(i & VMask, VClampVal) * Mip->USize;
		INT j = 0;
		do { //j_stop always >= 1;
			const FColor& Src = Base[Min<DWORD>(j & UMask, UClampVal)];
			*(DWORD *)Ptr = *((DWORD *)&Src) * 2; // because of 7777
			Ptr++;
		} while ((j += ij_inc) < j_stop);
	} while ((i += ij_inc) < i_stop);

#ifdef UTGLR_DEBUG_SHOW_TEX_CONVERT_COUNTS
	{
		static int si;
		dout << L"utglr: ConvertBGRA7777_BGRA8888 = " << si++ << std::endl;
	}
#endif
}

void UOpenGLRenderDevice::ConvertBGRA7777_BGRA8888_NoClamp(const FMipmapBase *Mip, INT Level) {
	FColor* Ptr = (FColor*)m_texConvertCtx.pCompose;
	INT StepBits = m_texConvertCtx.stepBits;
	DWORD VMask = Mip->VSize - 1;
	DWORD UMask = Mip->USize - 1;
	INT ij_inc = 1 << StepBits;
	INT i_stop = 1 << Max(0, m_texConvertCtx.pBind->VBits - Level + StepBits);
	INT j_stop = 1 << Max(0, m_texConvertCtx.pBind->UBits - Level + StepBits);
	INT i = 0;
	do { //i_stop always >= 1
		FColor* Base = (FColor*)Mip->DataPtr + (DWORD)(i & VMask) * Mip->USize;
		INT j = 0;
		do { //j_stop always >= 1;
			const FColor& Src = Base[(DWORD)(j & UMask)];
			*(DWORD *)Ptr = *((DWORD *)&Src) * 2; // because of 7777
			Ptr++;
		} while ((j += ij_inc) < j_stop);
	} while ((i += ij_inc) < i_stop);

#ifdef UTGLR_DEBUG_SHOW_TEX_CONVERT_COUNTS
	{
		static int si;
		dout << L"utglr: ConvertBGRA7777_BGRA8888_NoClamp = " << si++ << std::endl;
	}
#endif
}

void UOpenGLRenderDevice::ConvertBGRA7777_RGBA8888(const FMipmapBase *Mip, INT Level) {
	FColor* Ptr = (FColor*)m_texConvertCtx.pCompose;
	INT StepBits = m_texConvertCtx.stepBits;
	DWORD VMask = Mip->VSize - 1;
	DWORD VClampVal = m_texConvertCtx.pBind->VClampVal;
	DWORD UMask = Mip->USize - 1;
	DWORD UClampVal = m_texConvertCtx.pBind->UClampVal;
	INT ij_inc = 1 << StepBits;
	INT i_stop = 1 << Max(0, m_texConvertCtx.pBind->VBits - Level + StepBits);
	INT j_stop = 1 << Max(0, m_texConvertCtx.pBind->UBits - Level + StepBits);
	INT i = 0;
	do { //i_stop always >= 1
		FColor* Base = (FColor*)Mip->DataPtr + Min<DWORD>(i & VMask, VClampVal) * Mip->USize;
		INT j = 0;
		do { //j_stop always >= 1;
			const FColor& Src = Base[Min<DWORD>(j & UMask, UClampVal)];
			// vogel: optimize it.
			Ptr->R = 2 * Src.B;
			Ptr->G = 2 * Src.G;
			Ptr->B = 2 * Src.R;
			Ptr->A = 2 * Src.A; // because of 7777

			Ptr++;
		} while ((j += ij_inc) < j_stop);
	} while ((i += ij_inc) < i_stop);

#ifdef UTGLR_DEBUG_SHOW_TEX_CONVERT_COUNTS
	{
		static int si;
		dout << L"utglr: ConvertBGRA7777_RGBA8888 = " << si++ << std::endl;
	}
#endif
}

void UOpenGLRenderDevice::SetBlendNoCheck(DWORD PolyFlags) {
	guardSlow(UOpenGLRenderDevice::SetBlend);

	// Detect changes in the blending modes.
	DWORD Xor = CurrentPolyFlags ^ PolyFlags;
	if (Xor & (PF_Translucent | PF_Modulated | PF_Invisible | PF_Occlude | PF_Masked | PF_Highlighted)) {
		if (Xor & PF_Masked) {
			if (PolyFlags & PF_Masked) {
				glEnable(GL_ALPHA_TEST);
			}
			else {
				glDisable(GL_ALPHA_TEST);
			}
		}
		if (Xor & (PF_Translucent | PF_Modulated | PF_Highlighted)) {
			if(!(PolyFlags & (PF_Translucent | PF_Modulated | PF_Highlighted))) {
				glDisable(GL_BLEND);
			}
			else {
				if (!(CurrentPolyFlags & (PF_Translucent | PF_Modulated | PF_Highlighted))) {
					glEnable(GL_BLEND);
				}
				if (PolyFlags & PF_Translucent) {
					glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
				}
				else if (PolyFlags & PF_Modulated) {
					glBlendFunc(GL_DST_COLOR, GL_SRC_COLOR);
				}
				else if (PolyFlags & PF_Highlighted) {
					glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
				}
			}
		}
		if (Xor & PF_Invisible) {
			UBOOL flag = ((PolyFlags & PF_Invisible) == 0) ? GL_TRUE : GL_FALSE;
			glColorMask(flag, flag, flag, flag);
		}
		if (Xor & PF_Occlude) {
			UBOOL flag = ((PolyFlags & PF_Occlude) == 0) ? GL_FALSE : GL_TRUE;
			glDepthMask(flag);
		}

		CurrentPolyFlags = PolyFlags;
	}

	unguardSlow;
}

void UOpenGLRenderDevice::SetTexEnvNoCheck(INT Multi, DWORD PolyFlags) {
	guardSlow(UOpenGLRenderDevice::SetTexEnv);

	DWORD Xor = CurrentEnvFlags[Multi] ^ PolyFlags;

	if (Xor & PF_Modulated) {
		if (PolyFlags & PF_Modulated) {
			if (SUPPORTS_GL_EXT_texture_env_combine && (Multi != 0)) {
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);

				glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, GL_MODULATE);
				glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_EXT, GL_MODULATE);

				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_EXT, GL_TEXTURE);
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_EXT, GL_PREVIOUS_EXT);

				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_EXT, GL_TEXTURE);
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_ALPHA_EXT, GL_PREVIOUS_EXT);

				glTexEnvf(GL_TEXTURE_ENV, GL_RGB_SCALE_EXT, 2.0f);
			}
			else {
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			}
		}
	}
	if (Xor & PF_Highlighted) {
		if (PolyFlags & PF_Highlighted) {
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);

			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, GL_INTERPOLATE_EXT);
			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_EXT, GL_REPLACE);

			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_EXT, GL_TEXTURE);
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_EXT, GL_PRIMARY_COLOR_EXT);
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB_EXT, GL_PREVIOUS_EXT);

			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_EXT, GL_TEXTURE);

			glTexEnvf(GL_TEXTURE_ENV, GL_RGB_SCALE_EXT, 2.0f);
		}
	}
	if (Xor & PF_Memorized) {	// Abused for detail texture approach
		if (PolyFlags & PF_Memorized) {
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);

			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, GL_INTERPOLATE_EXT);
			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_EXT, GL_REPLACE);

			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_EXT, GL_TEXTURE);
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_EXT, GL_PRIMARY_COLOR_EXT);
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB_EXT, GL_PREVIOUS_EXT);

			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_EXT, GL_TEXTURE);

			glTexEnvf(GL_TEXTURE_ENV, GL_RGB_SCALE_EXT, 1.0f);
		}
	}
	if (Xor & PF_FlatShaded) {
		if (PolyFlags & PF_FlatShaded) {
		}
	}

	CurrentEnvFlags[Multi] = PolyFlags;

	unguardSlow;
}

void UOpenGLRenderDevice::RenderPassesNoCheck(void) {
	guard(UOpenGLRenderDevice::RenderPassesNoCheck);

	INT i;

	SetBlend(MultiPass.TMU[0].PolyFlags);

	for (i = 0; i < m_rpPassCount; i++) {
		if (i != 0) {
			DWORD texBit;

			glActiveTextureARB(GL_TEXTURE0_ARB + i);

			texBit = 1 << i;
			if ((m_texEnableBits & texBit) == 0) {
				glEnable(GL_TEXTURE_2D);

				m_texEnableBits |= texBit;
			}
			if ((m_clientTexEnableBits & texBit) == 0) {
				glClientActiveTextureARB(GL_TEXTURE0_ARB + i);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits |= texBit;
			}

			SetTexEnv(i, MultiPass.TMU[i].PolyFlags);
		}

		SetTexture(i, *MultiPass.TMU[i].Info, MultiPass.TMU[i].PolyFlags, MultiPass.TMU[i].PanBias);
	}

	//Check for additional enabled texture units that should be disabled
	DWORD desiredTexBits = ~(~0U << m_rpPassCount);
	if (m_texEnableBits != desiredTexBits) {
		DWORD texUnit;

		//Disable subsequent texture units
		for (texUnit = m_rpPassCount; m_texEnableBits != desiredTexBits; texUnit++) {
			DWORD texBit;

			//See if the texture unit is enabled
			texBit = 1U << texUnit;
			if (texBit & m_texEnableBits) {
				//Disable the texture unit
				glActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisable(GL_TEXTURE_2D);

				m_texEnableBits &= ~texBit;
			}
		}
	}
	if (m_clientTexEnableBits != desiredTexBits) {
		DWORD texUnit;

		//Disable subsequent client textures
		for (texUnit = m_rpPassCount; m_clientTexEnableBits != desiredTexBits; texUnit++) {
			DWORD texBit;

			//See if the client texture is enabled
			texBit = 1U << texUnit;
			if (texBit & m_clientTexEnableBits) {
				//Disable the texture unit
				glClientActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits &= ~texBit;
			}
		}
	}

	for (INT t = 0; t < m_rpPassCount; t++) {
		FLOAT UPan = TexInfo[t].UPan;
		FLOAT VPan = TexInfo[t].VPan;
		FLOAT UMult = TexInfo[t].UMult;
		FLOAT VMult = TexInfo[t].VMult;
		const FGLMapDot *pMapDot = &MapDotArray[0];
		FGLTexCoord *pTexCoord = &TexCoordArray[t][0];

		for (INT Index = 0; Index < m_csPtCount; Index++) {
			pTexCoord->u = (pMapDot->u - UPan) * UMult;
			pTexCoord->v = (pMapDot->v - VPan) * VMult;

			pMapDot++;
			pTexCoord++;
		}
	}

	if (UseMultiDrawArrays && (m_csPolyCount > 1)) {
		glMultiDrawArraysEXT(GL_TRIANGLE_FAN, MultiDrawFirstArray, MultiDrawCountArray, m_csPolyCount);
	}
	else {
		for (INT PolyNum = 0; PolyNum < m_csPolyCount; PolyNum++) {
			glDrawArrays(GL_TRIANGLE_FAN, MultiDrawFirstArray[PolyNum], MultiDrawCountArray[PolyNum]);
		}
	}

	glActiveTextureARB(GL_TEXTURE0_ARB);

#if 0
{
	dout << L"utglr: PassCount = " << m_rpPassCount << std::endl;
}
#endif
	m_rpPassCount = 0;

	unguard;
}

#ifdef INCLUDE_SSE_CODE
void UOpenGLRenderDevice::RenderPassesNoCheck_SSE(void) {
	guard(UOpenGLRenderDevice::RenderPassesNoCheck);

	INT i;

	SetBlend(MultiPass.TMU[0].PolyFlags);

	for (i = 0; i < m_rpPassCount; i++) {
		if (i != 0) {
			DWORD texBit;

			glActiveTextureARB(GL_TEXTURE0_ARB + i);

			texBit = 1 << i;
			if ((m_texEnableBits & texBit) == 0) {
				glEnable(GL_TEXTURE_2D);

				m_texEnableBits |= texBit;
			}
			if ((m_clientTexEnableBits & texBit) == 0) {
				glClientActiveTextureARB(GL_TEXTURE0_ARB + i);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits |= texBit;
			}

			SetTexEnv(i, MultiPass.TMU[i].PolyFlags);
		}

		SetTexture(i, *MultiPass.TMU[i].Info, MultiPass.TMU[i].PolyFlags, MultiPass.TMU[i].PanBias);
	}

	//Check for additional enabled texture units that should be disabled
	DWORD desiredTexBits = ~(~0U << m_rpPassCount);
	if (m_texEnableBits != desiredTexBits) {
		DWORD texUnit;

		//Disable subsequent texture units
		for (texUnit = m_rpPassCount; m_texEnableBits != desiredTexBits; texUnit++) {
			DWORD texBit;

			//See if the texture unit is enabled
			texBit = 1U << texUnit;
			if (texBit & m_texEnableBits) {
				//Disable the texture unit
				glActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisable(GL_TEXTURE_2D);

				m_texEnableBits &= ~texBit;
			}
		}
	}
	if (m_clientTexEnableBits != desiredTexBits) {
		DWORD texUnit;

		//Disable subsequent client textures
		for (texUnit = m_rpPassCount; m_clientTexEnableBits != desiredTexBits; texUnit++) {
			DWORD texBit;

			//See if the client texture is enabled
			texBit = 1U << texUnit;
			if (texBit & m_clientTexEnableBits) {
				//Disable the texture unit
				glClientActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits &= ~texBit;
			}
		}
	}

	for (INT t = 0; t < m_rpPassCount; t++) {
		__m128 uvPan;
		__m128 uvMult;
		const FGLMapDot *pMapDot = &MapDotArray[0];
		FGLTexCoord *pTexCoord = &TexCoordArray[t][0];

		uvPan = _mm_setzero_ps();
		uvMult = _mm_setzero_ps();
		uvPan = _mm_loadl_pi(uvPan, (const __m64 *)&TexInfo[t].UPan);
		uvMult = _mm_loadl_pi(uvMult, (const __m64 *)&TexInfo[t].UMult);
		uvPan = _mm_movelh_ps(uvPan, uvPan);
		uvMult = _mm_movelh_ps(uvMult, uvMult);

		for (INT Index = 0; Index < m_csPtCount; Index += 2) {
			__m128 data;

			data = _mm_load_ps((const float *)pMapDot);
			data = _mm_sub_ps(data, uvPan);
			data = _mm_mul_ps(data, uvMult);
			_mm_store_ps((float *)pTexCoord, data);

			pMapDot += 2;
			pTexCoord += 2;
		}
	}

	if (UseMultiDrawArrays && (m_csPolyCount > 1)) {
		glMultiDrawArraysEXT(GL_TRIANGLE_FAN, MultiDrawFirstArray, MultiDrawCountArray, m_csPolyCount);
	}
	else {
		for (INT PolyNum = 0; PolyNum < m_csPolyCount; PolyNum++) {
			glDrawArrays(GL_TRIANGLE_FAN, MultiDrawFirstArray[PolyNum], MultiDrawCountArray[PolyNum]);
		}
	}

	glActiveTextureARB(GL_TEXTURE0_ARB);

#if 0
{
	dout << L"utglr: PassCount = " << m_rpPassCount << std::endl;
}
#endif
	m_rpPassCount = 0;

	unguard;
}
#endif //INCLUDE_SSE_CODE

void UOpenGLRenderDevice::RenderPassesNoCheck_VP(void) {
	guard(UOpenGLRenderDevice::RenderPassesNoCheck_VP);

	INT i;

	SetBlend(MultiPass.TMU[0].PolyFlags);

	SetVertexProgram(m_vpComplexSurface[m_rpPassCount - 1]);

	for (i = 0; i < m_rpPassCount; i++) {
		DWORD texBit;

		texBit = 1 << i;

		if (i != 0) {
			glActiveTextureARB(GL_TEXTURE0_ARB + i);

			if ((m_texEnableBits & texBit) == 0) {
				glEnable(GL_TEXTURE_2D);

				m_texEnableBits |= texBit;
			}

			SetTexEnv(i, MultiPass.TMU[i].PolyFlags);
		}
		//Disable client textures
		if ((m_clientTexEnableBits & texBit) != 0) {
			glClientActiveTextureARB(GL_TEXTURE0_ARB + i);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);

			m_clientTexEnableBits &= ~texBit;
		}

		SetTexture(i, *MultiPass.TMU[i].Info, MultiPass.TMU[i].PolyFlags, MultiPass.TMU[i].PanBias);
		
		glVertexAttrib4fARB(i + 8, TexInfo[i].UPan, TexInfo[i].VPan, TexInfo[i].UMult, TexInfo[i].VMult);
	}

	//Check for additional enabled texture units that should be disabled
	DWORD desiredTexBits = ~(~0U << m_rpPassCount);
	if (m_texEnableBits != desiredTexBits) {
		DWORD texUnit;

		//Disable subsequent texture units
		for (texUnit = m_rpPassCount; m_texEnableBits != desiredTexBits; texUnit++) {
			DWORD texBit;

			//See if the texture unit is enabled
			texBit = 1U << texUnit;
			if (texBit & m_texEnableBits) {
				//Disable the texture unit
				glActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisable(GL_TEXTURE_2D);

				m_texEnableBits &= ~texBit;
			}
		}
	}
	DWORD desiredClientTexBits = 0;
	if (m_clientTexEnableBits != desiredClientTexBits) {
		DWORD texUnit;

		//Disable subsequent client textures
		for (texUnit = m_rpPassCount; m_clientTexEnableBits != desiredClientTexBits; texUnit++) {
			DWORD texBit;

			//See if the client texture is enabled
			texBit = 1U << texUnit;
			if (texBit & m_clientTexEnableBits) {
				//Disable the texture unit
				glClientActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits &= ~texBit;
			}
		}
	}

	if (UseMultiDrawArrays && (m_csPolyCount > 1)) {
		glMultiDrawArraysEXT(GL_TRIANGLE_FAN, MultiDrawFirstArray, MultiDrawCountArray, m_csPolyCount);
	}
	else {
		for (INT PolyNum = 0; PolyNum < m_csPolyCount; PolyNum++) {
			glDrawArrays(GL_TRIANGLE_FAN, MultiDrawFirstArray[PolyNum], MultiDrawCountArray[PolyNum]);
		}
	}

	glActiveTextureARB(GL_TEXTURE0_ARB);

#if 0
{
	dout << L"utglr: PassCount = " << m_rpPassCount << std::endl;
}
#endif
	m_rpPassCount = 0;

	unguard;
}

void UOpenGLRenderDevice::RenderPassesNoCheck_DualTextureAndDetailTexture(FTextureInfo &DetailTextureInfo) {
	guard(UOpenGLRenderDevice::RenderPassesNoCheck_DualTextureAndDetailTexture);

	INT i;
	FLOAT NearZ  = 380.0f;
	FLOAT RNearZ = 1.0f / NearZ;

	SetBlend(MultiPass.TMU[0].PolyFlags);

	//Detail texture uses first two texture units
	for (i = 2; i < 4; i++) {
		DWORD texBit;

		texBit = 1 << i;

		if (i != 0) {
			glActiveTextureARB(GL_TEXTURE0_ARB + i);

			if ((m_texEnableBits & texBit) == 0) {
				glEnable(GL_TEXTURE_2D);

				m_texEnableBits |= texBit;
			}
			if ((m_clientTexEnableBits & texBit) == 0) {
				glClientActiveTextureARB(GL_TEXTURE0_ARB + i);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits |= texBit;
			}

			SetTexEnv(i, MultiPass.TMU[i - 2].PolyFlags);
		}

		SetTexture(i, *MultiPass.TMU[i - 2].Info, MultiPass.TMU[i - 2].PolyFlags, MultiPass.TMU[i - 2].PanBias);
	}

	if (ColorizeDetailTextures) {
		glColor3f(0.25f, 0.5f, 0.25f);
	}
	else {
		glColor3f(0.5f, 0.5f, 0.5f);
	}

	glActiveTextureARB(GL_TEXTURE0_ARB);
	SetAlphaTexture(0);
	{
		DWORD texBit = 1 << 0;
		if ((m_texEnableBits & texBit) == 0) {
			glEnable(GL_TEXTURE_2D);

			m_texEnableBits |= texBit;
		}
		if ((m_clientTexEnableBits & texBit) == 0) {
			glClientActiveTextureARB(GL_TEXTURE0_ARB);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);

			m_clientTexEnableBits |= texBit;
		}
	}

	glActiveTextureARB(GL_TEXTURE1_ARB);
	SetTexEnv(1, PF_Highlighted);
	SetTextureNoPanBias(1, DetailTextureInfo, PF_Modulated);
	{
		DWORD texBit = 1 << 1;
		if ((m_texEnableBits & texBit) == 0) {
			glEnable(GL_TEXTURE_2D);

			m_texEnableBits |= texBit;
		}
		if ((m_clientTexEnableBits & texBit) == 0) {
			glClientActiveTextureARB(GL_TEXTURE1_ARB);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);

			m_clientTexEnableBits |= texBit;
		}
	}

	//Check for additional enabled texture units that should be disabled
	DWORD desiredTexBits = ~(~0U << m_rpPassCount);
	if (m_texEnableBits != desiredTexBits) {
		DWORD texUnit;

		//Disable subsequent texture units
		for (texUnit = m_rpPassCount; m_texEnableBits != desiredTexBits; texUnit++) {
			DWORD texBit;

			//See if the texture unit is enabled
			texBit = 1U << texUnit;
			if (texBit & m_texEnableBits) {
				//Disable the texture unit
				glActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisable(GL_TEXTURE_2D);

				m_texEnableBits &= ~texBit;
			}
		}
	}
	if (m_clientTexEnableBits != desiredTexBits) {
		DWORD texUnit;

		//Disable subsequent client textures
		for (texUnit = m_rpPassCount; m_clientTexEnableBits != desiredTexBits; texUnit++) {
			DWORD texBit;

			//See if the client texture is enabled
			texBit = 1U << texUnit;
			if (texBit & m_clientTexEnableBits) {
				//Disable the texture unit
				glClientActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits &= ~texBit;
			}
		}
	}

	//Alpha texture for detail texture uses texture unit 0
	{
		INT t = 0;
		const FGLVertex *pVertex = &VertexArray[0];
		FGLTexCoord *pTexCoord = &TexCoordArray[t][0];

		for (INT Index = 0; Index < m_csPtCount; Index++) {
			pTexCoord->u = pVertex->z * RNearZ;
			pTexCoord->v = 0.5f;

			pVertex++;
			pTexCoord++;
		}
	}
	//Detail texture uses texture unit 1
	//Remaining 2 textures use texture units 2 and 3
	for (INT t = 1; t < 4; t++) {
		FLOAT UPan = TexInfo[t].UPan;
		FLOAT VPan = TexInfo[t].VPan;
		FLOAT UMult = TexInfo[t].UMult;
		FLOAT VMult = TexInfo[t].VMult;
		const FGLMapDot *pMapDot = &MapDotArray[0];
		FGLTexCoord *pTexCoord = &TexCoordArray[t][0];

		for (INT Index = 0; Index < m_csPtCount; Index++) {
			pTexCoord->u = (pMapDot->u - UPan) * UMult;
			pTexCoord->v = (pMapDot->v - VPan) * VMult;

			pMapDot++;
			pTexCoord++;
		}
	}

	if (UseMultiDrawArrays && (m_csPolyCount > 1)) {
		glMultiDrawArraysEXT(GL_TRIANGLE_FAN, MultiDrawFirstArray, MultiDrawCountArray, m_csPolyCount);
	}
	else {
		for (INT PolyNum = 0; PolyNum < m_csPolyCount; PolyNum++) {
			glDrawArrays(GL_TRIANGLE_FAN, MultiDrawFirstArray[PolyNum], MultiDrawCountArray[PolyNum]);
		}
	}

	glActiveTextureARB(GL_TEXTURE0_ARB);

#if 0
{
	dout << L"utglr: PassCount = " << m_rpPassCount << std::endl;
}
#endif
	m_rpPassCount = 0;

	unguard;
}

#ifdef INCLUDE_SSE_CODE
void UOpenGLRenderDevice::RenderPassesNoCheck_DualTextureAndDetailTexture_SSE(FTextureInfo &DetailTextureInfo) {
	guard(UOpenGLRenderDevice::RenderPassesNoCheck_DualTextureAndDetailTexture);

	INT i;
	FLOAT NearZ  = 380.0f;
	FLOAT RNearZ = 1.0f / NearZ;

	SetBlend(MultiPass.TMU[0].PolyFlags);

	//Detail texture uses first two texture units
	for (i = 2; i < 4; i++) {
		DWORD texBit;

		texBit = 1 << i;

		if (i != 0) {
			glActiveTextureARB(GL_TEXTURE0_ARB + i);

			if ((m_texEnableBits & texBit) == 0) {
				glEnable(GL_TEXTURE_2D);

				m_texEnableBits |= texBit;
			}
			if ((m_clientTexEnableBits & texBit) == 0) {
				glClientActiveTextureARB(GL_TEXTURE0_ARB + i);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits |= texBit;
			}

			SetTexEnv(i, MultiPass.TMU[i - 2].PolyFlags);
		}

		SetTexture(i, *MultiPass.TMU[i - 2].Info, MultiPass.TMU[i - 2].PolyFlags, MultiPass.TMU[i - 2].PanBias);
	}

	if (ColorizeDetailTextures) {
		glColor3f(0.25f, 0.5f, 0.25f);
	}
	else {
		glColor3f(0.5f, 0.5f, 0.5f);
	}

	glActiveTextureARB(GL_TEXTURE0_ARB);
	SetAlphaTexture(0);
	{
		DWORD texBit = 1 << 0;
		if ((m_texEnableBits & texBit) == 0) {
			glEnable(GL_TEXTURE_2D);

			m_texEnableBits |= texBit;
		}
		if ((m_clientTexEnableBits & texBit) == 0) {
			glClientActiveTextureARB(GL_TEXTURE0_ARB);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);

			m_clientTexEnableBits |= texBit;
		}
	}

	glActiveTextureARB(GL_TEXTURE1_ARB);
	SetTexEnv(1, PF_Highlighted);
	SetTextureNoPanBias(1, DetailTextureInfo, PF_Modulated);
	{
		DWORD texBit = 1 << 1;
		if ((m_texEnableBits & texBit) == 0) {
			glEnable(GL_TEXTURE_2D);

			m_texEnableBits |= texBit;
		}
		if ((m_clientTexEnableBits & texBit) == 0) {
			glClientActiveTextureARB(GL_TEXTURE1_ARB);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);

			m_clientTexEnableBits |= texBit;
		}
	}

	//Check for additional enabled texture units that should be disabled
	DWORD desiredTexBits = ~(~0U << m_rpPassCount);
	if (m_texEnableBits != desiredTexBits) {
		DWORD texUnit;

		//Disable subsequent texture units
		for (texUnit = m_rpPassCount; m_texEnableBits != desiredTexBits; texUnit++) {
			DWORD texBit;

			//See if the texture unit is enabled
			texBit = 1U << texUnit;
			if (texBit & m_texEnableBits) {
				//Disable the texture unit
				glActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisable(GL_TEXTURE_2D);

				m_texEnableBits &= ~texBit;
			}
		}
	}
	if (m_clientTexEnableBits != desiredTexBits) {
		DWORD texUnit;

		//Disable subsequent client textures
		for (texUnit = m_rpPassCount; m_clientTexEnableBits != desiredTexBits; texUnit++) {
			DWORD texBit;

			//See if the client texture is enabled
			texBit = 1U << texUnit;
			if (texBit & m_clientTexEnableBits) {
				//Disable the texture unit
				glClientActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits &= ~texBit;
			}
		}
	}

	//Alpha texture for detail texture uses texture unit 0
	{
		INT t = 0;
		const FGLVertex *pVertex = &VertexArray[0];
		FGLTexCoord *pTexCoord = &TexCoordArray[t][0];

		for (INT Index = 0; Index < m_csPtCount; Index++) {
			pTexCoord->u = pVertex->z * RNearZ;
			pTexCoord->v = 0.5f;

			pVertex++;
			pTexCoord++;
		}
	}
	//Detail texture uses texture unit 1
	//Remaining 2 textures use texture units 2 and 3
	for (INT t = 1; t < 4; t++) {
		__m128 uvPan;
		__m128 uvMult;
		const FGLMapDot *pMapDot = &MapDotArray[0];
		FGLTexCoord *pTexCoord = &TexCoordArray[t][0];

		uvPan = _mm_setzero_ps();
		uvMult = _mm_setzero_ps();
		uvPan = _mm_loadl_pi(uvPan, (const __m64 *)&TexInfo[t].UPan);
		uvMult = _mm_loadl_pi(uvMult, (const __m64 *)&TexInfo[t].UMult);
		uvPan = _mm_movelh_ps(uvPan, uvPan);
		uvMult = _mm_movelh_ps(uvMult, uvMult);

		for (INT Index = 0; Index < m_csPtCount; Index += 2) {
			__m128 data;

			data = _mm_load_ps((const float *)pMapDot);
			data = _mm_sub_ps(data, uvPan);
			data = _mm_mul_ps(data, uvMult);
			_mm_store_ps((float *)pTexCoord, data);

			pMapDot += 2;
			pTexCoord += 2;
		}
	}

	if (UseMultiDrawArrays && (m_csPolyCount > 1)) {
		glMultiDrawArraysEXT(GL_TRIANGLE_FAN, MultiDrawFirstArray, MultiDrawCountArray, m_csPolyCount);
	}
	else {
		for (INT PolyNum = 0; PolyNum < m_csPolyCount; PolyNum++) {
			glDrawArrays(GL_TRIANGLE_FAN, MultiDrawFirstArray[PolyNum], MultiDrawCountArray[PolyNum]);
		}
	}

	glActiveTextureARB(GL_TEXTURE0_ARB);

#if 0
{
	dout << L"utglr: PassCount = " << m_rpPassCount << std::endl;
}
#endif
	m_rpPassCount = 0;

	unguard;
}
#endif //INCLUDE_SSE_CODE

void UOpenGLRenderDevice::RenderPassesNoCheck_DualTextureAndDetailTexture_VP(FTextureInfo &DetailTextureInfo) {
	guard(UOpenGLRenderDevice::RenderPassesNoCheck_DualTextureAndDetailTexture_VP);

	INT i;

	SetBlend(MultiPass.TMU[0].PolyFlags);

	SetVertexProgram(m_vpComplexSurfaceDualTextureAndDetailTexture);

	//Detail texture uses first two texture units
	for (i = 2; i < 4; i++) {
		DWORD texBit;

		texBit = 1 << i;

		if (i != 0) {
			glActiveTextureARB(GL_TEXTURE0_ARB + i);

			if ((m_texEnableBits & texBit) == 0) {
				glEnable(GL_TEXTURE_2D);

				m_texEnableBits |= texBit;
			}

			SetTexEnv(i, MultiPass.TMU[i - 2].PolyFlags);
		}
		//Disable client textures
		if ((m_clientTexEnableBits & texBit) != 0) {
			glClientActiveTextureARB(GL_TEXTURE0_ARB + i);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);

			m_clientTexEnableBits &= ~texBit;
		}

		SetTexture(i, *MultiPass.TMU[i - 2].Info, MultiPass.TMU[i - 2].PolyFlags, MultiPass.TMU[i - 2].PanBias);

		glVertexAttrib4fARB(i + 8, TexInfo[i].UPan, TexInfo[i].VPan, TexInfo[i].UMult, TexInfo[i].VMult);
	}

	if (ColorizeDetailTextures) {
		glColor3f(0.25f, 0.5f, 0.25f);
	}
	else {
		glColor3f(0.5f, 0.5f, 0.5f);
	}

	glActiveTextureARB(GL_TEXTURE0_ARB);
	SetAlphaTexture(0);
	{
		DWORD texBit = 1 << 0;
		if ((m_texEnableBits & texBit) == 0) {
			glEnable(GL_TEXTURE_2D);

			m_texEnableBits |= texBit;
		}
		if ((m_clientTexEnableBits & texBit) != 0) {
			glClientActiveTextureARB(GL_TEXTURE0_ARB);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);

			m_clientTexEnableBits &= ~texBit;
		}
	}

	glActiveTextureARB(GL_TEXTURE1_ARB);
	SetTexEnv(1, PF_Highlighted);
	SetTextureNoPanBias(1, DetailTextureInfo, PF_Modulated);
	{
		DWORD texBit = 1 << 1;
		if ((m_texEnableBits & texBit) == 0) {
			glEnable(GL_TEXTURE_2D);

			m_texEnableBits |= texBit;
		}
		if ((m_clientTexEnableBits & texBit) != 0) {
			glClientActiveTextureARB(GL_TEXTURE1_ARB);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);

			m_clientTexEnableBits &= ~texBit;
		}
	}
	glVertexAttrib4fARB(9, TexInfo[1].UPan, TexInfo[1].VPan, TexInfo[1].UMult, TexInfo[1].VMult);

	//Check for additional enabled texture units that should be disabled
	DWORD desiredTexBits = ~(~0U << m_rpPassCount);
	if (m_texEnableBits != desiredTexBits) {
		DWORD texUnit;

		//Disable subsequent texture units
		for (texUnit = m_rpPassCount; m_texEnableBits != desiredTexBits; texUnit++) {
			DWORD texBit;

			//See if the texture unit is enabled
			texBit = 1U << texUnit;
			if (texBit & m_texEnableBits) {
				//Disable the texture unit
				glActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisable(GL_TEXTURE_2D);

				m_texEnableBits &= ~texBit;
			}
		}
	}
	DWORD desiredClientTexBits = 0;
	if (m_clientTexEnableBits != desiredClientTexBits) {
		DWORD texUnit;

		//Disable subsequent client textures
		for (texUnit = m_rpPassCount; m_clientTexEnableBits != desiredClientTexBits; texUnit++) {
			DWORD texBit;

			//See if the client texture is enabled
			texBit = 1U << texUnit;
			if (texBit & m_clientTexEnableBits) {
				//Disable the texture unit
				glClientActiveTextureARB(GL_TEXTURE0_ARB + texUnit);
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);

				m_clientTexEnableBits &= ~texBit;
			}
		}
	}

	if (UseMultiDrawArrays && (m_csPolyCount > 1)) {
		glMultiDrawArraysEXT(GL_TRIANGLE_FAN, MultiDrawFirstArray, MultiDrawCountArray, m_csPolyCount);
	}
	else {
		for (INT PolyNum = 0; PolyNum < m_csPolyCount; PolyNum++) {
			glDrawArrays(GL_TRIANGLE_FAN, MultiDrawFirstArray[PolyNum], MultiDrawCountArray[PolyNum]);
		}
	}

	glActiveTextureARB(GL_TEXTURE0_ARB);

#if 0
{
	dout << L"utglr: PassCount = " << m_rpPassCount << std::endl;
}
#endif
	m_rpPassCount = 0;

	unguard;
}

INT UOpenGLRenderDevice::BufferStaticComplexSurfaceGeometry(const FSurfaceFacet& Facet) {
	FLOAT UDot = Facet.MapCoords.XAxis | Facet.MapCoords.Origin;
	FLOAT VDot = Facet.MapCoords.YAxis | Facet.MapCoords.Origin;
	INT Index = 0;

	// Buffer "static" geometry.
	m_csPolyCount = 0;
	for (FSavedPoly* Poly = Facet.Polys; Poly; Poly = Poly->Next) {
		MultiDrawFirstArray[m_csPolyCount] = Index;

		for (INT i = 0; i < Poly->NumPts; i++) {
			const FVector &Point = Poly->Pts[i]->Point;

			FGLMapDot &destMapDotArray = MapDotArray[Index];
			destMapDotArray.u = (Facet.MapCoords.XAxis | Point) - UDot;
			destMapDotArray.v = (Facet.MapCoords.YAxis | Point) - VDot;

			FGLVertex &destVertexArray = VertexArray[Index];
			destVertexArray.x = Point.X;
			destVertexArray.y = Point.Y;
			destVertexArray.z = Point.Z;
			Index++;
		}

		MultiDrawCountArray[m_csPolyCount] = Poly->NumPts;
		m_csPolyCount++;
	}

	return Index;
}

INT UOpenGLRenderDevice::BufferStaticComplexSurfaceGeometry_VP(const FSurfaceFacet& Facet) {
	INT Index = 0;

	// Buffer "static" geometry.
	m_csPolyCount = 0;
	for (FSavedPoly* Poly = Facet.Polys; Poly; Poly = Poly->Next) {
		MultiDrawFirstArray[m_csPolyCount] = Index;

		for (INT i = 0; i < Poly->NumPts; i++) {
			const FVector &Point = Poly->Pts[i]->Point;

			FGLVertex &destVertexArray = VertexArray[Index];
			destVertexArray.x = Point.X;
			destVertexArray.y = Point.Y;
			destVertexArray.z = Point.Z;
			Index++;
		}

		MultiDrawCountArray[m_csPolyCount] = Poly->NumPts;
		m_csPolyCount++;
	}

	return Index;
}

#ifdef INCLUDE_SSE_CODE
//TODO: update to support MultiDraw array updating
INT UOpenGLRenderDevice::BufferStaticComplexSurfaceGeometry_SSE(const FSurfaceFacet& Facet) {
	const FCoords &Coords = Facet.MapCoords;
	const FSavedPoly *Poly = Facet.Polys;
	FGLMapDot *pMapDotArray = &MapDotArray[0];
	FGLVertex *pVertexArray = &VertexArray[0];
	INT Index;

	// Buffer "static" geometry.
	__asm {
		xor eax, eax

		mov ecx, Coords

		movss xmm0, [ecx]Coords.Origin.X
		movss xmm1, [ecx]Coords.Origin.Y
		movss xmm2, [ecx]Coords.Origin.Z

		movss xmm3, [ecx]Coords.XAxis.X
		mulss xmm3, xmm0
		movss xmm4, [ecx]Coords.XAxis.Y
		mulss xmm4, xmm1
		movss xmm5, [ecx]Coords.XAxis.Z
		mulss xmm5, xmm2
		addss xmm3, xmm4
		addss xmm3, xmm5

		movss xmm4, [ecx]Coords.YAxis.X
		mulss xmm4, xmm0
		movss xmm5, [ecx]Coords.YAxis.Y
		mulss xmm5, xmm1
		movss xmm6, [ecx]Coords.YAxis.Z
		mulss xmm6, xmm2
		addss xmm4, xmm5
		addss xmm4, xmm6

		mov esi, pMapDotArray
		mov edi, pVertexArray

		mov ecx, Coords

		mov ebx, Poly

		test ebx, ebx
		jz l_poly_done

l_poly_loop:
		mov edx, [ebx]Poly.NumPts
		test edx, edx
		jz l_pts_done

		xor edx, edx

l_pts_loop:
			mov eax, [ebx + edx*4]Poly.Pts

			movss xmm0, [eax]FOutVector.Point.X
			movss xmm1, [eax]FOutVector.Point.Y
			movss xmm2, [eax]FOutVector.Point.Z

			movss xmm5, [ecx]Coords.XAxis.X
			mulss xmm5, xmm0
			movss xmm6, [ecx]Coords.XAxis.Y
			mulss xmm6, xmm1
			movss xmm7, [ecx]Coords.XAxis.Z
			mulss xmm7, xmm2
			addss xmm5, xmm6
			addss xmm5, xmm7
			subss xmm5, xmm3
			movss [esi]FGLMapDot.u, xmm5

			movss xmm5, [ecx]Coords.YAxis.X
			mulss xmm5, xmm0
			movss xmm6, [ecx]Coords.YAxis.Y
			mulss xmm6, xmm1
			movss xmm7, [ecx]Coords.YAxis.Z
			mulss xmm7, xmm2
			addss xmm5, xmm6
			addss xmm5, xmm7
			subss xmm5, xmm4
			movss [esi]FGLMapDot.v, xmm5

			add esi, TYPE FGLMapDot

			movss [edi]FGLVertex.x, xmm0
			movss [edi]FGLVertex.y, xmm1
			movss [edi]FGLVertex.z, xmm2

			add edi, TYPE FGLVertex

			inc Index

			inc edx
			cmp edx, [ebx]Poly.NumPts
			jnz l_pts_loop

l_pts_done:
		mov ebx, [ebx]Poly.Next
		test ebx, ebx
		jnz l_poly_loop

l_poly_done:
		mov Index, eax
	}

	return Index;
}
#endif

INT UOpenGLRenderDevice::BufferStaticComplexSurfaceGeometryAndDetailTextureData(const FSurfaceFacet& Facet) {
	FLOAT UDot = Facet.MapCoords.XAxis | Facet.MapCoords.Origin;
	FLOAT VDot = Facet.MapCoords.YAxis | Facet.MapCoords.Origin;
	FLOAT NearZ  = 380.0f;
	INT Index = 0;

	// Buffer "static" geometry.
	DWORD *pDetailTextureIsNear = DetailTextureIsNearArray;
	DetailTextureAnyIsNearBits = 0;
	m_csPolyCount = 0;
	for (FSavedPoly* Poly = Facet.Polys; Poly; Poly = Poly->Next) {
		DWORD isNear = 0;

		MultiDrawFirstArray[m_csPolyCount] = Index;

		for (INT i = 0; i < Poly->NumPts; i++) {
			const FVector &Point = Poly->Pts[i]->Point;

			FGLMapDot &destMapDotArray = MapDotArray[Index];
			destMapDotArray.u = (Facet.MapCoords.XAxis | Point) - UDot;
			destMapDotArray.v = (Facet.MapCoords.YAxis | Point) - VDot;

			FGLVertex &destVertexArray = VertexArray[Index];
			destVertexArray.x = Point.X;
			destVertexArray.y = Point.Y;
			destVertexArray.z = Point.Z;

			isNear <<= 1;
			if (Point.Z < NearZ) {
				isNear |= 1;
			}

			Index++;
		}

		*pDetailTextureIsNear++ = isNear;
		DetailTextureAnyIsNearBits |= isNear;

		MultiDrawCountArray[m_csPolyCount] = Poly->NumPts;
		m_csPolyCount++;
	}

	return Index;
}

INT UOpenGLRenderDevice::BufferStaticComplexSurfaceGeometryAndDetailTextureData_VP(const FSurfaceFacet& Facet) {
	FLOAT NearZ  = 380.0f;
	INT Index = 0;

	// Buffer "static" geometry.
	DWORD *pDetailTextureIsNear = DetailTextureIsNearArray;
	DetailTextureAnyIsNearBits = 0;
	m_csPolyCount = 0;
	for (FSavedPoly* Poly = Facet.Polys; Poly; Poly = Poly->Next) {
		DWORD isNear = 0;

		MultiDrawFirstArray[m_csPolyCount] = Index;

		for (INT i = 0; i < Poly->NumPts; i++) {
			const FVector &Point = Poly->Pts[i]->Point;

			FGLVertex &destVertexArray = VertexArray[Index];
			destVertexArray.x = Point.X;
			destVertexArray.y = Point.Y;
			destVertexArray.z = Point.Z;

			isNear <<= 1;
			if (Point.Z < NearZ) {
				isNear |= 1;
			}

			Index++;
		}

		*pDetailTextureIsNear++ = isNear;
		DetailTextureAnyIsNearBits |= isNear;

		MultiDrawCountArray[m_csPolyCount] = Poly->NumPts;
		m_csPolyCount++;
	}

	return Index;
}

void UOpenGLRenderDevice::EndBufferingNoCheck(void) {
	SetDefaultVertexProgramState();
	SetDefaultTextureState();

	clock(GouraudCycles);

	if (RenderFog) {
		glEnableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
		glEnable(GL_COLOR_SUM_EXT);
	}

		// gruenschloss:
#ifdef UTGLR_GRUENSCHLOSS_DEBUG_WIREFRAME
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
#endif

	// gruenschloss:
	if (UseTruForm) {
		glEnable(GL_PN_TRIANGLES_ATI);
		glEnableClientState(GL_NORMAL_ARRAY);
	}

	// Actually render the triangles.
	glDrawArrays(GL_TRIANGLES, 0, BufferedVerts);

	// gruenschloss:
	if (UseTruForm) {
		glDisable(GL_PN_TRIANGLES_ATI);
		glDisableClientState(GL_NORMAL_ARRAY);
	}

		// gruenschloss:
#ifdef UTGLR_GRUENSCHLOSS_DEBUG_WIREFRAME
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
#endif

	if (ColorArrayEnabled) {
		glDisableClientState(GL_COLOR_ARRAY);
	}
	if (RenderFog) {
		glDisableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
		glDisable(GL_COLOR_SUM_EXT);
	}
	BufferedVerts = 0;

	unclock(GouraudCycles);
}

// gruenschloss: Buffer3Verts only gets used when TruForm is enabled.
void UOpenGLRenderDevice::Buffer3Verts(FTransTexture** Pts) {
	FGLTexCoord *pTexCoordArray = &TexCoordArray[0][BufferedVerts];
	FGLColor *pColorArray = &ColorArray[BufferedVerts];
	FGLVertex *pVertexArray = &VertexArray[BufferedVerts];

	// gruenschloss:
	FGLNormal *pNormalArray = &NormalArray[BufferedVerts];

	for (INT i = 0; i < 3; i++) {
		const FTransTexture* P = Pts[i];
		(pTexCoordArray + i)->u = P->U * TexInfo[0].UMult;
		(pTexCoordArray + i)->v = P->V * TexInfo[0].VMult;
		if (RenderFog) {
			FLOAT One_Minus_FogW = 1.0f - P->Fog.W;
			(pColorArray + i)->color = RGBA_MAKE(
				appRound(P->Light.X * 255.0f * One_Minus_FogW), 
				appRound(P->Light.Y * 255.0f * One_Minus_FogW), 
				appRound(P->Light.Z * 255.0f * One_Minus_FogW),
				255);

			(pColorArray + i)->specular = RGBA_MAKE(
				appRound(255 * P->Fog.X), 
				appRound(255 * P->Fog.Y),
				appRound(255 * P->Fog.Z),
				0);
		}
		else if (ColorArrayEnabled) {
			(pColorArray + i)->color = RGBA_MAKE(
				appRound(P->Light.X * 255.0f), 
				appRound(P->Light.Y * 255.0f), 
				appRound(P->Light.Z * 255.0f),
				255);
		}
		(pVertexArray + i)->x = P->Point.X;
		(pVertexArray + i)->y = P->Point.Y;
		(pVertexArray + i)->z = P->Point.Z;

		// gruenschloss:
		// if (UseTruForm) {
			(pNormalArray + i)->x = P->Normal.X;
			(pNormalArray + i)->y = P->Normal.Y;
			(pNormalArray + i)->z = P->Normal.Z;
		// }
	}
	BufferedVerts += 3;
}

void UOpenGLRenderDevice::Buffer3BasicVerts(FTransTexture** Pts) {
	FGLTexCoord *pTexCoordArray = &TexCoordArray[0][BufferedVerts];
	FGLVertex *pVertexArray = &VertexArray[BufferedVerts];
	for (INT i = 0; i < 3; i++) {
		const FTransTexture* P = Pts[i];
		(pTexCoordArray + i)->u = P->U * TexInfo[0].UMult;
		(pTexCoordArray + i)->v = P->V * TexInfo[0].VMult;
		(pVertexArray + i)->x = P->Point.X;
		(pVertexArray + i)->y = P->Point.Y;
		(pVertexArray + i)->z = P->Point.Z;
	}
	BufferedVerts += 3;
}

void UOpenGLRenderDevice::Buffer3ColoredVerts(FTransTexture** Pts) {
	FGLTexCoord *pTexCoordArray = &TexCoordArray[0][BufferedVerts];
	FGLColor *pColorArray = &ColorArray[BufferedVerts];
	FGLVertex *pVertexArray = &VertexArray[BufferedVerts];
	for (INT i = 0; i < 3; i++) {
		const FTransTexture* P = Pts[i];
		(pTexCoordArray + i)->u = P->U * TexInfo[0].UMult;
		(pTexCoordArray + i)->v = P->V * TexInfo[0].VMult;
		(pColorArray + i)->color = RGBA_MAKE(
			appRound(P->Light.X * 255.0f), 
			appRound(P->Light.Y * 255.0f), 
			appRound(P->Light.Z * 255.0f),
			255);
		(pVertexArray + i)->x = P->Point.X;
		(pVertexArray + i)->y = P->Point.Y;
		(pVertexArray + i)->z = P->Point.Z;
	}
	BufferedVerts += 3;
}

#ifdef INCLUDE_SSE_CODE
void UOpenGLRenderDevice::Buffer3ColoredVerts_SSE(FTransTexture** Pts) {
	FTransTexture **PtsPlus12B = Pts + 3;
	float f255 = 255.0f;
	__asm {
		mov ecx, this

		lea eax, [ecx]this.TexInfo

		movss xmm0, [eax]FTexInfo.UMult
		movss xmm1, [eax]FTexInfo.VMult

		movss xmm2, f255

		mov eax, [ecx]this.BufferedVerts

		lea ebx, [eax*8]
		lea esi, [ecx]this.TexCoordArray[0]
		mov esi, [esi]
		add ebx, esi

		lea edx, [eax*8]
		lea esi, [ecx]this.ColorArray
		add edx, esi

		lea edi, [eax + eax*2]
		lea edi, [edi*4]
		lea esi, [ecx]this.VertexArray
		add edi, esi

		mov eax, Pts

v_loop:
			mov esi, [eax]
			add eax, 4

			movss xmm3, [esi]FTransTexture.U
			mulss xmm3, xmm0
			movss [ebx]FGLTexCoord.u, xmm3
			movss xmm3, [esi]FTransTexture.V
			mulss xmm3, xmm1
			movss [ebx]FGLTexCoord.v, xmm3
			add ebx, TYPE FGLTexCoord

			push ebx

			mov ebx, 0xFF000000
			movss xmm3, [esi]FTransSample.Light + 0
			mulss xmm3, xmm2
			cvtss2si ecx, xmm3
			and ecx, 255
			or ebx, ecx
			movss xmm3, [esi]FTransSample.Light + 4
			mulss xmm3, xmm2
			cvtss2si ecx, xmm3
			and ecx, 255
			shl ecx, 8
			or ebx, ecx
			movss xmm3, [esi]FTransSample.Light + 8
			mulss xmm3, xmm2
			cvtss2si ecx, xmm3
			and ecx, 255
			shl ecx, 16
			or ebx, ecx
			mov [edx]FGLColor.color, ebx
			add edx, TYPE FGLColor

			mov ebx, [esi]FOutVector.Point.X
			mov [edi]FGLVertex.x, ebx
			mov ebx, [esi]FOutVector.Point.Y
			mov [edi]FGLVertex.y, ebx
			mov ebx, [esi]FOutVector.Point.Z
			mov [edi]FGLVertex.z, ebx
			add edi, TYPE FGLVertex

			pop ebx

			cmp eax, PtsPlus12B
			jne v_loop
	}
	BufferedVerts += 3;
}
#endif

void UOpenGLRenderDevice::Buffer3FoggedVerts(FTransTexture** Pts) {
	FGLTexCoord *pTexCoordArray = &TexCoordArray[0][BufferedVerts];
	FGLColor *pColorArray = &ColorArray[BufferedVerts];
	FGLVertex *pVertexArray = &VertexArray[BufferedVerts];
	for (INT i = 0; i < 3; i++) {
		const FTransTexture* P = Pts[i];
		(pTexCoordArray + i)->u = P->U * TexInfo[0].UMult;
		(pTexCoordArray + i)->v = P->V * TexInfo[0].VMult;
		FLOAT One_Minus_FogW = 1.0f - P->Fog.W;
		(pColorArray + i)->color = RGBA_MAKE(
			appRound(P->Light.X * 255.0f * One_Minus_FogW), 
			appRound(P->Light.Y * 255.0f * One_Minus_FogW), 
			appRound(P->Light.Z * 255.0f * One_Minus_FogW),
			255);

		(pColorArray + i)->specular = RGBA_MAKE(
			appRound(255 * P->Fog.X), 
			appRound(255 * P->Fog.Y),
			appRound(255 * P->Fog.Z),
			0);
		(pVertexArray + i)->x = P->Point.X;
		(pVertexArray + i)->y = P->Point.Y;
		(pVertexArray + i)->z = P->Point.Z;
	}
	BufferedVerts += 3;
}


void UOpenGLRenderDevice::BufferAdditionalClippedVerts(FTransTexture** Pts, INT NumPts) {
	INT FirstVert = BufferedVerts - 3;

	for (INT i = 3; i < NumPts; i++) {
		TexCoordArray[0][BufferedVerts].u = TexCoordArray[0][FirstVert].u;
		TexCoordArray[0][BufferedVerts].v = TexCoordArray[0][FirstVert].v;
		if (RenderFog) {
			ColorArray[BufferedVerts].color = ColorArray[FirstVert].color;
			ColorArray[BufferedVerts].specular = ColorArray[FirstVert].specular;
		}
		else if (ColorArrayEnabled) {
			ColorArray[BufferedVerts].color = ColorArray[FirstVert].color;
		}
		VertexArray[BufferedVerts].x = VertexArray[FirstVert].x;
		VertexArray[BufferedVerts].y = VertexArray[FirstVert].y;
		VertexArray[BufferedVerts].z = VertexArray[FirstVert].z;

		// gruenschloss:
		if (UseTruForm)
			checkAndSetNormal(&NormalArray[FirstVert], &NormalArray[BufferedVerts]);

		BufferedVerts++;

		TexCoordArray[0][BufferedVerts].u = TexCoordArray[0][BufferedVerts - 2].u;
		TexCoordArray[0][BufferedVerts].v = TexCoordArray[0][BufferedVerts - 2].v;
		if (RenderFog) {
			ColorArray[BufferedVerts].color = ColorArray[BufferedVerts - 2].color;
			ColorArray[BufferedVerts].specular = ColorArray[BufferedVerts - 2].specular;
		}
		else if (ColorArrayEnabled) {
			ColorArray[BufferedVerts].color = ColorArray[BufferedVerts - 2].color;
		}
		VertexArray[BufferedVerts].x = VertexArray[BufferedVerts - 2].x;
		VertexArray[BufferedVerts].y = VertexArray[BufferedVerts - 2].y;
		VertexArray[BufferedVerts].z = VertexArray[BufferedVerts - 2].z;

		// gruenschloss:
		if (UseTruForm)
			checkAndSetNormal(&NormalArray[BufferedVerts - 2], &NormalArray[BufferedVerts]);

		BufferedVerts++;

		const FTransTexture* P = Pts[i];
		TexCoordArray[0][BufferedVerts].u = P->U * TexInfo[0].UMult;
		TexCoordArray[0][BufferedVerts].v = P->V * TexInfo[0].VMult;
		if (RenderFog) {
			FLOAT One_Minus_FogW = 1.0f - P->Fog.W;
			ColorArray[BufferedVerts].color = RGBA_MAKE(
				appRound(P->Light.X * 255.0f * One_Minus_FogW), 
				appRound(P->Light.Y * 255.0f * One_Minus_FogW), 
				appRound(P->Light.Z * 255.0f * One_Minus_FogW),
				255);

			ColorArray[BufferedVerts].specular = RGBA_MAKE(
				appRound(255 * P->Fog.X), 
				appRound(255 * P->Fog.Y),
				appRound(255 * P->Fog.Z),
				0);
		}
		else if (ColorArrayEnabled) {
			ColorArray[BufferedVerts].color = RGBA_MAKE(
				appRound(P->Light.X * 255.0f), 
				appRound(P->Light.Y * 255.0f), 
				appRound(P->Light.Z * 255.0f),
				255);
		}
		FGLVertex &destVertexArray = VertexArray[BufferedVerts];
		destVertexArray.x = P->Point.X;
		destVertexArray.y = P->Point.Y;
		destVertexArray.z = P->Point.Z;

		// gruenschloss:
		if (UseTruForm)
			checkAndSetNormal((FGLNormal *) &(P->Normal), &NormalArray[BufferedVerts]); // conversion *should* work.

		BufferedVerts++;
	}
}


// Static variables.
INT		UOpenGLRenderDevice::NumDevices    = 0;
INT		UOpenGLRenderDevice::LockCount     = 0;
#ifdef WIN32
HGLRC		UOpenGLRenderDevice::hCurrentRC    = NULL;
HMODULE		UOpenGLRenderDevice::hModuleGlMain = NULL;
HMODULE		UOpenGLRenderDevice::hModuleGlGdi  = NULL;
TArray<HGLRC>	UOpenGLRenderDevice::AllContexts;
#else
UBOOL 		UOpenGLRenderDevice::GLLoaded	   = false;
#endif
UOpenGLRenderDevice::DWORD_CTMap_t UOpenGLRenderDevice::m_sharedZeroPrefixBindMaps[NUM_CTMap_TREES];
UOpenGLRenderDevice::QWORD_CTMap_t UOpenGLRenderDevice::m_sharedNonZeroPrefixBindMaps[NUM_CTMap_TREES];
UOpenGLRenderDevice::TexIdPoolVector_t UOpenGLRenderDevice::m_sharedNonZeroPrefixTexIdPool;
UOpenGLRenderDevice::TexPoolMap_t UOpenGLRenderDevice::m_sharedRGBA8TexPool;

// OpenGL function pointers.
#define GL_EXT(name) UBOOL UOpenGLRenderDevice::SUPPORTS##name=0;
#define GL_PROC(ext,ret,func,parms) ret (STDCALL *UOpenGLRenderDevice::func)parms;
#include "OpenGLFuncs.h"
#undef GL_EXT
#undef GL_PROC

/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/


